open Common
open List
open Types
open Printf
open Print
open Rew
open Rewrite

let debug = Flags.debug_env
let onlylocal = debug

(*
   this ugly variable is necessary because of
   - closures which take their own environment, we have more than one current
   environment which we can't handle
   - when going from type_inference to interpreter (via "eval") and again to type_inference 
    (via "help_interpreter_resolving_constraints")
*)
let noenv_check = ref false

let val_(v,_,_,_) = v
let typ_(_,t,_,_) = t
let fla_(_,_,f,_) = f
let nam_(_,_,_,s) = s

let set_val_(_,t,f,s) v = (v,t,f,s)
let set_typ_(v,_,f,s) t = (v,t,f,s)
let set_fla_(v,t,_,s) f = (v,t,f,s)
let set_nam_(v,t,f,_) s = (v,t,f,s)

let def_flags = { para_or_local = `Local ; is_lazy_and_alike = None ; is_constr = false ; is_attached_to_a_variable = true ; is_constraintable = true ; is_closed = false ; is_type = false ; inoutness = None ; pattern_matching__implying_closed_world = false }
let def_tconstr = { lower = subval ; greater = superval }

let is_constraint_op i = i = builtin_op__has_type || i = builtin_op__subtype || i = builtin_op__supertype

let local_i2rew f s i four = f (Local((i, s, !(nam_ four)), bpos)) four
let global_i2rew f pkg i four = f (Global((i, pkg, !(nam_ four)), bpos)) four

let env_iter_local f env = 
  Array.iteri (fun s locals -> Array.iteri (local_i2rew f s) locals) env.local

let env_iter f env =
  env_iter_local f env ; 
  iter (fun (pkg, global) -> Array.iteri (global_i2rew f pkg) global) env.global

let env_iter_non_closed f env =
  env_iter (fun var four -> 
    if not (fla_ four).is_closed then f var four
  ) env

let env_iter_non_closed_withenv f accu env = do0_withenv2 env_iter_non_closed f accu env

let env_iter_constr f = env_iter (fun var t -> f var (typ_ t))

let env_map f env =
  let env2 =  
    { env with local = 
      Array.mapi (fun s locals -> 
	Array.mapi (fun i four -> some_or (local_i2rew f s i four) four) locals
      ) env.local } in
  let env3 = 
    { env2 with global = 
      map (fun (pkg, global) -> pkg, Array.mapi (fun i four -> some_or (global_i2rew f pkg i four) four) global) env2.global } in
  env3

let env_map_non_closed f env =
  env_map (fun var four -> 
    if (fla_ four).is_closed then None else f var four
  ) env

let env_local_collect f env = do_collect Array.iteri (local_i2rew f 0) env.local.(0)
let env_local_map f env = Array.to_list (Array.mapi (local_i2rew f 0) env.local.(0))

let env_map_constr f env = 
  let wrap var four =
    match f var (typ_ four) with
    | Some constr -> 
	if debug && constr <> typ_ four then
	  printf "iter_index_constr: %s is now %s (was %s)\n" (rew2s var) (constr2s constr) (constr2s (typ_ four));
	Some (set_typ_ four constr)
    | None -> None
  in env_map wrap env

let env_scan_while_modifying f env =
  let renv = ref env in

  Array.iteri (fun stack _ -> 
    Array.iteri (fun i _ -> renv := local_i2rew (f !renv) stack i !renv.local.(stack).(i)) (!renv.local.(stack))
  ) !renv.local ;

  iter (fun (pkg, global) -> Array.iteri (fun i _ -> renv := global_i2rew (f !renv) pkg i (assoc pkg !renv.global).(i)) global) !renv.global ;

  !renv

let dump_forced env = 
  printf "--------------------------------------------------------------------------------\n";
  env_iter (fun v e ->
    if ! ! (val_ e) <> bottom then printf "%s = %s\n" (rew2s_ true false v) (rew2s ! ! (val_ e))
  ) env

let dumpt_forced_ env onlylocal = 
  printf "--------------------------------------------------------------------------------\n";
  env_iter (fun var four ->
    if not (is_global_var var && onlylocal) then
      let closeness = if (fla_ four).is_closed then " " else "*" in
      match typ_ four with
      |	TFree -> ()
      |	TExact(any) when is_anyval any -> ()
      |	constr -> printf "%20s%s !! %s\n" (rew2s var) closeness (constr2s constr)
  ) env ;
  flush stdout 
let dumpt_forced env = dumpt_forced_ env onlylocal

let dump  env = if debug then dump_forced  env
let dumpt env = if debug then dumpt_forced env

let dump_package pkg =
  printf "- %s::Vars -------------------------------------------------------------------------------\n" pkg.full_name;
  let display (v,tc,fl,s) =
    match tc with
    | TExact t when not (is_anyval t) -> 
	let closeness = if fl.is_closed then " " else "*" in
	printf "%20s%s !! %s\n" !s closeness (rew2s t) ;
	if is_overload ! !v then (	
	  match map snd (unoverload ! ! v) with
	  | [t2] -> if is_same t t2 then () else printf "%20s  := Overload(%s)\n" !s (rew2s t2)
	  | l -> printf "%20s  := Overload\n" !s ; iter (fun typ -> printf "                                  %s\n" (rew2s typ)) l 
	)
    | _ -> ()
  in
  let display_ p =
    let s = rew2s_ false false p.overload_var in
    if p.overload_is_open_type then
      printf "%20s  := Overload(%s)\n" s (rew2s p.overload_type)
    else
      printf "%20s%s !! %s\n" s " " (rew2s p.overload_type)
  in
  Array.iter display pkg.p_global ;
  iter display_ pkg.overloading_funcs

let constr_map f = function
  | TFree -> TFree
  | TExact(t) -> TExact(f t)
  | TConstr(t) -> TConstr({ lower = f t.lower ; greater = f t.greater })
let constr_map_withenv f env c = do_withenv constr_map f env c

let constr_iter f = function
  | TFree -> ()
  | TExact(t) -> f t
  | TConstr(t) -> f t.lower ; f t.greater

let constr_collect f = function
  | TFree -> []
  | TExact(t) -> f t
  | TConstr(t) -> f t.lower @ f t.greater

let constr2lower = function
  | TFree -> None
  | TExact(t) -> Some t
  | TConstr { lower = e } -> or2or_option e

let constr2greater = function
  | TFree -> None
  | TExact(t) -> Some t
  | TConstr { greater = e } -> allt2allt_option e

let is_tfree = function
  | TFree -> true
  | TExact _ -> false
  | TConstr c -> is_subval c.lower && is_superval c.greater

let constr2tconstr = function
  | TFree -> def_tconstr
  | TExact(t) -> { lower = t ; greater = t }
  | TConstr tconstr -> tconstr

let constr2lower_or_greater funct c =
  (if is_or funct then constr2lower else constr2greater) c

let pack_constr var = function
  | TFree -> []
  | TExact(_) -> []
  | TConstr(t) -> 
      let build s_op op = function
	| None -> []
	| Some v when is_same v var -> failwith "cleanup missing"
	| Some t -> [ Call(mark_as_val_simple (Global((op, "Pervasives::Builtin", s_op), bpos)), Sep(",", [ var; t ])) ]
      in
      build "!>" builtin_op__supertype (or2or_option t.lower) @ build "!<" builtin_op__subtype (allt2allt_option t.greater)

let unpack_constr_raw l = 
  fold_left_option (fun constrs -> function
    | Call(Call(Constr("Pervasives::Builtin", "Val", _), Global((op, _, _), _)), 
	   Sep(",", [Local((i, 0, _), _); e ])) ->
	let addit f = Some (update_assoc_by_with_default def_tconstr (=) f i constrs) in
	if op = builtin_op__subtype then
	  addit (fun constr -> { constr with greater = l2allt [ e ; constr.greater ] })
	else if op = builtin_op__supertype then
	  addit (fun constr -> { constr with lower = l2or [ e ; constr.lower ] })
	else None
    | _ -> None
  ) [] l

let unpack_constr body = 
  match body with
  | Sep(";", l) ->
      let l2, last = split_last l in
      (match unpack_constr_raw l2 with
      | Some constrs -> true, last, constrs
      | None -> false, body, [])
  | _ -> true, body, []

let pack_local_constr_with_kind kind name t =
  let var = Local((0, 0, if name = "" then "h" else name), bpos) in
  let t' = constr_map (rew_remap_locals 1) t in
  Call(Constr("Pervasives::Builtin", kind, bpos), SepL(";", [ [] ], pack_constr var t' @ [ var ]))

let pack_local_constr name t = pack_local_constr_with_kind "Var_forall" name t

let unpack_local_constr_raw kind = function
  | Call(Constr("Pervasives::Builtin", kind', _), SepL(";", [_], l)) when kind = kind' ->
      (match split_last l with
      |	[], Local((0, 0, name), _) -> Some(name, def_tconstr)
      |	l', Local((0, 0, name), _) ->
	  (match unpack_constr_raw l' with
	  | Some [ 0, constr ] -> 
	      Some(name, { lower = rew_remap_locals (-1) constr.lower ; greater = rew_remap_locals (-1) constr.greater })
	  | _ -> failwith "unpack_local_constr")
      |	_ -> None)
  | _ -> None

let unpack_local_constr_ kind e = map_option snd (unpack_local_constr_raw kind e)
let unpack_local_constr e = unpack_local_constr_ "Var_forall" e

let constr_remap_locals nb constr = constr_map (rew_remap_locals nb) constr

let remap_stack_locals nb ((v,t,f,_) as four) =
  if f.is_closed then four
  else (
    !v := rew_remap_locals nb ! ! v ;
    set_typ_ four (constr_remap_locals nb t)
  )
  
let push_stackt env nb flag =
  if debug then printf "push_stack\n";
  let global' = map_assoc_val (Array.map (remap_stack_locals 1)) env.global in
  let local' = Array.map (Array.map (remap_stack_locals 1)) env.local in
  let env' = { env with
    local = Array.append [| Array.init nb (fun _ -> reff bottom, TFree, flag, ref "") |] local' ;
    global = global'
  } in
  env'

let pop_stackt env =
  if debug then printf "pop_stack\n";
  let global' = map_assoc_val (Array.map (remap_stack_locals (-1))) env.global in
  let local' = Array.map (Array.map (remap_stack_locals (-1))) env.local in
  { env with local = array_shift local' ; global = global' }

let pop_stackt_ignore env = ignore (pop_stackt env)

let realloc_stack env stack nb flag =
  let local_stack = Array.copy env.local in
  let l = local_stack.(stack) in

  local_stack.(stack) <- 
    Array.append l (Array.init nb (fun _ -> reff bottom, TFree, { flag with is_attached_to_a_variable = false }, ref "")) ;

  Array.length l, { env with local = local_stack }

let realloc_global env package nb flag =
  let s = ref 0 in
  let globals = 
    update_assoc (fun global ->
      s := Array.length global ;
      Array.append global (Array.init nb (fun _ -> reff bottom, TFree, { flag with is_attached_to_a_variable = false }, ref ""))
    ) package env.global in
  !s, { env with global = globals }

let get_local_internal env i stack = env.local.(stack).(i)
let get_global_internal env i pkg = (assoc pkg env.global).(i)

let get_local_val   env i stack = val_ (get_local_internal env i stack)
let get_local_typ   env i stack = typ_ (get_local_internal env i stack)
let get_local_flags env i stack = fla_ (get_local_internal env i stack)
let get_local_name  env i stack = nam_ (get_local_internal env i stack)
let get_global_val   env i pkg = val_ (get_global_internal env i pkg)
let get_global_typ   env i pkg = typ_ (get_global_internal env i pkg)
let get_global_flags env i pkg = fla_ (get_global_internal env i pkg)
let get_global_name  env i pkg = nam_ (get_global_internal env i pkg)

let set_local_internal env i stack four =
  let local' = Array.copy env.local in
  local'.(stack) <- Array.copy local'.(stack) ;
  local'.(stack).(i) <- four ;
  { env with local = local' }

let set_global_internal env i pkg four =
  { env with global = 
    update_assoc (fun global ->
      let global' = Array.copy global in
      global'.(i) <- four ;
      global'
    ) pkg env.global }

let set_local_typ   new_typ env i stack = set_local_internal env i stack (set_typ_ (get_local_internal env i stack) new_typ)
let set_local_flags new_fla env i stack = set_local_internal env i stack (set_fla_ (get_local_internal env i stack) new_fla)
let set_global_typ   new_typ env i pkg = set_global_internal env i pkg (set_typ_ (get_global_internal env i pkg) new_typ)
let set_global_flags new_fla env i pkg = set_global_internal env i pkg (set_fla_ (get_global_internal env i pkg) new_fla)

let dispatch_local_or_global env loc glob = function
  | Local((i,stack,n),pos) -> (try loc env i stack with _ -> internal_error (sprintf "bad local variable %s%s" n (Info.pos2sfull pos)))
  | Global((i,pkg,n),pos) -> (try glob env i pkg with _ -> internal_error (sprintf "bad global variable %s::%s%s" pkg n (Info.pos2sfull pos)))
  | e -> internal_error (sprintf "%s is not a variable%s" (rew2s e) (Info.pos2sfull (getpos e)))

let get_var_raw   env = dispatch_local_or_global env get_local_internal get_global_internal
let get_var_val   env = dispatch_local_or_global env get_local_val get_global_val
let get_var_flags env = dispatch_local_or_global env get_local_flags get_global_flags
let get_var_name  env = dispatch_local_or_global env get_local_name get_global_name
let get_var_typ   env = dispatch_local_or_global env get_local_typ get_global_typ

let set_var_typ_bare_raw env var new_typ = dispatch_local_or_global env (set_local_typ new_typ) (set_global_typ new_typ) var
let set_var_flags        env var new_fla = dispatch_local_or_global env (set_local_flags new_fla) (set_global_flags new_fla) var

let merge_inoutness io1 io2 = prefer_some max_io io1 io2

let rew_collect_with_io f (io, e) =
  let rec rew_collect_with_io_rec io depth e =
    match e with
    | Call(Constr(_, "Out",     _), e) -> rew_collect_with_io_rec (inv_io io) depth e
    | Call(Constr(_, "Inout",   _), e) -> rew_collect_with_io_rec Inout depth e
    | Call(Constr(_, "Exact",   _), e) -> rew_collect_with_io_rec (max_io Exact io) depth e
    | Call(e1, e2) -> rew_collect_with_io_rec io depth e1 @ rew_collect_with_io_rec io depth e2
    | Sep(_,l) -> collect (rew_collect_with_io_rec io depth) l
    | SepL(_,_,l) -> collect (rew_collect_with_io_rec io (depth+1)) l
    | Func(_,para,body) -> rew_collect_with_io_rec io (depth+1) body @ collect (rew_collect_with_io_rec (inv_io io) (depth+1)) para
    | _ -> f(io,depth,e)
  in rew_collect_with_io_rec io 0 e

let merge_flags_no_inoutness flag1 flag2 =
  { para_or_local = 
    (match flag1.para_or_local, flag2.para_or_local with
    | `Parameter, _ -> `Parameter
    | _, `Parameter -> `Parameter
    | _ -> `Local) ;
    is_lazy_and_alike = merge_some
      (fun a b -> if a = b then a else failwith "merge_flags")
      flag1.is_lazy_and_alike flag2.is_lazy_and_alike ;
    is_constr = flag1.is_constr || flag2.is_constr ;
    is_closed = flag1.is_closed && flag2.is_closed ;
    is_type = flag1.is_type && flag2.is_type ;
    is_attached_to_a_variable = flag1.is_attached_to_a_variable ;
    is_constraintable = flag1.is_constraintable ;
    pattern_matching__implying_closed_world = flag1.pattern_matching__implying_closed_world ;
    inoutness = flag1.inoutness
  }

let merge_flags flag1 flag2 =
  let flag = merge_flags_no_inoutness flag1 flag2 in
  { flag with inoutness = merge_inoutness flag1.inoutness flag2.inoutness }

let set_var_inoutness env io v =
  if debug then printf "set_var_inoutness %s %s\n" (rew2s v) (io2s (some io));
  let flags = get_var_flags env v in
  let io' = merge_some max_io flags.inoutness io in
  io', set_var_flags env v { flags with inoutness = io' } 

let rew2best_io env io e =
  if is_var e then
    let io', env' = set_var_inoutness env (Some io) e in
    (some io', e), env'
  else rew2io e, env


let rec check_expr_ is_value depth env var expr =
  let err s = 
    dump_forced env ; 
    failwith
      (match var with 
      | None -> sprintf "check_expr: %s in %s\n" s (rew2s expr)
      | Some var -> sprintf "check_expr(%s): %s in %s\n" (rew2s var) s (rew2s expr))
  in

  if Flags.check then
    (* skip fully bottom values *)
    if is_value && is_same expr bottom then () else
    let rec check_once depth e =
      match e with
      | Call(Constr("Pervasives::Builtin", "Overload", _), _) -> 
	  if not is_value then err "bad Overload construct" else
	  iter (fun (v,t) -> check_expr_ true depth env var v ; check_expr_ false depth env var t) (unoverload e) ; true
      |	Call(Constr("Pervasives::Builtin", "Val", _), e') when not (is_var e') || is_value -> err (sprintf "bad Val construct %s" (rew2s e))
      (* |	Call(Constr("Pervasives::Builtin", "Type", _), e') when not (is_var e') -> err (sprintf "bad Type construct %s" (rew2s e)) *)
      |	Call(Constr("Pervasives::Builtin", "Or", _), Sep _) -> for_all (check_once depth) (or2l e)
      |	Call(Constr("Pervasives::Builtin", "All_T", _), Sep _) -> for_all (check_once depth) (allt2l e)
      | Local((i, stack, n), pos) when stack >= depth -> let _ = get_var_raw env (Local((i, stack - depth, n), pos)) in true
      | Global _ -> let _ = get_var_raw env e in true
      |	Constr("Pervasives::Builtin", "Bottom", _) -> err "bottom found"
      |	Func(_, [], _) -> err "function doesn't have parameter"
      |	Sep(_, []) -> err "empty Sep"
      |	Sep(";", [_]) -> err "one element sequence"
      |	_ -> false
    in
    let _ = rew_iter_once check_once depth expr in
    let vars = rew_collect_all_accessible_vars expr in
    iter (fun v ->
      try ignore (get_var_raw env v) 
      with _ -> err (sprintf "bad variable %s" (rew2s v))
    ) vars

let check_expr is_value env var expr = check_expr_ is_value 0 env var expr

let check_env env =
  let check var (val_,typ,_,_) =
    iter (fun v ->
      try ignore (get_var_raw env v) with _ -> 
	dumpt_forced env ; 
	failwith (sprintf "check_env(%s): bad variable %s in constraint %s\n" (rew2s var) (rew2s v) (constr2s typ))
    ) (constr_collect rew_collect_all_accessible_vars typ) ;
    check_expr true env (Some var) ! ! val_
  in
  if Flags.check && not !noenv_check then env_iter check env

let rew_is_closed_t env e = 
  let rec f depth = function
    | Call(e1, e2) -> f depth e1 && f depth e2
    | Sep(_,l) -> for_all (f depth) l
    | SepL(_,_,l) -> for_all (f (depth+1)) l
    | Func(_, para, body) -> for_all (f (depth+1)) para && f (depth+1) body
    | Local((_,stack,_), _) -> stack < depth
    | Global _ as v -> (get_var_flags env v).is_closed
    | _ -> true
  in f 0 e

let rew_is_closed_v _env e = 
  let rec f depth = function
    | Call(e1, e2) -> f depth e1 && f depth e2
    | Sep(_,l) -> for_all (f depth) l
    | SepL(_,_,l) -> for_all (f (depth+1)) l
    | Func(_, para, body) -> for_all (f (depth+1)) para && f (depth+1) body
    | Local((_,stack,_), _) -> stack < depth
    | _ -> true
  in not (is_same e bottom) && f 0 e

let constr_is_closed env = function
  | TExact(t) -> rew_is_closed_t env t
  | _ -> false

let update_closeness_t env var constr =
  let flags = get_var_flags env var in
  let value = ! !(get_var_val env var) in
  if flags.is_closed then env else (
    (* suppose the variable is closed (for recursive functions) *)
    let env2 = set_var_flags env var { flags with is_closed = true } in
    if constr_is_closed env2 constr && rew_is_closed_v env2 value then (
      if debug then printf "setting %s is_closed (%s / %s)\n" (rew2s var) (rew2s value) (constr2s constr); 
      env2
    ) else (* nope *) env
  ) 

let update_closeness_v env var value = 
  let flags = get_var_flags env var in
  let constr = get_var_typ env var in
  if flags.is_closed then env else (
    (* suppose the variable is closed (for recursive functions) *)
    let env2 = set_var_flags env var { flags with is_closed = true } in
    if rew_is_closed_v env2 value && constr_is_closed env2 constr then (
      if debug then printf "setting %s is_closed (%s / %s)\n" (rew2s var) (rew2s value) (constr2s constr);
      env2
    ) else (* nope *) env
  )

let get_var env v =
  check_env env ;
  let (val_, typ_, flags_, _) = get_var_raw env v in
  ! !val_, typ_, flags_
  
let set_var_val env v value = 
  if debug then printf "setting %s val to %s\n" (rew2s v) (rew2s value);
  let env2 = update_closeness_v env v value in
  !(get_var_val env2 v) := value ;
  env2

let follow_var env v =
  let rec f l v =
    match get_var_typ env v with
    | TExact(v') when is_var v' -> 
	if mem v' l then l else f (v' :: l) v'
    | _ -> l
  in f [v] v

let rew_merge_flags merge env var expr =
  fold_left (fun env v' -> 
    let env2 = set_var_flags env v' (merge (get_var_flags env v') (get_var_flags env var)) in
    if debug then printf "merging flags of %s and %s\n" (rew2s v') (rew2s var) ;
    env2
  ) env (rew_collect_all_accessible_vars_no_val expr)

let pre_set_var_typ env v = function
  | TExact(e) ->
      if Flags.check then check_no_mixed_var_in_or e ;
      let env2 = if (get_var_flags env v).para_or_local = `Parameter then rew_merge_flags merge_flags_no_inoutness env v e else env in
      let c = if is_var e && exists (is_same v) (follow_var env e) then TFree else TExact(e) in
      c, env2
  | c -> c, env

let set_var_typ_raw env v constr = 
  if debug then printf "setting %s typ to %s\n" (rew2s v) (constr2s constr);

  constr_iter (check_expr false env (Some v)) constr ;

  let constr', env2 = pre_set_var_typ env v constr in
  let env3 = update_closeness_t env2 v constr' in

  set_var_typ_bare_raw env3 v constr'

let set_var_name env v =
  let name = 
    match v with
    | Local((_,_,name), _)
    | Global((_,_,name), _) -> name
    | _ -> failwith "set_var_name"
  in
  if name <> "" then get_var_name env v := name


let allocate_new_vars env nb flags = function
  | Local((_,stack,n),pos) ->
      let offset, env2 = realloc_stack env stack nb flags in
      let vars = map (fun i -> Local((offset + i, stack, n ^ "_" ^ string_of_int i), pos)) (range 0 nb) in
      iter (set_var_name env2) vars ;
      vars, env2
  | Global((_,_,n),pos) ->
      let offset, env2 = realloc_global env env.current_package nb flags in
      let vars = map (fun i -> Global((offset + i, env.current_package, n ^ "_" ^ string_of_int i), pos)) (range 0 nb) in
      iter (set_var_name env2) vars ;
      vars, env2
  | _ -> failwith "allocate_new_var"

let allocate_new_vars_in_depth env nb flags = function
  | Local((_,stack,n),pos) ->
      if stack = 0 then failwith "allocate_new_vars_in_depth";
      let offset, env2 = realloc_stack env (stack-1) nb flags in
      map (fun i -> Local((offset + i, stack-1, n ^ "_" ^ string_of_int i), pos)) (range 0 nb), env2
  | Global((_,_,n),pos) ->
      let offset, env2 = realloc_stack env 0 nb flags in
      map (fun i -> Local((offset + i, 0, n ^ "_" ^ string_of_int i), pos)) (range 0 nb), env2
  | _ -> failwith "allocate_new_var"

let var2lower_or_greater f env var =
  let seen = ref [] in
  let rec var2 var =
    (* detect circularity, dead loops *)
    if exists (is_same var) !seen then None else (
    seen := var :: !seen ;
    map_optionoption (rew_map_leaf__option (function 
      | depth, (Local((_,stack,_),_) as v) when stack >= depth -> 
	  var2 (rew_remap_locals (-depth) v)
      | _, e -> Some e)) (f (get_var_typ env var)))
  in
  let e = var2 var in
  e

let rec collect_allvars__rec env l_in var =
  let l = constr_collect rew_collect_all_accessible_vars (get_var_typ env var) in
  let l' = difference_ is_same l l_in in
  fold_left (collect_allvars__rec env) (l_in @ l') l'


let rew_is_type env t = 
  let rec is = function
  | Call(Constr("Pervasives::Builtin", "Overload", _), _) -> false (* overloading is not allowed on types *)
  | Constr("Pervasives::Builtin", "Type", _)
  | Constr("Pervasives::Builtin", "Or", _)
  | Constr("Pervasives::Builtin", "Or_range", _)
  | Constr("Pervasives::Builtin", "Abstract", _) -> true

  | Call(Global((i,"Pervasives::Builtin",_),_), _) when is_constraint_op i -> false

  | Call(e1, e2) -> is e1 || is e2
  | Func(_, para, body) -> exists is para || is body
  | Sep(_, l)
  | SepL(_, _, l) -> exists is l
  | External _ -> false
  | Global _ as v -> (get_var_flags env v).is_type
  | _ -> false
  in is t


let rec is_abstract3 env = function
  | Call(c, _) when not (is_special_constr c) -> is_abstract3 env c
  | Local _ as v ->
      (match get_var_typ env v with
      |	TExact t | TConstr { greater = t } -> is_abstract2 t
      |	TFree -> false)
  | e -> is_abstract2 e
  
let is_var_constraintable env v = is_var v && (get_var_flags env v).is_constraintable

let texact = function
  | TExact t -> t
  | _ -> failwith "texact"

let find_created_vars_now_dead env =
  let created_vars_now_dead = ref [] in
  let vars = ref [] in

  let add var t =
    if some(var_depth var) > 0 then
      let a = if (fla_ t).is_attached_to_a_variable then vars else created_vars_now_dead in
      a := var :: !a
  in
  env_iter_non_closed add env ;

  let created_vars_still_alive = fold_left (collect_allvars__rec env) [] !vars in
    
  let d = difference_ is_same !created_vars_now_dead created_vars_still_alive in
  d



let find_still_alive_local_vars env =
  let local_vars = env_local_map (fun v _ -> v) env in

  let vars = ref [] in
  let add var _t = if some(var_depth var) > 0 then vars := var :: !vars
  in
  env_iter_non_closed add env ;

  let local_vars2 = filter (fun var ->
    let l = constr_collect rew_collect_all_accessible_vars_no_val (get_var_typ env var) in
    let l' = filter (fun v -> some (var_depth v) > 0) l in
    l' <> []
  ) local_vars in

  let local_vars3 = fold_left (collect_allvars__rec env) local_vars2 !vars in
  filter (fun v -> some (var_depth v) = 0) local_vars3

let rew_map_once_adapting_env f env e = 
  let rec ff env e =
    match e with
    | Constr _ when is_special_constr e -> e, env (* for rew_map_once_adapting_env called with Builtin::Or *)
    | Call(Constr("Pervasives::Builtin", "Val", _), _) -> some_or (f env e) (e, env)
    | SepL(s,locals,l) -> 
	let env2 = push_stackt env (length locals) { def_flags with para_or_local = `Local ; is_constraintable = false } in
	let l', env3 = map_withenv ff env2 l in
	if length locals <> Array.length env3.local.(0) then internal_error "rew_map_once_adapting_env";
	SepL(s, locals, l'), pop_stackt env3
    | _ ->
	match f env e with
	| Some (e', env') -> e', env'
	| None ->
	    match e with
	    | Call(c, e2) when is_special_constr c -> let e2', env' = ff env e2 in Call(c, e2'), env'
	    | Call(e1, e2) -> 
		let e1', env2 = ff env e1 in
		let e2', env3 = ff env2 e2 in
		Call(e1', e2'), env3
	    | Sep(s,l) -> let l', env' = map_withenv ff env l in Sep(s, l'), env'
	    | Func(nb, para, body) -> 
		let env2 = push_stackt env nb { def_flags with para_or_local = `Local ; is_constraintable = false } in
		let para', env3 = map_withenv ff env2 para in
		let body', env4 = ff env3 body in
		let nb' = Array.length env4.local.(0) in
		Func(nb', para', body'), pop_stackt env4
	    | e -> e, env
  in ff env e

let rew_remove_type_constraints e =
  let rec ff e =
    match e with
    | Call(Constr("Pervasives::Builtin", "Overload", _) as o, Sep(",", l)) -> 
	let l' = map (function 
	  | Sep(",", [ value ; typ ]) -> Sep(",", [ ff value ; typ ])
	  | _ -> failwith "bad overload"
        ) l in
	Call(o, Sep(",", l'))
    | Call(Global((i,"Pervasives::Builtin",_),_), _) ->
	if is_constraint_op i then unit else e
    | Call(e1, e2) -> Call(ff e1, ff e2)
    | Sep(s,l) -> Sep(s, map ff l)
    | SepL(s,locals,l) -> SepL(s, locals, map ff l)
    | Func(nb, para, body) -> Func(nb, para, ff body)
    | _ -> e
  in ff e


let traverse_contraints env in_expr out_expr =
  let rec traverse_var io accu var =
    let io_list() = if var_depth var = Some 0 then [ var_indice var, io ] else [] in
    try
      let _ = assoc_by is_same var accu in
      io_list(), accu
    with Not_found ->
      match get_var_typ env var with
      |	TFree -> io_list(), accu

      |	TExact(t) -> 
	  if var_depth var = Some 0 then
	    (* special case, let it flow through *)
	    traverse_expr io accu t
	  else
	    (* ensure it is kept *)
	    let accu1 = (var, []) :: accu in
	    let l, accu2 = traverse_expr Inout accu1 t in
	    io_list(), update_assoc_by is_same (fun _ -> l) var accu2

      | TConstr(t) -> 
	  let accu1 = (var, []) :: accu in
	  let l1, accu2 = traverse_expr Out accu1 t.lower in
	  let l2, accu3 = traverse_expr In accu2 t.greater in
	  let l = l1 @ l2 in
	  io_list(), update_assoc_by is_same (fun _ -> l) var accu3
		    
  and traverse_expr io accu e = 
    do_withenv rew_collect_with_io (fun accu (io, depth, v) ->
      match var_depth v with
      |	Some stack when stack >= depth ->
	let io_list', accu2 = traverse_var io accu (rew_remap_locals (-depth) v) in
	io_list', accu2
      |	_ -> [], accu
    ) accu (io, e)
  in

  let accu0 = [] in
  let l1, accu1 = collect_withenv (traverse_expr In) accu0 in_expr in
  let l2, accu2 = collect_withenv (traverse_expr Out) accu1 out_expr in
  let accu3 =     
    env_iter_non_closed_withenv (fun accu v t -> 
      if some(var_depth v) > 0 && (fla_ t).is_attached_to_a_variable then 
	let _, accu2 = traverse_var In (* not important as we discard the result *) accu v in accu2
      else
	accu) accu2 env in
  l1 @ l2, accu3


(* simplified comparison:
  - variables not handled (so no environment needed)
  - simplified function accepted: (_ -> whatever)
 *)
let rec raw_compare env a b = 
  let unpack depth l =
    let constrs, t = split_last l in
    let senv = fold_left (fun senv -> function
      | Call(Global((g,"Pervasives::Builtin",_),_), Sep(",", [ Local((i,stack,_),_) ; t ])) ->
	  let var = i, stack + depth in
	  let lo,gr = 
	    if g = builtin_op__subtype then [], [t] else
	    if g = builtin_op__supertype then [t], [] else
	    failwith "raw_compare.unpack" in
	  (try update_assoc_by (=) (fun c -> { lower = l2or (c.lower :: lo) ; greater = l2allt (c.greater :: gr) }) var senv
	  with Not_found -> (var, { lower = l2or lo ; greater = l2allt gr }) :: senv)
      |	_ -> failwith "raw_compare.unpack"
    ) [] constrs in
    t, senv
  in
  let var2constr i stack senv =
    try assoc (i, stack) senv with Not_found -> def_tconstr
  in
  let rec raw_is_greater allow_closing_world loc_env a b =
    try raw_compare_ allow_closing_world loc_env a b >= 0 with Not_comparable -> false
  and raw_compare_ allow_closing_world (depth_a, senv_a, depth_b, senv_b as loc_env) a b =
    if debug then printf "raw_compare_ %s and %s\n" (rew2s a) (rew2s b) ;
    match a,b with
      (* same value *)
    | Float(a,_), Float(b,_)           -> if a = b then 0 else raise Not_comparable
    | String(a,_), String(b,_)         -> if a = b then 0 else raise Not_comparable
    | Constr(qa,ea,_), Constr(qb,eb,_) -> if qa = qb && ea = eb then 0 else raise Not_comparable
  
    | Func(1, [ Local((0,0,_), _) ], body_a), Func(1, [ Local((0,0,_), _) ], body_b)
    | Func(1, [ Local((0,0,_), _) ], body_a), Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], body_b) ->
  	raw_compare_ true (depth_a+1,senv_a,depth_b+1,senv_b) body_a body_b
    | Func(0, [any], body_a), Func(1, [ Local((0,0,_), _) ], body_b) when is_anyval any ->
  	raw_compare_ true (depth_a+1,senv_a,depth_b+1,senv_b) body_a body_b

    | Func(0, pa, ba), Func(0, pb, bb) ->
	forall2_compare (raw_compare_ true (depth_a+1,senv_a,depth_b+1,senv_b)) [ l2sep_ pb ; ba ] [ l2sep_ pa  ; bb ]

      (* HACK: functions are considered equal *)
    | Func _, Func _ ->
	raise Not_comparable

    | Sep(";", l), _ ->
	let a', senv_a_ = unpack depth_a l in
	raw_compare_ true (depth_a, senv_a_ @ senv_a, depth_b, senv_b) a' b
    | _, Sep(";", l) ->
	let b', senv_b_ = unpack depth_b l in
	raw_compare_ true (depth_a, senv_a, depth_b, senv_b_ @ senv_b) a b'

    | Sep(sa,la), Sep(sb,lb) -> if sa = sb then forall2_compare (raw_compare_ true loc_env) la lb else raise Not_comparable

      (* special case for "Object", is it needed? *)
    | Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])), _ -> 0
    | _, Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) -> 0

    | Call(Constr("Pervasives::Builtin", "Var_forall", _), _), _ ->
	(match unpack_local_constr a with
	| None -> internal_error "raw_compare_"
	| Some { lower = lo ; greater = gr } ->
	    if raw_is_greater allow_closing_world loc_env gr b && raw_is_greater allow_closing_world loc_env b lo 
	    then 0
	    else raise Not_comparable)

    | _, Call(Constr("Pervasives::Builtin", "Var_forall", _), _) ->
	- (raw_compare_ allow_closing_world (depth_b,senv_b,depth_a,senv_a) b a)
  
      (* special range cases *)
    | Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min, _) ; Float(max, _) ])), Float(nb, _) ->
        if is_int nb && min <= nb && nb <= max then 1 else raise Not_comparable
    | Float(nb, _), Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min, _) ; Float(max, _) ])) ->
        if is_int nb && min <= nb && nb <= max then -1 else raise Not_comparable
    | Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min_a, _) ; Float(max_a, _) ])), 
      Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min_b, _) ; Float(max_b, _) ])) ->
        if min_a = min_b && max_b = max_a then 0 else 
        if min_a <= min_b && max_b <= max_a then 1 else 
        if min_b <= min_a && max_a <= max_b then -1 else 
        raise Not_comparable

      (* special string case *)
    | Constr("Pervasives::Builtin", "String", _), String _ -> 1
    | String _, Constr("Pervasives::Builtin", "String", _) -> -1
    | Constr("Pervasives::Builtin", "String", _), Call(Constr("Pervasives::Builtin", "Or", _), _) when for_all is_string (or2l b) -> 1
    | Call(Constr("Pervasives::Builtin", "Or", _), _), Constr("Pervasives::Builtin", "String", _) when for_all is_string (or2l a) -> -1

      (* special char case *)
    | Constr("Pervasives::Builtin", "Char", _), String _ when is_char b -> 1
    | String _, Constr("Pervasives::Builtin", "Char", _) when is_char a -> -1
    | Constr("Pervasives::Builtin", "Char", _), Call(Constr("Pervasives::Builtin", "Or", _), _) when for_all is_char (or2l b) -> 1
    | Call(Constr("Pervasives::Builtin", "Or", _), _), Constr("Pervasives::Builtin", "Char", _) when for_all is_char (or2l a) -> -1
  
    | Constr("Pervasives::Builtin", "Numbers", _), Float _ -> 1
    | Constr("Pervasives::Builtin", "Numbers", _), Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float _ ; Float _ ])) -> 1
    | Float _, Constr("Pervasives::Builtin", "Numbers", _) -> -1
    | Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float _ ; Float _ ])), Constr("Pervasives::Builtin", "Numbers", _) -> -1

      (* small hack: rewrite (X|&|Y)(...) in X(...) |&| Y(...) *)
    | Call(Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", la)), pa), Call(c, _) when is_orallt c ->
	raw_compare_ allow_closing_world loc_env (l2allt (map (fun ca -> Call(ca, pa)) la)) b
    | Call(c, _), Call(Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", lb)), pb) when is_orallt c ->
	raw_compare_ allow_closing_world loc_env a (l2allt (map (fun cb -> Call(cb, pb)) lb))

    | Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", la)), Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", lb)) ->
	exists2_compare 1 (raw_compare_ allow_closing_world loc_env) la lb
    | Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", la)), Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", lb)) ->
	let la' = filter (fun e -> not (is_strict e)) (collect allt_expand_call la) in
	let lb' = filter (fun e -> not (is_strict e)) (collect allt_expand_call lb) in
	exists2_compare (-1) (raw_compare_ allow_closing_world loc_env) la' lb'

      (* Or must be first, All_T next *)
      (* needed for A|&|B | C > A|&|B *)
    | Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", la)), _   -> 
	(* answer is either -1, 1 or Not_comparable *)
	(* -1 can happen because of 0|2 !< 0....2 *)
	if for_all (fun a -> raw_is_greater allow_closing_world loc_env b a) la then -1 else
	if exists (fun a -> raw_is_greater allow_closing_world loc_env a b) la then  1 else
	if allow_closing_world && can_close_the_world true env a b then  1 else raise Not_comparable

    | _, Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", _)) -> 
	- (raw_compare_ allow_closing_world (depth_b,senv_b,depth_a,senv_a) b a)

    | _, Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", lb)) -> 
	if lb = [] then -1 else
	if exists (fun b -> raw_is_greater allow_closing_world loc_env a b) lb then 1 else 
	if allow_closing_world && can_close_the_world true env b a then -1 else
	raise Not_comparable

    | Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", _)), _ ->
	- (raw_compare_ allow_closing_world (depth_b,senv_b,depth_a,senv_a) b a)

    | Call(Constr("Pervasives::Builtin", "Abstract", _), ca), Call(Constr("Pervasives::Builtin", "Abstract", _), cb) ->
	if is_same ca cb then 0 else
	if allow_closing_world && can_close_the_world true env b a then -1 else
	if allow_closing_world && can_close_the_world true env a b then  1 else
	raise Not_comparable
 
    | _, Call(Constr("Pervasives::Builtin", "Abstract", _), _) ->
	if allow_closing_world && can_close_the_world true env b a then -1 else raise Not_comparable
    | Call(Constr("Pervasives::Builtin", "Abstract", _), _), _ ->
	if allow_closing_world && can_close_the_world true env a b then  1 else raise Not_comparable
  
    | Call(fa,pa), Call(fb,pb) -> forall2_compare (raw_compare_ true loc_env) [fa ; pa] [fb ; pb]

    | Local((i_a,stack_a,_),_), Local((i_b,stack_b,_),_) ->
	let c_a = var2constr i_a (stack_a+depth_a) senv_a in
	let c_b = var2constr i_b (stack_b+depth_b) senv_b in
	if raw_is_greater allow_closing_world loc_env c_a.greater c_b.lower &&
	   raw_is_greater allow_closing_world loc_env c_b.greater c_a.lower then
	  0
	else raise Not_comparable

    | Local((i,stack,_),_), _ ->
	let c_a = var2constr i (stack+depth_a) senv_a in
	if raw_is_greater allow_closing_world loc_env c_a.greater b && raw_is_greater allow_closing_world loc_env b c_a.lower 
	then 0
	else raise Not_comparable

    | _, Local _ -> - (raw_compare_ allow_closing_world (depth_b,senv_b,depth_a,senv_a) b a)

    |      Ref _, _ | _, Ref _
    |     SepL _, _ | _, SepL _
    |   Global _, _ | _, Global _
    | External _, _ | _, External _ -> failwith (sprintf "raw_compare (%s vs %s)" (rew2s a) (rew2s b))
  
    | _, _ -> raise Not_comparable
  in
  raw_compare_ false (0, [], 0, []) a b

and can_close_the_world want_upper env abstract concrete =
  is_abstract2_bounded abstract &&
  try
    let _ = close_the_world want_upper env abstract concrete in
    true
  with Not_found -> false

and close_the_world want_upper env abstract concrete =
  let upper_than = if want_upper then Lattice.greatest_upper_than else Lattice.least_upper_than in
  let is_strict, abstract = 
    let strict, l' = partition is_strict (allt2l abstract) in
    if l' = [] 
    then raise Not_found (* occurs when the ensure_greater tries to enumerate the allt, fail! *)
    else strict <> [], l2allt l'
  in
  match concrete with
  | _ ->      
      let close_one one_abstract = 
	let r = assoc_by is_same one_abstract env.virtual_classes.class2info in
	match upper_than (raw_compare env) r.instanced_by concrete with
	| [] -> 
	    if is_strict then raise Not_found else
	    (* if non strict, we can use A|B !< &C if A !< &C and B !< &C *)
	    let ll = map (upper_than (raw_compare env) r.instanced_by) (or2l concrete) in
	    if for_all ((<>) []) ll then l2or (map l2allt ll) else raise Not_found
	| l -> l2allt l
      in

      match Lattice.least_upper_than (raw_compare env) env.virtual_classes.lattice abstract with
      |	[] -> raise Not_found
      |	[e] -> close_one e
      |	l ->
          (* "l_abstract" has more than one element when one more than one instance is needed.
	     eg: 1 !< &Eq |&| &Addable *)
	  (* special case for sake of simplicity and speed *)
	  let l2 = map close_one l in
          let l3 = fold_left1 (intersection_by is_same) (map allt2l l2) in
	  l2allt l3
	

let class2info_add_instance env virtual_classes (class_, instance) =
  update_assoc_by is_same (fun class_info ->
    let instanced_by = Lattice.add (raw_compare env) class_info.instanced_by instance in
    { class_info with instanced_by = instanced_by }
  ) class_ virtual_classes

let env_add_instances env instances =
  let class2info = fold_left (class2info_add_instance env) env.virtual_classes.class2info instances in
  { env with virtual_classes = { env.virtual_classes with class2info = class2info } }


let env_add_class env class_ methods =
  let class_info = { instanced_by = Lattice.create() ; class_methods = methods } in

  let class_, latt = Lattice.add_ (raw_compare env) env.virtual_classes.lattice class_ in

  { env with virtual_classes = 
    { env.virtual_classes with
        lattice = latt ; 
        class2info = (class_, class_info) :: env.virtual_classes.class2info ;
    } }


let new_env_ pkgs =
  let env = {
    local = [| [||] |] ;
    global = map (fun p -> p.full_name, p.p_global) pkgs ;
    virtual_classes = { lattice = Lattice.create() ; class2info = [] ; class4pkg = { p_instances = [] ; p_classes = [] }  } ;
    overloading_funcs4pkg = [] ;
    current_package = "NONE" ;
  } in

  let env = fold_left (fun env pkg -> 
    fold_left (fun env (class_, methods) ->
      env_add_class env class_ methods.p_methods
    ) env pkg.p_virtual_classes.p_classes
  ) env pkgs in

  let env = fold_left (fun env pkg ->
    env_add_instances env pkg.p_virtual_classes.p_instances
  ) env pkgs in

  let env = fold_left (fun env pkg ->
    fold_left (fun env o ->
      let instances = unoverload ! !(get_var_val env o.overload_var) in
      let env = set_var_val env o.overload_var (overload (instances @ [ o.overload_val, o.overload_type ])) in
      if o.overload_is_open_type then env else
      match get_var_typ env o.overload_var with
      | TExact(prev) -> 
	  set_var_typ_raw env o.overload_var (TExact (l2allt [ prev ; o.overload_type]))
      |	_ -> internal_error "handling overloading_funcs"
    ) env pkg.overloading_funcs
  ) env pkgs in

  env

let new_env packages l_globals package_name =
  let pkgs = map snd packages in
  let globals = Array.of_list (map (fun n -> reff bottom, TFree, def_flags, ref n) l_globals) in

  if package_name = "Pervasives::Builtin" then
    begin
      (* adding some things that can't be done any other way *)

      let r = val_ (globals.(builtin_op__assign)) in
      !r := External(":=", bpos) ;

      iter (fun i ->
	globals.(i) <- set_fla_ globals.(i) { def_flags with is_closed = true }
      ) [ builtin_op__assign ; builtin_op__has_type ; builtin_op__supertype ; builtin_op__subtype ]

    end ;

  let env = new_env_ pkgs in

  { env with 
    global = (package_name, globals) :: env.global ;
    current_package = package_name }
