open Common
open Print
open Printf
open Types
open List
open Rew
open Rewrite
open Info
open Env

let debug = Flags.debug_interpreter
let during_type_inference = ref true
let (args : string list ref) = ref []
let (resolve_constraints : (env -> fromrewrite -> fromrewrite -> (fromrewrite * fromrewrite) option * env) ref) = ref (fun env _ _ -> None, env)

exception Not_compile_time
exception Parameter_number_mismatch of string
exception Interpreter_call of fromrewrite
exception Interpreter_rew of fromrewrite

let may_failwith s =
  if !during_type_inference then (
    if debug then printf "interpretation failed (%s)\n" s ;
    raise (Interpreter_rew (Constr("Pervasives::Builtin", "Defer", bpos))) 
  ) else failwith s

let rec deref = function
  | Ref(r,name,pos) -> 
      (match deref !r with
      |	Constr("Pervasives::Builtin", "Bottom", _) -> may_failwith (sprintf "bottom found in \"%s\"%s" name (Info.pos2sfull pos))
      |	e -> e)
  | e -> e

let create_closure env expr =
  (* closure only if needed *)
  rew_map_leaf (function
    | depth, (Local((i, stack, n), pos)) when stack >= depth -> 
	Ref(!(get_local_val env i (stack - depth)), n, pos)
    | _, Global((i, pkg, n), pos) ->
	Ref(!(get_global_val env i pkg), n, pos)
    | _, e -> e
  ) expr

let rec remap_locals_propa env nb e = 
  rew_map_leaf (function 
    | depth, Local((i,stack,n),pos) when stack >= depth -> 
	if stack+nb < depth 
	then 
	  let v = ! !(get_local_val env i (stack-depth)) in
	  if is_same v bottom then may_failwith "bottom found" else
	  remap_locals_propa env (depth + nb) v
	else Local((i,stack+nb,n),pos)
    | _, e -> e) e

let rec remap_locals nb e = 
  rew_map_leaf (function 
    | depth, Local((i,stack,n),pos) when stack >= depth -> 
	if stack+nb < depth then may_failwith "bottom found" else Local((i,stack+nb,n),pos)
    | _, e -> e) e


let rec rew_map_ref f e = 
  let e' = f e in
  match e' with
  | Call(e1, e2) -> Call(rew_map_ref f e1, rew_map_ref f e2)
  | Sep(s,l) -> Sep(s, map (rew_map_ref f) l)
  | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_ref f) l)
(*  | Func(nb, para, body) -> Func(nb, map (rew_map_ref f) para, rew_map_ref f body) *)
  | Ref(r,n,pos) -> Ref(ref (rew_map_ref f !r), n, pos)
  | e -> e

let debuginfo f para = 
  if debug then printf "%s called with %s\n" f (rew2s para)

let sep2l_env env e f =
  match e with
  | SepL(",", locals, l) -> 
      let env2 = push_stackt env (length locals) def_flags in
      (try
	let r = f true env2 l in
	pop_stackt_ignore env2 ; r
      with e -> pop_stackt_ignore env2 ; raise e)
  | Sep(",", l) -> f false env l
  | _ -> f false env [e]

let func_env env e f = 
  match e with
  | Func(nblocal, in_para, body) ->
      let env2 = push_stackt env nblocal { def_flags with para_or_local = `Parameter } in
      (try
	let r = f env2 in_para body in
	pop_stackt_ignore env2 ; r
      with e -> pop_stackt_ignore env2 ; raise e)
  | _ -> failwith "func_env"

let sep2l_env_binary env e f =
  sep2l_env env e (fun must_remap env' l ->
    match l with
    | [a;b] -> f must_remap env' (a,b)
    | _ -> failwith "sep2l_env_binary"
  )

let rec unify env (a,b) = 
  let a = deref a in
  (* printf "unifying %s and %s\n" (rew2s a) (rew2s b) ;*)
  match a, b with
  | v, Ref(r,_,_) -> r := v ; true

  | _, Local(_) -> 
      if debug then printf "setting %s to %s\n" (rew2s b) (rew2s a) ;
      !(get_var_val env b) := a ; true
  | _, Global(_) ->
      (* HACK: ensuring globals are eval'ed. 
	 This is hacky since Local's should have the same treatment when non strictly local *)
      let b' = ! ! (get_var_val env b) in
      unify env (a, b')

  | Local(_), _ 
  | Global(_), _ -> may_failwith "unify"

  | _, Call(c, _) when is_allt c ->
      (match partition (function Local _ | Global _ | Ref _ -> true | _ -> false) (allt2l b) with
      |	[], lb ->
	  (* HACK: should be "for_all" instead of "exists" (not short-circuit'ed) *)
	  (* that way for unifying () with *** *)
	  (* this comes from *** being transformed in None|&|""|&|() *)
	  (* whereas a closing the world should be used resulting in *** giving () *)
	  exists id (map (fun b -> unify env (a,b)) lb)
      |	[var], lb ->
	  (* handling X(1)          -> X(x) |&| p  -->  x=1 and p = superval *)
	  (* handling X(1) |&| Y(2) -> X(x) |&| p  -->  x=1 and p = Y(2) *)
	  (match (fold_left_option (fun (used_la, unused_la) b ->
	      match break_at (fun a -> unify env (a,b)) unused_la with
	      | _, [] ->
		  if exists (fun a -> unify env (a,b)) used_la then
		    failwith "uh?" (*Some(used_la, unused_la)*)
		  else
		    None
	      | unused_la1, used :: unused_la2 ->
		  Some(used :: used_la, unused_la1 @ unused_la2)
	    ) ([], allt2l a) lb) with
	  | None -> false
	  | Some(_used_la, unused_la) ->
	      unify env (l2allt unused_la, var))
	  
      |	_ -> failwith (sprintf "unify: more than one variable in %s\n" (rew2s b)))

  | Call(c, _), _ when is_allt c -> exists (fun a -> unify env (a,b)) (allt2l a)

  | _, Call(Constr("Pervasives::Builtin", "Overload", _), _) ->
      unify env (a, l2allt (map fst (unoverload b)))

  | Call(fa,pa), Call(fb,pb) -> for_all (unify env) [fa,fb;pa,pb]

  | Sep(sa,la), Sep(sb,lb) -> sa = sb && length la = length lb && for_all (unify env) (combine la lb)
  | SepL _, SepL _ -> internal_error "unify"

  | _,_ -> is_same a b


let rec resolve_overloading env fname f para =
  let fs = unoverload f in
  if length fs = 1 then 
    (* do not need to evaluate parameters, it means non overloaded function are allowed "Lazy" *)
    (* no need to check if the function is ok to call, was checked at compile-time (at least it should!) *)
    [ fst (hd fs) ], para, env
  else
    let para2 = eval_ env para in
    let para2_ = rew_map_ref (function
      | Call(Constr("Pervasives::Builtin", "Overload", _), _) as e -> 
	  let _vals, typs = split (unoverload e) in
	  if for_all is_func typs
	  then anyval
	  else unoverload_const e
      | Func _ -> anyval
      | e -> e) para2 in

    if debug then printf "overloaded function choosing paras: %s\n\n" (rew2s_ Print.verbose true para2_) ;

    let fs', (env2, latt) = map_withenv (fun (env,latt) (f, tf) ->
      let r, env_latt = match !resolve_constraints env tf para2_ with
      |	Some (in_para, upper_para), env -> 
	  let (in_para', latt2) = Lattice.add_ (raw_compare env) latt in_para in
	  (in_para', (upper_para, f)), (env, latt2)
      |	None, env -> (bottom, (bottom, f)), (env, latt)
      in
      if debug then printf "overloaded function choosing: %s  -->  %s\n\n" (rew2s tf) (rew2s (fst r));
      r, env_latt
    ) (env, Lattice.create()) fs in
    let best_fs = 
      match Lattice.lower_bounds latt with
      | [] -> may_failwith (sprintf "can't call %s with %s%s" (rew2s fname) (rew2s para) (pos2sfull (getpos para)))
      | l -> map (fun e -> 
	  let choices = all_assoc_by (== (*on purpose*)) e fs' in
	  let choices2 =
	    let reals, abstracts = partition (fun (a,_) -> no_abstract a) choices in
	    if reals = [] then abstracts else reals
	  in
	  (* if there are still choices, arbitrary take the first one *)
	  snd (hd choices2)
	) l
    in
    if debug then printf "overloaded function %s, chosen %s\n" (rew2s fname) (String.concat " AND " (map rew2s best_fs));
    best_fs, para2, env2

and reduce_seq env s l = 
  match s with
  | ";" ->
      let l', e = split_last l in
      let _ = map (eval_val env) l' in e
  | _ -> Sep(s, l)

and caml_defined name env para =
  let ops2 deref_a deref_b f = 
    sep2l_env_binary env para (fun must_remap env' (a,b) ->
      let a' = (if deref_a then eval_val else eval_) env' a in
      let b' = (if deref_b then eval_val else eval_) env' b in
      let r = f(a', b') in
      if must_remap then remap_locals_propa env' (-1) r else r
    )
  in
  let ops_choose deref chooser f =
    let para' = (if deref then eval_val else eval_) env para in
    l2allt (map f (filter chooser (allt2l para')))
  in
  let ops2_choose deref_a deref_b chooser_a chooser_b f = 
    ops2 deref_a deref_b (fun (a,b) -> 
      let la = filter chooser_a (allt2l a) in
      let lb = filter chooser_b (allt2l b) in
      l2allt (map f (vector_product2 la lb))
    )
  in
  let ops2_choose_sym deref chooser f = ops2_choose deref deref chooser chooser f in

  let infix_num_op f = 
    ops2_choose_sym true is_float (function
      | Float(a,posa), Float(b,posb) -> Float(f a b, unipos posa posb)
      | a, b -> may_failwith (sprintf "infix_num_op: bad values %s and %s" (rew2s a) (rew2s b))
    )
  in
  let prefix_num_op f = 
    ops_choose true is_float (function
      | Float(a,pos) -> Float(f a, pos)
      | e -> failwith ("prefix_num_op: bad value " ^ rew2s e)
    )
  in
  let infix_string_op f = 
    ops2_choose_sym true is_string (function
    | String(a,posa), String(b,posb) -> String(f a b, unipos posa posb)
    | a,b -> failwith (sprintf "infix_string_op: bad values (%s,%s)" (rew2s a) (rew2s b)))
  in
  match name with
  | "=" | ":=" ->
      ops2 false true (fun (a,b) ->
	(match b with
	| Ref(r,_,_) -> get_var_val env a := r
	| _ -> if not (unify env (b,a)) then failwith "= failed");
	Constr("Pervasives::Builtin", "Empty_tuple", unipos (Rew.getpos a) (Rew.getpos b)))

  | "unify" ->
      ops2 true true (fun (a,b) -> bool2rew (unify env (a,b)))

  | "sqrt" -> prefix_num_op sqrt
  | "+" -> infix_num_op ( +. )
  | "*" -> infix_num_op ( *. )
  | "**" -> infix_num_op ( ** )
  | "-" -> 
      if sep2nb para = 2 then 
	infix_num_op (-.) 
      else 
	prefix_num_op ((-.) 0.)

  | "concat" -> infix_string_op ( ^ )

  | "string_length" ->
      ops_choose true is_string (function
	| String(s,pos) -> Float(float_of_int (String.length s), pos)
	| e -> failwith ("string_length: bad value " ^ rew2s e)
      )

  | "string_get" ->
      ops2_choose true true is_string is_float (function
	| String(s,_), Float(n,_) when is_int n -> 
	    (try String(String.sub s (int_of_float n) 1, bpos)
	    with _ -> raise (Interpreter_rew (String("string_get", bpos))))
	| s, n -> failwith (sprintf "to_string: bad values %s and %s" (rew2s s) (rew2s n))
      )

  | "random" -> prefix_num_op (fun n -> float_of_int (Random.int (int_of_float n)))

  | "to_string" -> 
      let l = 
	map (function
	  | String(s, _) -> s
	  | e -> rew2s e
        ) (allt2l (eval_val env para))
      in l2allt (map (fun s -> String(s, Rew.getpos para)) l)

  | "==" -> ops2 true true (fun (a,b) -> bool2rew (is_same a b))
  | "!=" -> ops2 true true (fun (a,b) -> bool2rew (not (is_same a b)))
  | "compare_strings" -> ops2_choose_sym true is_string (function
      | String(a,posa), String(b,posb) -> Float(float_of_int (compare a b), unipos posa posb)
      | _ -> failwith "compare_string")
  | "compare_numbers" -> ops2_choose_sym true is_float (function
      | Float(a,posa), Float(b,posb) -> Float(float_of_int (compare a b), unipos posa posb)
      | _ -> failwith "compare_string")

  | "print" -> 
      ops_choose true is_string (function
	| String(s, pos) -> print_string s ; flush stdout ; Constr("Pervasives::Builtin", "Empty_tuple", pos)
	| _ -> failwith "print"
      )      

  | "unlazy" -> 
      let e = eval_ env para in
      let e2 = eval_ env (deref e) in
      e2

  | "args" -> l2sep (map (fun s -> String(s, bpos)) !args)

  | "raise" -> raise (Interpreter_rew (eval_val env para))

  | "cond" ->
      debuginfo "cond" para;
      (match para with
      |	Sep(",", l) ->
	  (try
	    find_some (fun e ->
	      sep2l_env_binary env e (fun must_remap env' (a,b) ->
		let va = eval_val env' a in
		if rew2bool va
		then (if debug then printf "cond : chosen %s\n" (rew2s b) ; 
		      Some(let b' = eval_ env' b in
		           if must_remap then remap_locals_propa env' (-1) b' else b'))
		else (if debug then printf "cond : Trying next\n" ; None)
              )
	    ) l
	  with Not_found -> may_failwith (sprintf "default cond condition%s" (pos2sfull (getpos para))))
      |	_ -> failwith "bad cond args")

    (* nearly identity *)
  | "to_list" -> l2tuple (sep2non_empty_l (eval_val env para))

  | "list_cons" -> ops2 false false (fun (a,b) -> l2tuple(a :: tuple2l (deref b)))

  | "list_value" ->
      (match tuple2l (eval_val env para) with
      |	[] -> Constr("Pervasives::Builtin", "None", bpos)
      |	e :: _ -> Call(Constr("Pervasives::Builtin", "Some", bpos), e))

  | "list_next" ->
      (match tuple2l (eval_val env para) with
      |	[] -> raise (Interpreter_rew (String("list_next: empty list", bpos)))
      |	_ :: l -> l2tuple l)

  | "list_deconstruct" ->
      (match eval_val env para with
      |	Constr("Pervasives::Builtin", "Empty_tuple", pos) -> Constr("Pervasives::Builtin", "None", pos)
      |	Sep(",", e :: l) -> Call(Constr("Pervasives::Builtin", "Some", bpos), Sep(",", [ e ; l2tuple l ]))
      |	_ -> may_failwith "bad args for list_deconstruct")

  | "tuple_add" ->
      let is_nonempty = function 
	| Constr("Pervasives::Builtin", "Empty_tuple", _) -> false
	| _ -> true
      in
      (match eval_val env para with
      |	Sep(",", l) -> l2sep (collect sep2l (filter is_nonempty l))
      |	_ -> may_failwith "bad args for tuple_add")

  | "l_unify" ->
      debuginfo "l_unify" para;
      (match eval_val env para with
      |	SepL(",", [], l) -> remap_locals (-1) (l2or l)
      |	Sep(",", l) -> l2or l
      |	a -> if is_simple_const2 a then a else may_failwith "bad args for l_unify")

  | _ -> failwith ("unknown caml op " ^ name)

and call_func envpara para env f = 
  let in_func env in_para body =
    let nb_in_para_min, strictness = compute_nbpara_bounds in_para in

    let unif must_remap (para, envpara) =
      let local_eval e =
	let e' = eval_ envpara e in
	(* the result must be remapped (+1), so if must_remap keep it unchanged *)
	if must_remap then e' else remap_locals 1 e'
      in
      for_all2 (fun a b ->
	match a with
	| Call(Constr(_, "Out", _), var)
	| Call(Constr(_, "Inout", _), var) -> 
	    let b' = local_eval b in
	    (match b' with
	    | Call(Constr("Pervasives::Builtin", "Ref", _), Ref(r,_,_))
	    | Ref(r,_,_) -> get_var_val env var := r ; true
	    | _ -> if !during_type_inference then raise Not_compile_time else failwith "beurk")	    
	| Call(Constr(_, "Lazy", _), var) ->
	    (*if must_remap && exists (fun v -> var_indice v > 0) (rew_collect_all_accessible_vars b) then
	      failwith "variable in Lazy evaluation" ;*)
	    let bclosed = match b with
	    | Call(External("unlazy", _), b') -> 
		(* already closed *)
		if is_var b' then ! !(get_var_val envpara b') else deref b'
	    | _ -> create_closure envpara b in
	    !(get_var_val env var) := bclosed ; true
	| Call(Constr(_, "Dynamic", _), a') -> 
	    unify env (local_eval b, a')
	| a' -> 
	    unify env (local_eval b, a')
      ) in_para para
    in
    if !during_type_inference && exists is_dynamic in_para then
      (* keep it un-eval'ed *)
      raise Not_compile_time
    else if (
      let para' = 
	if sep2nb para = 1 && 1 < nb_in_para_min && is_var para then ! !(get_var_val envpara para) else para in
      let nb_para = sep2nb para' in
      if nb_para < nb_in_para_min then raise (Parameter_number_mismatch (sprintf "missing parameters (%d < %d)%s" nb_para nb_in_para_min (pos2sfull (getpos para))))
      else if strictness = `Strict then
	if nb_in_para_min = 1 then
	  unif false ([ para' ], envpara)
	else if nb_para > nb_in_para_min then
	  raise (Parameter_number_mismatch (sprintf "too many parameters (%d > %d)%s" nb_para nb_in_para_min (pos2sfull (getpos para))))
	else
	  sep2l_env envpara para' (fun must_remap envpara' l -> unif must_remap (l, envpara'))
	else (* `NoLimit *)
	  sep2l_env envpara para' (fun must_remap envpara' l ->
	    let a,b = break_at_indice nb_in_para_min l in
	    let l' = a @ [ Call(Constr("Pervasives::Builtin", "Remaining", bpos), l2sep b) ] in
	    unif must_remap (l', envpara'))
    ) then 
      let r = eval_including_func env body in
      (* TODO: only deref when "ref" is not use *)
      let r' = full_deref_no_func r in
      remap_locals_propa env (-1) r'
    else
      if !during_type_inference then raise Not_compile_time else failwith "can't call function with those parameters"
  in
  func_env env f in_func

and (call : env -> fromrewrite -> fromrewrite -> fromrewrite -> fromrewrite) = fun env fname f para ->
  let f = deref f in
  match f with

  | Constr("Pervasives::Builtin", "Freeze", _) ->
      let e = rew_remove_unused_sepl para in
      let e', _ = rew_map_once_adapting_env (fun env -> function
	| Call(Constr("Pervasives::Builtin", "Expand", _), e) -> Some (eval_ env e, env)
	| _ -> None
      ) env e in
      e'

  | Constr _ -> Call(f, eval_ env para)
  | Local _ -> 
      if !during_type_inference then Call(f, eval_ env para) else
      failwith (sprintf "no way to have this at runtime (variable %s is bottom)" (rew2s_ true false f))

  | Global((i,pkg,_),_) ->
      if !during_type_inference then Call(f, eval_ env para) else
      if pkg = "Pervasives::Builtin" && is_constraint_op i then unit (* ignore it *) else
      failwith (sprintf "no way to have this at runtime (variable %s is bottom)" (rew2s_ true false f))

  | External(name,_) -> caml_defined name env para

  | Func _ -> call_func env para env f

  | Call(Constr("Pervasives::Builtin", "Val", _), _) -> failwith "Pervasives::Builtin::Val in interpreter is bad!"

  | Call(Constr("Pervasives::Builtin", "Overload", _), _) ->
      let fs, para', env2 = resolve_overloading env fname f para in
      l2allt (map (fun f' -> call env2 fname (eval_ env2 f') para') fs)

  | Call(Constr _, _) -> Call(f, eval_ env para)

  | _ -> Call(f, para)

and eval_including_func env expr = 
  match expr with
  | Func(nb, _, _) ->
      func_env env expr (fun env in_para body ->
	Func(nb, map (eval_ env) in_para, eval_ env body)
      )
  | _ -> eval_ env expr

and eval_ env expr =
  if debug then printf "eval_ (beg): %s\n" (rew2s expr);
  let expr' = match expr with

  | Call(Constr("Pervasives::Builtin", "Overload", _), _) -> expr

    (* bounded vals *)
  | Call(Constr("Pervasives::Builtin", "Var_forall", _), _)
  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) ->
      if !during_type_inference then expr else failwith "no way to have this at runtime"

  | Call(f, para) -> call env f (eval_ env f) para

  | Local((_,_,n),pos)
  | Global((_,_,n),pos) -> 
      let r = !(get_var_val env expr) in
      if is_same !r bottom then expr else
      (try
	if debug then printf "eval_ (end): local var %s is (derefed) %s\n" (rew2s expr) (rew2s_ false true !r) ;
	let l = unoverload ! r in
	if for_all (fun (v,t) -> is_same v t) l then
	  l2allt (map fst l) else Ref(r, n, pos)
      with _ -> Ref(r, n, pos))

  | Sep(s, l) -> 
      let l' = map (eval_ env) l in
      reduce_seq env s l'
      
  | SepL(s, locals, l) -> 
      let env' = push_stackt env (length locals) def_flags in
      (try
	let l' = map (eval_ env') l in
	let expr' = remap_locals_propa env' (-1) (reduce_seq env' s l') in
	pop_stackt_ignore env' ;
	expr'
	 
      with e -> pop_stackt_ignore env' ; raise e)

  | Func _ -> create_closure env expr
  | Ref _
  | External _
  | Float _
  | String _
  | Constr _ -> expr
  in
  if debug then printf "eval_ (end): %s --> %s\n" (rew2s expr) (rew2s expr');
  expr'

and eval_val env e =
  let e2 = eval_ env e in
  let e3 = full_deref_no_func e2 in
  let e4 = rew_map_ref (function
    | Call(Constr("Pervasives::Builtin", "Overload", _), _) as e -> unoverload_const e
    | e -> e
  ) e3 in
  let e5 = Type_handling.normalize e4 in
  e5

let eval_during_type_inference env e =
  let noenv_check = !Env.noenv_check in
  Env.noenv_check := true ;
  let e2 = eval_val env e in
  let e3 = full_deref e2 in
  Env.noenv_check := noenv_check ;
  e3

let eval_main env e =
  noenv_check := true ;
  during_type_inference := false ;
  eval_val env e
