{
open Types
open Stack
open Common
open Lexing

type error =
  | Unterminated_string
  | Unterminated_pstring
  | Unterminated_rstring
  | Unterminated_bstring
  | Bad_hex_in_string
exception Error of error

let lines_starts = Stack.create()
let current_file = ref ""

let rel_pos = ref 0
let l_start lexbuf = !rel_pos + lexeme_start lexbuf
let l_end lexbuf = !rel_pos + lexeme_end lexbuf
let pos lexbuf = !current_file, l_start lexbuf, l_end lexbuf

let pstring_ = ref 0
let bstring_ = ref 0
let got_string = ref ""
let beg_string = ref 0
let next_s s t lexbuf =
  got_string := !got_string ^ s ;
  t lexbuf
let next t lexbuf = next_s (lexeme lexbuf) t lexbuf

let lexer_pos2sfull lexbuf = 
  let line, line_offset = length lines_starts, try Stack.top lines_starts with Stack.Empty -> 0 in
  let n1 = lexeme_start lexbuf - line_offset in
  let n2 = lexeme_end lexbuf - line_offset in
  Printf.sprintf ", file %s, line %d, characters %d-%d" !current_file (line + 1) n1 n2

let lexeme_check_identifier_style lexbuf =
  let s = lexeme lexbuf in
  (if Flags.check_style then
     match to_CamelCase s with
     | Some s' -> Printf.eprintf "Bad style identifier \"%s\" better written \"%s\"%s\n" s s' (lexer_pos2sfull lexbuf)
     | None -> () 
  );
  s

let ins t lexbuf = 
  got_string := ""; beg_string := l_start lexbuf;
  let old_pstring = !pstring_ in
  pstring_ := 1;
  let s = t lexbuf in
  pstring_ := old_pstring ;
  s

let insb t lexbuf = 
  got_string := ""; beg_string := l_start lexbuf;
  let old_bstring = !bstring_ in
  bstring_ := 1;
  let s = t lexbuf in
  bstring_ := old_bstring ;
  s


let rec lexbuf2list t beg lexbuf =
  let old_rel_pos = !rel_pos in rel_pos := !rel_pos + beg ;
  let rec f () =
    match t lexbuf with
    | EOF -> 
	rel_pos := lexeme_end lexbuf;
	[]
    | e -> e :: f()
  in
  let l = f() in
  got_string := ""; beg_string := l_start lexbuf;
  rel_pos := old_rel_pos;
  l

let s_list = function
  | [e] -> e
  | l -> LIST([], l)
let s_concat (a,b) = s_list [ BIGIDENT("String", bpos) ; OP("::", bpos) ; IDENT("concat", bpos) ; s_list (a @ [ OP(",", bpos) ] @ b) ]

let main_lexbuf2list t lexbuf file = 
  current_file := file ;
  Stack.clear lines_starts ;
  rel_pos := 0 ;
  let l = 
    try lexbuf2list t 0 lexbuf 
    with 
    | Error(Unterminated_string)
    | Error(Unterminated_pstring)
    | Error(Unterminated_rstring)
    | Error(Unterminated_bstring) -> failwith (Printf.sprintf "Unterminated string%s" (lexer_pos2sfull lexbuf))
    | Error(Bad_hex_in_string) -> failwith (Printf.sprintf "Bad_hex_in_string%s" (lexer_pos2sfull lexbuf))
  in
  l, 0 :: stack2list lines_starts
} 

let ident = ['0'-'9' 'A'-'Z' 'a'-'z' '_']*
let ident_end = ['?' '!' ''']*
let disturbing_ident_end = ['?' '!']
let free_chars = [ '&' '~' '-' '|' 
                   '`' '\\' '^' '@' '='
                   '' '+' '$' '' '%' '' '*' '!' ''
                   '/' ';' '?' '<' '>' '' '''  ]

rule token = parse
| [ '(' '[' '{' ')' ']' '}' ] { OP(lexeme lexbuf, pos lexbuf) }
| ','         { OP(lexeme lexbuf, pos lexbuf) }
| '.'         { OP(lexeme lexbuf, pos lexbuf) }
| ".."        { OP(lexeme lexbuf, pos lexbuf) }
| ":"         { OP(lexeme lexbuf, pos lexbuf) }
| "::"        { OP(lexeme lexbuf, pos lexbuf) }
| ":"? free_chars (free_chars | ':')* { OP(lexeme lexbuf, pos lexbuf) }
| [' ' '\t']+ { SPACE(pos lexbuf) }
| [' ' '\t']* '\n' { space lexbuf }
| [' ' '\t']* '#' [^ '\n']* { token lexbuf }
| ['a'-'z' '_'] ident ident_end { IDENT(lexeme_check_identifier_style lexbuf, pos lexbuf) }
| ['A'-'Z']     ident ident_end { BIGIDENT(lexeme_check_identifier_style lexbuf, pos lexbuf) }
| ['a'-'z' '_'] ident disturbing_ident_end free_chars { failwith (Printf.sprintf "Confusion between indentifier ending with \"?\" or \"!\"%s" (lexer_pos2sfull lexbuf)) }
| ['0'-'9']* '.' ['0'-'9']+ (['e' 'E']['-' '+']?['0'-'9']+)? { NUM(float_of_string(lexeme lexbuf), pos lexbuf) }
| ['0'-'9']+ { NUM(float_of_string(lexeme lexbuf), pos lexbuf) }
| "0x" ['0'-'9' 'a'-'f' 'A'-'F']+ { NUM(float_of_string(lexeme lexbuf), pos lexbuf) }
| "q("  { let beg = l_start lexbuf in let s = ins rstring lexbuf in STRING(s, (!current_file, beg, l_end lexbuf)) }
| '"'   { ins  string lexbuf }
| "qq(" { ins pstring lexbuf }
| eof   { EOF }
| _ { let (s,a,b) = pos lexbuf in
	failwith (Printf.sprintf "error tokenizing ``%s'' at (%s,%d,%d)\n" (lexeme lexbuf) s a b) }

and space = parse
| [' ' '\t']* '\n' {
    push (l_start lexbuf) lines_starts;
    space lexbuf
  } 
| [' ' '\t']* { 
    push (l_start lexbuf) lines_starts;
    INDENT(lexeme lexbuf)
  }

and string = parse
  '"' { STRING(!got_string, (!current_file, !beg_string, l_end lexbuf)) }
| '\\' { string_escape lexbuf }
| '{' {
     let sbeg = !beg_string in
     let beg = l_start lexbuf in
     let before = !got_string in
     let code = insb bstring lexbuf in

     let l = lexbuf2list token (beg + 1) (from_string code) in
     let after = l_start lexbuf in

     let e = string lexbuf in
     s_concat([ STRING(before, (!current_file,sbeg,beg)) ],
	      [ s_concat([ IDENT("to_string",(!current_file,beg,after)) ; s_list l ], [e]) ])
  }
| [^ '\n' '\\' '{' '"']+ { next string lexbuf }
| '\n' { 
    push (l_end lexbuf) lines_starts;
    next string lexbuf
  }
| eof { raise (Error Unterminated_string) }

and pstring = parse
  ')' { 
    decr pstring_; 
    if !pstring_ = 0 then STRING(!got_string, (!current_file,!beg_string, l_end lexbuf)) else next pstring lexbuf
  }
| '(' {
    incr pstring_;
    next pstring lexbuf
  }
| '\\' { pstring_escape lexbuf }
| '{' {
     let sbeg = !beg_string in
     let beg = l_start lexbuf in
     let before = !got_string in
     let code = insb bstring lexbuf in

     let l = lexbuf2list token (beg + 1) (from_string code) in

     let e = pstring lexbuf in

     s_concat([ STRING(before, (!current_file,sbeg,beg)) ],
	      [ s_concat([ IDENT("to_string",bpos) ; s_list l ], [e]) ])
  }
| [^ '\n' '(' ')' '\\' '{']+ { next pstring lexbuf }
| '\n' { 
    push (l_end lexbuf) lines_starts;
    next pstring lexbuf
  }
| eof { raise (Error Unterminated_pstring) }

and pstring_escape = parse
| '0' { next_s "\000" string lexbuf }
| 'n' { next_s "\n" pstring lexbuf }
| 't' { next_s "\t" pstring lexbuf }
| '{' { next_s "{" pstring lexbuf }
| 'x' _ _ { 
  try
    let s = String.make 1 (Char.chr (int_of_string ("0" ^ lexeme lexbuf))) in
    next_s s string lexbuf 
  with Failure("int_of_string") -> raise (Error Bad_hex_in_string)
  }
| _ { next_s ("\\" ^ lexeme lexbuf) pstring lexbuf }

and string_escape = parse
| '0' { next_s "\000" string lexbuf }
| '"' { next_s "\"" string lexbuf }
| 'n' { next_s "\n" string lexbuf }
| 't' { next_s "\t" string lexbuf }
| '{' { next_s "{" string lexbuf }
| 'x' _ _ { 
  try
    let s = String.make 1 (Char.chr (int_of_string ("0" ^ lexeme lexbuf))) in
    next_s s string lexbuf 
  with Failure("int_of_string") -> raise (Error Bad_hex_in_string)
  }
| _ { next_s ("\\" ^ lexeme lexbuf) string lexbuf }

and rstring = parse
| ')' { 
    decr pstring_; 
    if !pstring_ = 0 then !got_string else next rstring lexbuf
  }
| '(' {
    incr pstring_;
    next rstring lexbuf
  }
| [^ '\n' '(' ')']+ { next rstring lexbuf }
| '\n' { 
    push (l_end lexbuf) lines_starts;
    next rstring lexbuf
  }
| eof { raise (Error Unterminated_rstring) }

and bstring = parse
  '}' { 
    decr bstring_; 
    if !bstring_ = 0 then !got_string else next bstring lexbuf
  }
| '{' {
    incr bstring_;
    next bstring lexbuf
  }
| [^ '\n' '{' '}']+ { next bstring lexbuf }
| '\n' { 
    push (l_end lexbuf) lines_starts;
    next bstring lexbuf
  }
| eof { raise (Error Unterminated_bstring) }


