open Common
open Print
open Printf
open List
open Types
open Env
open Info
open Type_handling
open Type_inference

let debug = Flags.debug_infere_and_interpret


let dumpt_forced_concise env = 
  let type_vars = Type_inference.get_type_vars env in
  let s = ref "--------------------------------------------------------------------------------\n" in
  env_iter_constr (fun var constr ->
    eprintf "Computing short name for %s\n" (rew2s var) ; flush stderr ;
    if constr <> TFree then s := !s ^ sprintf "%20s !! %s\n" (rew2s var) (constr2s (constr_map (fun e -> fst (rew2names_ type_vars env e)) constr))
  ) env ;
  !s

let find_package name = 
  let inc = (try [ Sys.getenv "MERD_LIB" ] with Not_found -> []) @ [ "/usr/share/merd" ; "." ; "../lib" ] in
  let findfile f = find Sys.file_exists (map (fun dir -> dir ^ "/" ^ f ^ ".me") inc) in

  let file = String.concat "/" (split_at_two_colons name) in

  try findfile file
  with Not_found ->
    let file' = String.copy file in
    file'.[0] <- Char.lowercase file'.[0] ;
    if file' <> file then findfile file' else raise Not_found

let save_mec file (package : package) =
  let out = open_out_bin (file ^ "c") in
  Marshal.to_channel out package [];
  close_out out

let use_mec file =
  let mtime f = (Unix.stat f).Unix.st_mtime in
  not (has_env "MERD_NO_MEC") && Sys.file_exists(file ^ "c") && mtime (file ^ "c") >= mtime file

let read_mec file =
  if not (use_mec file) then None else
  let f = open_in_bin (file ^ "c") in
  let package = Marshal.from_channel f in
  close_in f;
  Some { package with from_mec = true }

let rec load_file packages file name =
  let package, packages =
    match read_mec file with
    | Some package -> 
	let l, packages = map_withenv (load_package None) packages package.dependencies in
	(if for_all (fun (_,p) -> p.from_mec) l then Some package else None), packages
    | None ->
	None, packages
  in
  let package, packages =
  match package with
  | Some package ->
      package, packages
  | None ->
      let lexbuf = Lexing.from_channel (open_in file) in
      let package, packages = infere_package packages file name lexbuf in
      (try save_mec file package with _ -> ()) ;
      package, packages
  in
  package, packages @ [ name, package ]

and load_package hierarchical_name packages name =
  let load_package_ packages name short_name =
    try
      (short_name, assoc name packages), packages
    with Not_found ->
      let file = find_package name in (* exception catched below *)
      let package, packages = load_file packages file name in
      (short_name, package), packages
  in
  try
    match hierarchical_name with
    | Some s ->
	(try load_package_ packages (s ^ "::" ^ name) name
	with Not_found -> load_package_ packages name name)
    | None -> load_package_ packages name name
  with Not_found -> failwith (sprintf "can't find package %s" name)

and infere_package packages file name lexbuf =
  let (tokens, line_starts) = Lexer.main_lexbuf2list Lexer.token lexbuf file in
  Info.add file line_starts ;

  let new_ops, s_imports, s_packages_used, tokens = Parser.parse_all1 tokens in

  if s_packages_used <> [] && not Flags.quiet then eprintf "%s needs %s\n" name (String.concat ", " s_packages_used) ;

  let hierarchical_name = get_package_name name in

  let packages_used, packages = map_withenv (load_package hierarchical_name) packages s_packages_used in
  let packages_used = 
    (* Pervasives::Builtin must always be accessible *)
    try packages_used @ [ "Pervasives::Builtin", assoc "Pervasives::Builtin" packages ]
    with Not_found -> packages_used
  in
  let imports = map (fun pkg -> assoc_or_fail pkg packages_used) s_imports in

  (* !! packages_used is not indexed the same as packages !! *)
  (* in case of relative packages like hierarchical_name being A, *)
  (* s_packages_used can be C and loads A::C *)
  (* => variable "packages" is A::C => package *)
  (* => variable "packages_used" is C => package *)

  if not Flags.quiet then eprintf "Starting package %s\n" name ;


  let ops = 
    let imported_ops = map (fun pkg -> pkg.operators) imports in
    fold_left Parser.add_ops Parser.ops_empty (new_ops :: imported_ops) in
  let c = Parser.parse_all2 ops tokens in  
  let grouped = Grouper.grouper ops c in
  let rewritten, globals = Rewrite.rewrite name grouped imports packages_used packages in

  let l_packages_used = map snd packages_used in

  let env = new_env packages globals name in

  let valtyp, env =
    try infere_main env rewritten
    with Type_inference.Type_greater(a,b) -> 
      let s = (Printf.sprintf "can't ensure %s > %s%s%s" (rew2s a) (rew2s b) (pos2sfull (Rew.getpos a)) (pos2sfull (Rew.getpos b))) in
      print_endline s ; flush stdout ; failwith s
  in
  let bytecode, bytecode_type = to_run_time valtyp in
  let package = {
   full_name = name ;
   from_mec = false ;
   bytecode = bytecode ; bytecode_type = bytecode_type ;
   overloading_funcs = env.overloading_funcs4pkg ;
   p_global = assoc_or_fail name env.global ;
   line_starts = line_starts ;
   operators = ops ;
   dependencies = map (fun pkg -> pkg.full_name) l_packages_used ;
   re_exported_packages = uniq (name :: collect (fun pkg -> pkg.full_name :: pkg.re_exported_packages) imports) ;
   p_virtual_classes = env.virtual_classes.class4pkg ;
  } in
  if !debug then dump_package package ;
  package, packages


let chop_extension file =
  try Filename.chop_extension file with Invalid_argument _ -> file

let file2package_name file =
  let name = String.copy (chop_extension (Filename.basename file)) in
  name.[0] <- Char.uppercase name.[0] ;
  name  

let doit file_or_name lexbuf_optional =
  let packages = 
    let empty_package name = name, { full_name = name ; from_mec = false ; bytecode = Rew.bottom ; bytecode_type = Rew.bottom ; dependencies = [] ; overloading_funcs = [] ; p_global = [| |] ; operators = Parser.ops_empty ; line_starts = [] ; re_exported_packages = [] ; p_virtual_classes = { p_instances = [] ; p_classes = [] } } in
    [ empty_package "Pragma" ] in
  let packages =
    match lexbuf_optional, file_or_name with
    | Some lexbuf, name ->
	let package, packages = infere_package packages name name lexbuf in
	packages @ [ name, package ]
    | None, file ->
	let package, packages = load_file packages file (file2package_name file) in
	if package.from_mec && !debug then dump_package package ; (* always dump the last package *)
	packages
  in
  packages
 
