(* Type_handling *)
open Types
open Array
open List
open Common
open Printf
open Print
open Rew
open Rewrite
open Env


let debug = Flags.debug_type_handling

let part_of_eta_expand e = Func(0, [anyval], rew_remap_locals 1 e)

let eta_expand_again para e =
  match rew_collect_local_vars e with
  | [] -> Func(0, [anyval], e)
  | _ -> Func(1, [para], e)

let unpack_one_local_constr body =
  let bo, body', constrs = unpack_constr body in
  bo, body', (match constrs with
  | [] -> def_tconstr
  | [ 0, constr ] -> constr
  | _ -> failwith "unpack_one_local_constr")

let normalize_funcs_handling_constraints l =

  (* form1: C1 |&| Cn -> CC1 |&| CCn  (including _ -> CC1 |&| CCn) *)
  (* form2: x -> (x !< C1 |&| Cn) ; CC1 |&| CCn |&| (D1 |&| Dn)(x) *)
  (* form3: D1  (tis form2 simplified when no constraint) *)

  (* form1 |&| form1 examples *)	  
  (*  (_ -> CC1) |&| (C1 -> CC1)  -->  C1 -> CC1 *)
  (*  (_ -> CC1) |&| (C1 -> CC1 |&| CC2)  -->  C1 -> CC1 |&| CC2 *)
  (*  (_ -> CC1 |&| CC2) |&| (C1 -> CC1 |&| CC3)  -->  (_ -> CC2) |&| (C1 -> CC1 |&| CC3) *)
  (*  (C1 -> CC1) |&| (C2 -> CC1)  -->  (C1 |&| C2 -> CC1) *)

  (* form2 |&| form2 examples *)
  (* (x -> (x !< C1) ; D1(x)) |&| (x -> (x !< C2) ; D1(x))  -->  x -> (x !< C1 |&| C2) ; D1(x) *)
  (* (x -> (x !< C1) ; D1(x)) |&| (x -> (x !< C1) ; D2(x))  -->  x -> (x !< C1) ; (D1 |&| D2)(x) *)
  (* (x -> (x !< C1) ; CC1 |&| D1(x)) |&| (x -> (x !< C2) ; CC1 |&| D2(x))  -->  (C1 |&| C2 -> CC1) |&| (x -> (x !< C1) ; D1(x)) |&| (x -> (x !< C2) ; D2(x)) *)

  (* form1 and form3 can't provoque any simplifications *)
  let has_forms_allowing_simplifications = ref `No in
  let r_para = ref None in

  let extract_forms e =
    collect (function
      |	Call(e, Local((0, 0, _), _)) -> 
	  has_forms_allowing_simplifications := `Use_form3 ;
	  map (fun e -> `Form2, e) (allt2l e)
      |	e -> 
	  has_forms_allowing_simplifications := 
	    if !has_forms_allowing_simplifications = `No then `Maybe else `Only_form1 ;
	  [ `Form1, e ]
    ) (allt2l e) in

  let raw_cc2c, other = fpartition (function
    | Func(0, [c_], cc) -> Some (extract_forms cc, { lower = subval ; greater = if is_anyval c_ then superval else c_ })
    | Func(1, [ Local((0,0,_), _) as para ], body) ->
	r_para := Some para ;
	let bo, cc, c = unpack_one_local_constr body in
	if bo then Some (extract_forms cc, c) else None
    | _ -> None
  ) l in

  let raw_cc2c', other' =
    if !has_forms_allowing_simplifications = `Use_form3 then
      map (fun e -> [ `Form2, rew_remap_locals 1 e ], def_tconstr) other @ raw_cc2c, []
    else
      raw_cc2c, other
  in

  if mem !has_forms_allowing_simplifications [ `No ; `Maybe ] then l (* don't even bother *) else

  let cc2c = collect (fun (l_cc, c) -> map (fun cc -> cc, c) l_cc) raw_cc2c' in

  let constrs_is_same a b =
    is_same a.lower b.lower && is_same a.greater b.greater in
  let constrs_flatten_uniq l =
    { lower   = l2or   (uniq_ is_same (collect (fun c -> or2l   c.lower)   l)) ;
      greater = l2allt (uniq_ is_same (collect (fun c -> allt2l c.greater) l)) } in
  let form_is_same a b =
    match a, b with
    | (`Form1, a), (`Form1, b)
    | (`Form2, a), (`Form2, b) -> is_same a b
    | _ -> false
  in

  let c2cc = map (fun (cc,cs) -> constrs_flatten_uniq cs, cc) (prepare_want_all_assoc_by_ form_is_same constrs_is_same cc2c) in
  let c2ccs = prepare_want_all_assoc_by_ constrs_is_same form_is_same c2cc in

  let l' = collect (fun (c,ccs) ->
    let form1, form2 = map_t2 (map snd) (partition (fun (e,_) -> e = `Form1) ccs) in
    if form2 = [] || is_tfree(TConstr c) then
      if not (is_subval c.lower) then internal_error "normalize_funcs_handling_constraints" else
      (* can use form3 for form2's: x -> D1(x)  -->  D1 *)
      let form1' = if form1 = [] then [] else 
      [ Func(0, [superval2anyval c.greater], l2allt form1) ] in
      let form2' = map (rew_remap_locals (-1)) form2 in
      form1' @ form2'
    else
      let para = some !r_para in
      let form2' = if form2 = [] then [] else [ Call(l2allt form2, para) ] in
      [ Func(1, [para], add_packed_constr (pack_constr para (TConstr c)) (l2allt (rew_sort (form2' @ form1)))) ]
  ) c2ccs 
  in
  l' @ other'



let rec normalize e =
  let rec normalize_ = function 
    (* try to factorize *)
    | Call(funct, Sep _) as e when is_orallt funct ->
	let funct_is_or = is_or funct in

	let factorize (a,b) = 
	  match a, b with

	    (* special range cases *)

	    (* 1|(2..4)  --> 1..4 *)
	  | Float(nb, _) as e_nb, Call(Constr("Pervasives::Builtin", "Or_range", _) as c, Sep(",", [ Float(min, _) as e_min ; Float(max, _) as e_max ])) when funct_is_or && min -. 1. <= nb && nb <= max +. 1. && is_int nb ->
	      Some(Call(c, Sep(",", [ if nb < min then e_nb else e_min ;
				      if nb > max then e_nb else e_max ])))

	    (* 1|2  --> 1..2 *)
	  | Float(nb, pos) as e_nb, (Float(nb2, _) as e_nb2) when funct_is_or && is_int nb && nb +. 1. = nb2 -> 
	      Some(Call(Constr("Pervasives::Builtin", "Or_range", pos), Sep(",", [ e_nb ; e_nb2 ])))
	  | Float(nb, pos) as e_nb, (Float(nb2, _) as e_nb2) when funct_is_or && is_int nb && nb = nb2 +. 1. -> 
	      Some(Call(Constr("Pervasives::Builtin", "Or_range", pos), Sep(",", [ e_nb2 ; e_nb ])))

	    (* range merging  1..3 | 4..6  -->  1..6 *)
	  | Call(Constr("Pervasives::Builtin", "Or_range", _) as c, Sep(",", [ Float(min_a, _) as e_min_a ; Float(max_a, _) as e_max_a ])), 
	    Call(Constr("Pervasives::Builtin", "Or_range", _),      Sep(",", [ Float(min_b, _) as e_min_b ; Float(max_b, _) as e_max_b ])) ->
	      if min_a > max_b +. 1. || min_b > max_a +. 1. 
	      then None (* no merging possible *)
	      else Some(Call(c, Sep(",", [ if min_a < max_b then e_min_a else e_min_b ;
					   if max_a > max_b then e_max_a else e_max_b ])))

	    (* have (o !< C1 ; o) | (o' !< C2 ; o')  -->  (o !< C1 ; o !< C2 ; o) *)
	  | (Call(Constr("Pervasives::Builtin", ("Var_forall" as kind), _), _) as a), (Call(Constr("Pervasives::Builtin", "Var_forall", _), _) as b)
	  | (Call(Constr("Pervasives::Builtin", ("Var_exists" as kind), _), _) as a), (Call(Constr("Pervasives::Builtin", "Var_exists", _), _) as b) ->
	      (match unpack_local_constr_raw kind a, unpack_local_constr_ kind b with
	      |	Some(name, constr_a), Some constr_b ->
		  let lo = l2orallt funct [ constr_a.lower   ; constr_b.lower   ] in
		  let gr = l2orallt funct [ constr_a.greater ; constr_b.greater ] in
		  Some(pack_local_constr_with_kind kind name (TConstr { lower = lo ; greater = gr }))
	      |	_ -> None)

	    (* (A |&| B) | A  -->  A *)
	    (* (A |&| B) | (A |&| C)  -->  A |&| (B | C) *)
	    (* (A |&| B |&| C) | (A |&| D |&| E)  -->  A |&| (B|&|C | D|&|E) *)
	  | Call(c, Sep _) as a, b when funct_is_or && is_allt c ->
	      let la, lb = allt2l a, allt2l b in
	      (match intersection_by is_same la lb with
	      |	[] -> None
	      |	inter ->
		  let la', lb' = difference_ is_same la inter, difference_ is_same lb inter in
		  let lab = if la' = [] || lb' = [] then [] else [ l2or [ l2allt la' ; l2allt lb' ] ] in
		  Some (l2allt (lab @ inter)))

           (* A(a) |&| A(b)  -->  A(a |&| b) *)
	   (* A(a)  |  A(b)  -->  A(a | b) *)

           (* a(c) |&| b(c)  -->  (a |&| b)(c) *)
	   (* a(c)  |  b(c)  -->  (a | b)(c) *)

	  | Call(ca, pa), Call(cb, pb) when not (is_special_constr ca || is_special_constr cb) ->
	      if is_same ca cb then
		Some(Call(ca, l2orallt funct [ pa ; pb ]))
	      else if is_same pa pb then
		Some(Call(l2orallt funct [ ca ; cb ], pa))
	      else
		None

	  | Sep(",", la), Sep(",", lb) ->
	      if length la = length lb then
		Some(Sep(",", map2 (fun a b -> l2orallt funct [ a ; b ]) la lb))
	      else
		None

	  | _ -> None
	in
	let l = normalize_l (orallt2l funct e) in
	let l = keep_bests factorize l in
	let l = if funct_is_or then l else normalize_funcs_handling_constraints l in
	let r = l2orallt funct (rew_sort l) in
	r

    | Call(any, _) when is_anyval any -> anyval

    | Sep(sep, l) -> 
	let l2 = map normalize_ l in

	(* remove Builtin::Unit's in ( ; ; ), but keeping the last one *)
	if sep = ";" then 
	  let l, e = split_last l2 in
	  create_sequence (filter (fun e -> not (is_same e unit)) l) e
	else Sep(sep, l2)

    | SepL(sep, locals, l) -> SepL(sep, locals, map normalize_ l)

    (* normalize ((x,y)) -> ...    in (x,y) -> ... (see f400 & f401) *)
    | Func(n, [ Sep(",", l) ], body) -> Func(n, l, body)

    (* normalize 
         x -> A(x) |&| B  in  A |&| x->B
         x -> A(x)  |  B  in  A  |  x->B
     *)

    | Func(1, [ Local((0,0,_), _) as para ], (Call(funct, _) as e)) when is_orallt funct ->
	let higher, normal = fpartition (function 
	  | Call(f, Local((0,0,_),_)) -> Some f 
	  | _ -> None
        ) (orallt2l funct (normalize_ e)) in
	let normal' = if normal = [] then [] else [ eta_expand_again para (l2orallt funct normal) ] in
	l2orallt funct (normal' @ map (rew_remap_locals (-1)) higher)

    (* eta reduction *)
    | Func(1, [ Local((0,0,_), _) ], Call(f, Local((0,0,_),_))) 
      when rew_collect_local_vars f = [] && rew_collect_leaf (function _, Constr("Pervasives::Builtin", "Val", _) -> [1] | _ -> []) f = [] -> 
	(* don't eta reduct when the body has a val otherwise strange things happen (see f44 and F42 in infere.me) *)
	rew_remap_locals (-1) f

     (* remove v !< v *)
    | Call(Call(Constr("Pervasives::Builtin", "Val", _), Global((i, "Pervasives::Builtin", _),_)), Sep(",", [e;e'])) when is_constraint_op i && is_same e e' ->
	failwith "unused134"; 
	unit

    | Call(f, e) -> Call(normalize_ f, normalize_ e)
    | Func(nb, para, body) -> Func(nb, map normalize_ para, normalize_ body)
    | e -> e
  in
  fix_point normalize_ e
 
and normalize_l l = uniq_ is_same (rew_sort (map normalize l))

let l2ors = function
  | [] -> []
  | l -> normalize_l l

let l2allts = function
  | [] -> []
  | l -> normalize_l l

let rec get_var_typ_unaliased env var =
  match get_var_typ env var with
  | TExact(t) -> if is_var t then get_var_typ_unaliased env t else TExact(t)
  | c -> c

let rec set_var_typ_raw_unaliased env var constr =
  match get_var_typ env var with
  | TExact(v) when is_var v -> set_var_typ_raw_unaliased env v constr
  | _ -> set_var_typ_raw env var constr


let propagate_vartype_simple env e =
  match get_var_typ env e with
  | TExact(Call(Constr("Pervasives::Builtin", "Open_function", _), t)) -> t
  | TExact(t) -> t
  | _ -> e

let rec propagate_vartype env e = 
  rew_map_leaf_no_val (function
    | depth, (Local((_,stack,_),_) as e) when stack >= depth -> 
	let e2 = propagate_vartype_simple env (rew_remap_locals (-depth) e) in
	let e3 = if is_var e2 then e2 else propagate_vartype env e2 in
	rew_remap_locals depth e3
    | _, (Global(_,_) as e) ->
	propagate_vartype_simple env e
    | _, e -> e
  ) e

let propagate_vartype_no env var e = 
  rew_map_leaf_no_val (function
    | depth, (Local((_,stack,_),_) as e) when stack >= depth && not (is_same e var) -> 
	rew_remap_locals depth (propagate_vartype_simple env (rew_remap_locals (-depth) e))
    | depth, (Global(_,_) as e) ->
	rew_remap_locals depth (propagate_vartype_simple env e)
    | _, e -> e
  ) e
	
let normalize_constr c = { lower = normalize c.lower ; greater = normalize c.greater }

(* for newly created locals that don't depend on entry parameters, create them in the stack of caller *)
let allocate_new_locals def_depth expr env =
  let localvar2depth = flatten (map_index
    (fun locals s ->
      collect (fun t ->
	let l = constr_collect (rew_collect_vars 0) (typ_ t) in
	map (fun v -> var_indice v, s+1) l
      ) (Array.to_list locals)
    ) (tl (Array.to_list env.local))) in (* do not look in top of the stack *)

  let locals = Hashtbl.create 10 in

  let e, pre_allocate_e = match expr with
  | `None -> bottom, false
  | `NoPreAllocate e -> e, false
  | `PreAllocate e -> e, true
  in    
  let env2 =
    let r_env = ref env in
    let pre_allocate = rew_map_leaf (function 
      | depth, (Local((i,stack,n),_) as e) when stack = depth + def_depth -> 
	  (try let _ = Hashtbl.find locals i in ()
	  with Not_found -> 
	    let stack_off = try assoc i localvar2depth with Not_found -> 1 in
	    let (i', env) = realloc_stack !r_env stack_off 1 (get_local_flags !r_env i 0) in
	    get_local_name env i' stack_off := n ^ "_" ;
	    let env2 = set_local_typ (get_local_typ env i 0) env i' stack_off in
	    if debug then printf "creating new local Local(%d,1,%s) (was Local(%d,0,%s))\n" i' (n ^ "_") i n;
	    Hashtbl.add locals i (i',stack_off) ;
	    r_env := env2) ; e
      | _, e -> e) in
    env_iter_constr (fun var constr -> 
      if pre_allocate_e || some (var_depth var) <> 0 then
	let _ = constr_map pre_allocate constr in ()
    ) !r_env;
    if pre_allocate_e then let _ = pre_allocate e in () else () ;
    !r_env
  in
  let allocate = rew_map_leaf (function 
    | depth, (Local((i,stack,n),pos) as e) when stack = depth + def_depth -> 
	(try 
	  let (i', stack_off) = Hashtbl.find locals i in
	  Local((i',stack+stack_off,n ^ "_"), pos)
	with Not_found -> 
	  e)
    | _, e -> e) in

  let env3 = env_map_constr (fun _var constr -> Some (constr_map allocate constr)) env2 in

  let e' = allocate e in
  if debug && e <> e' then
    printf "allocate_new_locals\n";
  e', env3

let pop_stackt_allocate env =
  let _, env' = allocate_new_locals 0 `None env in pop_stackt env'

let lazyfy_and_alike env e = 
  rew_map_leaf (function 
    | depth, (Local((_,stack,_),_) as var) when stack = depth -> 
	let flags = get_var_flags env (rew_remap_locals (-depth) var) in
	let var' = io2rew var (some_or flags.inoutness In) in
	lazy_and_alike2rew var' flags.is_lazy_and_alike
    | _, e -> e
  ) e

let simplify_once propagate_or_not env e = 
  let e' = normalize e in
  rew_map_withenv (fun env (depth,e) ->
    let renv = ref env in
    let e' = match e with
    (* x |&| a  => x    when x !< a *)
    (* x  |  a  => x    when x !> a *)
    | Call(funct, _) when propagate_or_not && is_orallt funct ->
	let l = orallt2l funct e in
	let l2 = scan_list_while_modifying (fun l1 l2 e -> 
	  match e with
	  | Local((_, stack, _), _) when stack >= depth ->
	      let var = rew_remap_locals (-depth) e in
	      (match constr2lower_or_greater funct (get_var_typ_unaliased !renv var) with
	      | None -> None
	      | Some t ->
		  let needed e = not (is_same t (normalize (l2orallt funct [ e ; t ]))) in
		  Some(filter needed l1 @ [e], filter needed l2))
	  | Call(Local((_, stack, _), _) as v, _) when stack >= depth ->
	      let var = rew_remap_locals (-depth) v in
	      let constr = get_var_typ_unaliased !renv var in
	      (match constr2lower_or_greater funct constr with
	      | None -> None
	      | Some t ->
		  let needed e = 
	            let t' = normalize (l2orallt funct [ part_of_eta_expand e ; t ]) in
		    not (is_same t t') in
		  let l1', l2' = filter needed l1, filter needed l2 in
		  let l1'' = match l1' with
		  | [ strict ] when is_allt funct && is_strict strict ->
		      renv := set_var_typ_raw_unaliased !renv var 
			(match constr with
			| TConstr c -> TConstr { c with greater = l2allt [ strict ; c.greater ] }
			| _ -> internal_error "simplify_once") ;
		      []
		  | l -> l
		  in
		  Some(l1'' @ [e], l2'))
	  | _ -> None
        ) l in

	let e' = l2orallt funct l2 in
	if length l <> length l2 && debug then
	  (printf "simplified %s in %s\n" (rew2s e) (rew2s e'); dumpt !renv);
	e'

    (* try to find beta expansions: (x->None |&| Some)(x)  ==>  None |&| Some(x)  *)
    | Call(Call(c, _) as e, p) when is_orallt c ->
	let (ano, other) = fpartition (function Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], body) -> Some body | _ -> None) (orallt2l c e) in
	let ano' = map (fun body -> 
	  let p' = rew_remap_locals 1 p in
	  let body2 = rew_replace_localvar 0 0 p' body in
	  rew_remap_locals (-1) body2
	) ano in
	let other' = if other = [] then [] else [ Call(l2orallt c other, p) ] in
	l2orallt c (ano' @ other')


      (* para -> 
	     o !< a -> (a !< expr) ; lo
	     body
         => 
	 para -> 
	     o !< a -> lo
	     x !< expr
	     body
	 when every occurence of "o" is applied to "x", aka "o(x)" *)
    | Func(nblocal, para, Sep(";", 
	[Call(Call(Constr("Pervasives::Builtin", "Val", _), Global((lower1_ind,"Pervasives::Builtin",_), _)) as lower1, 
	      Sep(",", [ Local((o_ind,o_stack,_), _) as o ; 
			 Func(1, [ Local((0,0,_), _) as a ], 
			      Sep(";", [Call(Call(Constr("Pervasives::Builtin", "Val", _), Global((lower2_ind, "Pervasives::Builtin", _), _)) as lower2, 
					     Sep(",", [ Local((0,0,_), _) ; expr ])) ; 
					lo]))])) ;
	 body ])) when is_constraint_op lower1_ind && is_constraint_op lower2_ind ->
	 (match collect (rew_collect_ (function
	   | depth, Call(Local((ind,stack,_),_), (Local _ as x)) when ind=o_ind && stack+depth=o_stack -> [ x ]
	   | _ -> []
         )) (lo :: para) with
	 | x :: l when for_all (is_same x) l ->
	     let constr1 = Call(lower1, Sep(",", [ o ; eta_expand_again a lo ])) in
	     let constr2 = Call(lower2, Sep(",", [ x ; rew_remap_locals (-1) expr ])) in
	     Func(nblocal, para, Sep(";", [ constr1 ; constr2 ; body ]))
	 | _ -> e)

    | _ -> e
    in
    e', !renv
  ) env e'

let simplify_ propagate_or_not env e = fix_point_withenv (simplify_once propagate_or_not) env e
let simplify env e = simplify_ true env e
let simplify_no_propagate env e = fst(simplify_ false env e)

let wrap_close_the_world want_upper env abstract concrete =
  let concrete2 = normalize (full_deref concrete) in
  let concrete3, env =
    match concrete2 with
    | Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], e) ->
	(match partition (function Constr("Pervasives::Builtin", "Empty_tuple", _) -> true | _ -> false) (allt2l e) with
	| [], _ -> concrete2
	| _, other ->
	    let other' = if other = [] then [] else [ Func(0, [], l2allt other) ] in
	    l2allt (Constr("Pervasives::Builtin", "List", bpos) :: other')), env
    | Sep(",", _) ->
	failwith "unused380" ;
	let b', env = simplify env (l2or (or2l concrete2)) in
	Call(Constr("Pervasives::Builtin", "List", bpos), b'), env
    | e -> e, env
  in
  let r = close_the_world want_upper env abstract concrete3 in
  let r = simplify_no_propagate env r in
  r

(*
let eta_expand env e =
  if debug then printf "eta_expand: %s\n" (rew2s e);
  let pos = getpos e in

  let env2 = push_stackt env 1 def_flags in
  let call = simplify env2 (Call(rew_remap_locals 1 e, Local((0,0,"x"), pos))) in
  pop_stackt' env2 ;
  eta_expand_again (Local((0,0,"x"), pos)) call
*)
let rec reformat_to_exhibit_call_ = function
  | Call(funct, _) as e when is_orallt funct ->
      let l = orallt2l funct e in
      let calls, other = fpartition (fun e ->
	  match reformat_to_exhibit_call e with
	  | Some(c, Some e) -> Some(c, e)
	  | _ -> None
      ) l in
      let other' = if other = [] then None else Some (funct, l2orallt funct other) in
      let calls' = 
	match calls with
	| [] -> None
	| [c, e] -> Some(c, e)
	| (c, e) :: l -> 
	    if for_all (is_same c) (map fst l) then
	      Some(c, l2orallt funct (e :: (map snd l)))
	    else if for_all (is_same e) (map snd l) then
	      Some(l2orallt funct (c :: (map fst l)), e)
	    else
	      None in
      other', calls'
  | Call(c, e) when not (is_special_constr c) -> None, Some(c, e)
  | _ -> None, None

and reformat_to_exhibit_call e =
  match reformat_to_exhibit_call_ e with
  | None, None -> None
  | Some(_, other), None -> Some(part_of_eta_expand other, None)
  | None, Some(c, e) -> Some(c, Some e)
  | Some(funct, other), Some(c, e) -> Some(l2orallt funct [ c ; part_of_eta_expand other ], Some e)

let reformat_to_exhibit_call_factorizing_para e =
  let other, call = reformat_to_exhibit_call_ e in
  match call with
  | None -> None
  | Some(c, e) ->
      match other with
      | None -> Some(part_of_eta_expand (Call(c, e)), None)
      | Some(funct, other) -> Some(part_of_eta_expand (l2orallt funct [ other ; Call(c, e) ]), None)
 
let rec add_strict_to_abstract_constraints is_needed = function
  | Func(nb, para, Sep(";", l)) ->
      let (constraints, body) = split_last l in
      let constraints' = map (function
	| Call(op, Sep(",", [ Local((i, 0, n), pos) ; abstr ])) when is_abstract2 abstr && is_needed i ->
	    let l_abstr = map (add_strict_to_abstract_constraints (fun _ -> true)) (allt2l abstr) in
	    Call(op, Sep(",", [ Local((i, 0, n), pos) ; normalize (l2allt (abstract_strict :: l_abstr)) ]))
	| e -> e
      ) constraints in
      Func(nb, para, add_packed_constr constraints' body)
  | e -> e
