(* Type_inference *)
open Types
open Array
open List
open Common
open Printf
open Print
open Rew
open Rewrite
open Env
open Info
open Type_handling


exception Found
exception No_match_found
exception Resolve_best_type of string
exception Type_greater of fromrewrite * fromrewrite

let debug = Flags.debug_type_inference


let eval_some env e =
  try
    let e2 = Interpreter.eval_during_type_inference env e in
    let e3 = rew_map (function
      |	Call(Constr("Pervasives::Builtin", "Overload", _), _) as e ->
	  failwith "unused28" ;
	  let l = unoverload e in
	  if for_all (fun (v,t) -> is_same v t) l then
	    l2allt (map fst l)
	  else raise Interpreter.Not_compile_time
      |	e -> e
    ) e2 in
    Some e3
  with 
  | Interpreter.Not_compile_time
  | Interpreter.Interpreter_rew _
  | Interpreter.Interpreter_call _
  | Interpreter.Parameter_number_mismatch _ -> None


let eval_type env e = failwith "unused43?" ; mark_as_val (Interpreter.eval_during_type_inference env (mark_as_typ e))
let eval_type_some env e = map_option mark_as_val (eval_some env (mark_as_typ e))
let eval_type_no_error env e = some_or (eval_type_some env e) e

let type_declare env v value = 
  if rew_is_type env value then (
    if debug then printf "type declaration: %s = %s\n" (rew2s v) (rew2s value) ;
    let flags = get_var_flags env v in
    true, set_var_flags env v { flags with is_type = true }
  ) else false, env

let are_simple_paras para =
  failwith "unused56?" ;
  try
    for_all2 (=) (map var_indice para) (range 0 (length para))
  with
    _ -> false

let to_compile_time = function
  | Compile_time v -> v
  | _ -> raise Interpreter.Not_compile_time

let to_run_time = function
  | Compile_time v -> v, mark_as_val v
  | Run_time(v,t) -> v, t

let valtyp2s = function
  | Compile_time v -> sprintf "Compile_time %s" (rew2s v)
  | Run_time(v,t) -> sprintf "Run_time %s / %s" (rew2s v) (rew2s t)

let map_valtyp fv ft = function
  | Compile_time v -> Compile_time(fv v)
  | Run_time(v,t) -> Run_time(fv v, ft t)

let iter_valtyp fv ft = function
  | Compile_time v -> fv v
  | Run_time(v,t) -> fv v ; ft t

let eval_marked_val env e = 
  let e', _ = rew_map_once_adapting_env (fun env -> function
    | Call(Call(Constr("Pervasives::Builtin", "Val", _), _), _) as e -> 
	if debug then printf "eval_marked_val %s\n" (rew2s e);
	Some (eval_type_no_error env e, env)
    | _ -> None
  ) env e in
  e'

let rec set_var_typ env v constr =
  let constr' = get_var_typ env v in
  let set_it env constr = 
    let constr2, env2 = propagate_constr env constr v in
    set_var_typ_raw env2 v constr2
  in
  match constr', constr with
  | TFree, _ -> set_it env constr
  | _, TFree -> env

    (* need when the Trusted thing is not a function (eg: Builtin::Ref) *)
  | TExact(Call(Constr(_, "Trusted", _), t)), _ -> set_var_typ_raw env v (TExact(t))

  | TExact(v2), _ when is_var v2 ->
      if is_same v v2 then failwith (sprintf "bad news: ``%s'' points to itself" (rew2s v));
      if debug then printf "set_var_typ: forwarding from %s to %s\n" (rew2s v) (rew2s v2);
      set_var_typ env v2 constr

  | TExact(old), TExact(new_) -> 
      (match ensure_same_or_greater_Var_exists env old new_ with
      |	(Some new_', env), changed ->
	  if changed then set_it env (TExact new_') else env
      |	_ -> raise (Type_greater(old,new_)))
	  
  | TExact(a), TConstr(b) -> 
      if debug then printf "constr_unify: %s %s\n" (rew2s a) (constr2s (TConstr(b)));
      let env = ensure_greater env a b.lower in
      let env = ensure_greater env b.greater a in
      env

  |  TConstr(a), TExact(b) -> 
      if debug then printf "constr_unify: %s %s\n" (rew2s b) (constr2s (TConstr(a)));
      let env = set_it env constr in
      let env = ensure_greater env b a.lower in
      let env = ensure_greater env a.greater b in
      env

  | TConstr(a), TConstr(b) when (get_var_flags env v).pattern_matching__implying_closed_world ->
      let a_greater, env = fold_left (fun (a_greater, env) -> 
	function
	  | Call(c, (Local _ as v)) ->
	      collect_withenv (fun env e ->
		let consts, other = fpartition (function Call(c', v) when is_same c c' -> Some v | _ -> None) (or2l e) in
		l2or_ other, ensure_greater_e_l env v consts
	      ) env a_greater
	  | _ -> a_greater, env
      ) (allt2l a.greater, env) (or2l b.lower) in
      let one_way (lo, gr) = 
	let gr' = match lo, partition is_var gr with
	| [], (gr1, gr2) -> gr1 @ gr2
	| [lo], ([], gr2) -> allt2l lo @ gr2
	| lo, (gr1, gr2) -> l2or (l2allt_ gr2 @ lo) :: gr1
	in
	TConstr { lower = subval ; greater = l2allt gr' }
      in
      let constr = one_way (or2l a.lower @ or2l b.lower, a_greater @ allt2l b.greater) in
      let constr, env = propagate_constr env constr v in
      let constr = 
	match constr with
	| TFree -> TFree
	| TConstr c -> one_way(or2l c.lower, allt2l c.greater)
	| _ -> failwith "set_var_typ"
      in
      set_var_typ_raw env v constr

  | TConstr(a), TConstr(b) ->
      let constr = TConstr { lower = l2or [ a.lower ; b.lower ] ; greater = l2allt [ a.greater ; b.greater ] } in
      let env = set_var_typ_raw env v constr in
      let env = propagate_var_constr env v in
      let constr = get_var_typ env v in

      let ab = constr2tconstr constr in

      (* needed for f39: B(v1) < v < B(v2)  must implies  v1 < v2 *)
      let novar = filter (fun e -> not (is_var e)) in
      let env =
	if ab.greater = a.greater then env (* don't do anything if nothing changed *)
	else ensure_greater_l_l env (novar (allt2l b.greater)) (novar (or2l ab.lower)) in
      let env = 
	if ab.lower = a.lower then env (* don't do anything if nothing changed *)
	else ensure_greater_l_l env (novar (allt2l ab.greater)) (novar (or2l b.lower)) in
      env


and var2lower_and_greater env var =
  let lo = var2lower_or_greater constr2lower env var in
  let gr = var2lower_or_greater constr2greater env var in
  lo, gr

and add_var_typ env v t = 
  let l = match get_var_typ env v with
  | TFree -> []
  | TExact(t') -> allt2l t'
  | _ -> failwith "add_var_typ"
  in
  match is_open_function t with
  | Some t' ->
      let l' = filter_some is_open_function l in
      if length l' <> length l then failwith (sprintf "Function %s already has type contraints (%s)%s" (rew2s v) (String.concat ", " (map rew2s l)) (pos2sfull (getpos v))) ;
      set_var_typ_raw env v (TExact(Call(Constr("Pervasives::Builtin", "Open_function", bpos), l2allt(l' @ [t']))))
  | None ->
      set_var_typ_raw env v (TExact(l2allt(l @ [t])))

and set_var_allt env v value typ =
  let value = if is_var value then
    (* deep copying *)
    let v = ! !(get_var_val env value) in
    if is_same v bottom then value else v
  else value in
  let instances = unoverload ! !(get_var_val env v) in

  match type_declare env v value with
  | true, env2 ->
      if instances <> [] then
	failwith (sprintf "you can't overload function %s with %s which works over types%s" (rew2s v) (rew2s typ) (pos2sfull (getpos typ)))
      else (
	let env3 = set_var_typ env2 v (TExact typ) in
	let value2 = rew_remove_unused_sepl value in
	let value3, env4 = simplify env3 value2 in
	set_var_val env4 v value3
      )
  | _ ->
  let typ', (env, need_changing_type_super_package) = 
    match get_var_typ env v with
    | TFree -> 
	(* non pre-declared function *)
	typ, (set_var_typ_raw env v (TExact typ), false)
    | TExact(prev) -> 
	let open_funcs_, l = fpartition is_open_function (allt2l prev) in
	let open_funcs = if length open_funcs_ = 1 then allt2l (hd open_funcs_) else open_funcs_ in

	let typ', l' =
	  match length l - (if open_funcs = [] then length instances else 0) with
	  | 0 -> 
	      (* non pre-declared function *)
	      typ, l
	  | 1 ->
	      (* pre-declared function:  *)
	      (let l', prev_typ = split_last l in

	      if is_var typ then (failwith "unused255?"; prev_typ, l') else

	      match prev_typ with
	      |	Call(Constr(_, "Trusted", _), prev_typ') -> 
		  (* booh, allowing a hole in type system: do not verify the implementation comply *)
		  prev_typ', l'
	      |	_ ->
		  if func_is_greater env `Greater prev_typ typ then 
	  	    (* the type declaration was stricter. Keeping it *)
	  	    prev_typ, l'
		  else
		    match func_is_greater env `Greater typ prev_typ with
		    | true ->
	  	        (* the type declaration was not strict enough. It was just a
	  		   guide. Replacing it with the real type *)
	  		typ, l'
		    | _ -> 
	  		failwith (sprintf "%s doesn't match the pre-declared type %s%s" (rew2s typ) (rew2s prev_typ) (pos2sfull (getpos v))))
	  | _ -> failwith (sprintf "you gave 2 type declarations for %s%s\n" (rew2s v) (pos2sfull (getpos v)))
	in

	typ', 
	if open_funcs = [] then
	  set_var_typ_raw env v (TExact (l2allt (l' @ [typ']))), true
	else (
	  if exists (fun englobing -> func_have_common_subtype env typ' englobing) open_funcs then
	    (* instancing an open type *) 
	    ()
	  else
	    (*failwith (sprintf "%s doesn't match open type %s%s\n%s doesn't match open type %s" (rew2s_c env typ) (rew2s_c env prev) (pos2sfull (getpos v)) (rew2s typ) (rew2s prev)) ;*)
	    failwith (sprintf "%s doesn't match open type %s%s" (rew2s typ') (rew2s prev) (pos2sfull (getpos v))) ;
	  set_var_typ_raw env v (TExact (Call(Constr("Pervasives::Builtin", "Open_function", bpos), l2allt (open_funcs @ l')))), false
	 )
    | TConstr _ -> internal_error "set_var_allt"
  in
  let value2 = unoverload_simple (rew_remove_type_constraints value) in
  let env = set_var_val env v (overload (instances @ [ value2, typ' ])) in
  let env =
    if is_local_var v || var_package v = env.current_package then env else 
    let overload_func = { overload_var = v ; overload_val = value2 ; overload_type = typ' ; overload_is_open_type = not need_changing_type_super_package } in
    { env with overloading_funcs4pkg = env.overloading_funcs4pkg @ [ overload_func ] }
  in
  env

  (* works on types *)
and full_simplify env e = full_simplify_ true env e
and full_simplify_ propagate_or_not env e =
  let e3, env2 = simplify_ propagate_or_not env e in
  let e4, env3 = rew_map_once_adapting_env (fun env e ->
    match e with

    (* beta expansion *)
    | Call(Func(_, _, _) as f, p) -> 
	Some (infere_simple_type_call_noerror env p f)

    (* beta expansion *)
    | Call(var, para) when is_var var ->
	(match get_var_typ_unaliased env var with
	| TExact(Func _ as f) -> 
	    failwith "unused310?" ;
	    (try Some(eval_type env (Call(f, para)), env) with _ -> None)
	| _ -> None)

    (* try to find beta expansions: (x->None |&| Some)(x)  ==>  None |&| Some(x)  *)
    | Call(Call(c, _) as e, p) when is_orallt c ->
	let (ano, other) = partition (function Func _ -> true | _ -> false) (orallt2l c e) in
	let ano', env' = map_withenv (fun env f -> infere_simple_type_call_noerror env p f) env ano in
	let other' = if other = [] then [] else [ Call(l2orallt c other, p) ] in
	Some (l2orallt c (ano' @ other'), env')

    (* aliasing vars *)
    | Call(c, l) when is_orallt c ->
	(match partition is_local_var (sep2l l) with
	| var :: (_ :: _ as vars), non_vars ->
	    let env2 = fold_left (fun env v -> set_var_typ env v (TExact var)) env vars in
	    Some(Call(c, l2sep (var :: non_vars)), env2)
	| _ -> None)

    | _ -> None

  ) env2 e3 in
  e4, env3

and propagate_constr env constr var =
  let constr' = constr_map (fun t -> simplify_no_propagate env (propagate_vartype_no env var t)) constr in
  let t', env' = match constr' with
  | TFree -> TFree, env
  | TExact t -> TExact t, env
  | TConstr c ->
      (* expand local constraints *)
      (* replace v !> (c1!<o!<c2 ; o)(_)  with  c1(_) !< v !< c2(_) *)
      (*     and v !> (c1!<o!<c2 ; o)     with  c1    !< v !< c2    *)
      (*     and v !> Object -> (c1!<o!<c2 ; o) with Object -> c1 !< v !< Object -> c2 *)
      (* NOTE: we're not dropping the constraint on c2 as we could *)
      (*     and v !< (c1!<o!<c2 ; o)(_)  with  c1(_) !< v !< c2(_) *)
      (*     and v !< (c1!<o!<c2 ; o)     with  c1    !< v !< c2    *)
      (*     and v !< Object -> (c1!<o!<c2 ; o) with Object -> c1 !< v !< Object -> c2 *)
      (* NOTE: we're not dropping the constraint on c1 as we could *)
      let lower, greater =
	let (c, lower), (c', greater) = map_t2 (
	  fpartition (function
	    | Call(Call(Constr("Pervasives::Builtin", "Var_forall", _), _) as c, any) when is_anyval any ->
		map_option (fun { lower = lo ; greater = gr } ->
		  { lower = if is_subval lo then lo else Call(lo, anyval) ; 
		    greater = if is_superval gr then gr else Call(gr, anyval) }
	        ) (unpack_local_constr c)
	    | Func(0, [any], (Call(Constr("Pervasives::Builtin", "Var_forall", _), _) as c)) when is_anyval any ->
		map_option (fun { lower = lo ; greater = gr } ->
		  { lower = if is_subval lo then lo else Func(0, [any], lo) ; 
		    greater = if is_superval gr then gr else Func(0, [any], gr) }
	        ) (unpack_local_constr c)
	    | Call(Constr("Pervasives::Builtin", "Var_forall", _), _) as c ->
		let e = unpack_local_constr c in
		e
	    | _ -> None
          )) (or2l c.lower, allt2l c.greater) in
        let lower', greater' = fold_left (fun (lo,gr) c -> or2l c.lower @ lo, allt2l c.greater @ gr) (lower, greater) (c @ c') in
	(* renormalize it if modified *)
	if c <> [] || c' <> [] then normalize_l lower', normalize_l greater' else lower', greater'
      in
      let lower, greater =
	let lo1, lo2 = fpartition (function
	  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) as e ->
	      map_option (fun c -> or2l c.lower @ or2l c.greater) (unpack_local_constr_ "Var_exists" e)
	  | _ -> None) lower in
	let gr1, gr2 = fpartition (function
	  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) as e ->
	      map_option (fun c -> allt2l c.greater @ allt2l c.lower) (unpack_local_constr_ "Var_exists" e)
	  | _ -> None) greater in
	let lower', greater' = flatten lo1 @ lo2, flatten gr1 @ gr2 in
	(* renormalize it if modified *)
	if lo1 <> [] || gr1 <> [] then normalize_l lower', normalize_l greater' else lower', greater'
      in

      (* remove v !< v   and   v !> v *)
      let lower, greater = map_t2 (filter (fun e -> not (is_same var e))) (lower, greater) in

      (* replace v !> Object -> c(_)       with  v !> c *)
      (* replace v !< Object -> c(Object)  with  v !< c *)
      (*   plus the more complex versions *)
      (*         v !> Object -> e |&| c(_)       gives  v !< c |&| (Object -> e) *)
      (*         v !< Object -> e |&| c(Object)  gives  v !< c |&| (Object -> e) *)
	 
      let lower =
	map (function
	  | Func(0, [any], Call(c, any')) when is_anyval any && is_anyval any' -> c
	  | Func(0, [any], Call(c, l)) as initial when is_anyval any && is_allt c ->
	      (match fpartition (function Call(c, any) when is_anyval any -> Some c | _ -> None) (sep2l l) with
	      | [], _ -> initial
	      | cs, [] -> l2allt cs
	      | cs, unchanged -> l2allt (Func(0, [any], l2allt unchanged) :: cs))
	  | e -> e
	) lower in
      let greater = 
	map (function
	  | Func(0, [any], Call(c, s')) when is_anyval any && is_superval s' -> c
	  | Func(0, [any], Call(c, l)) as initial when is_allt c && is_anyval any ->
	      (match fpartition (function Call(c, s) when is_superval s -> Some c | _ -> None) (sep2l l) with
	      | [], _ -> initial
	      | cs, [] -> l2allt cs
	      | cs, unchanged -> l2allt (Func(0, [any], l2allt unchanged) :: cs))
	  | e -> e
	) greater in

      (* remove !< Eq  when (a, ...) !< Eq
	 propagate it to the tuple elements *)
      let greater, env =
	let is_eq = function 
	  | Call(Constr("Pervasives::Builtin", "Abstract", _), Constr("Pervasives::Standard_classes", "Eq", _)) -> true
	  | _ -> false
	in
        match lower with
	| [ Sep(",", _) as tuple ] when exists is_eq greater ->
	    let (eq, other) = partition is_eq greater in
	    other, ensure_greater_l_e env eq tuple
	| _ -> greater, env
      in    


      let find_call = function
	| Call(_, Sep _) -> None
	| Call(c, e) when not (is_special_constr c) -> Some(c, e) 
	| _ -> None 
      in

      (* add constraints of the type C(a) < var < C(b) => a < b *)
      let env = 
	let lower_constr, _ = fpartition find_call lower in
	let greater_constr, _ = fpartition find_call greater in

	fold_left (fun env (c, e) ->
	  ensure_greater_l_e env (all_assoc_by is_same c greater_constr) e
        ) env lower_constr in
     
      (* for things like v < o(a) |&| p(b) |&| C(c) *)
      (* -> ensure o=p and o < C *)
      (* -> gives v < o(a |&| b |&| c) *)
      (* *)
      (* for things like v > o(a) | p(b) | C(c) *)
      (* -> ensure o=p and o > C *)
      (* -> gives v > o(a | b | c) *)
      let handle_var_constr funct way env l =
	let handle_var_constr_l env l =
	  let calls, other = fpartition find_call l in
	  let calls', env = 
	    match partition (fun (e,_) -> is_var e) calls with
	    | (var,_) :: vars, consts ->
		let env = fold_left (fun env (e,_) -> set_var_typ env e (TExact var)) env vars in
		let env = fold_left (fun env (e,_) ->
		  let (e1,e2) = way var e in
		  ensure_greater env e1 e2
	        ) env consts in
		map (fun (_,e) -> (var, e)) calls, env
	    | _ -> calls, env
	  in other @ map (fun (c,e) -> Call(c, e)) calls', env
	in
	let handle_var_constr_once env e =
	  match e with
	  | Call(c, _) when is_same c funct ->
	      let l', env = handle_var_constr_l env (orallt2l funct e) in
	      l2orallt funct l', env
	  | _ -> e, env
	in
	let l', env = map_withenv (rew_map_shallow_withenv handle_var_constr_once) env l in
	handle_var_constr_l env l'
      in
      let lower,   env = handle_var_constr (Constr("Pervasives::Builtin", "Or",   bpos)) (fun x y -> x,y) env lower in
      let greater, env = handle_var_constr (Constr("Pervasives::Builtin", "Allt", bpos)) (fun x y -> y,x) env greater in


      (* for things like v > o(var | C)  or  v > o(var(p) | C) *)
      (* -> ensure       var > C         or  var > _ -> C *)
      (* -> gives        v > o(var)      or  v > o(var(p)) *)
      let lower, env =
	map_withenv (rew_map_shallow_withenv (fun env e ->
	  match e with
	  | Call(c, _) when is_or c ->
	      let l = or2l e in
	      (match partition is_var l with
	      | [var], constrs -> var, ensure_greater_e_l env var constrs
	      | _ -> 
		  (match partition (function Call(Local _, _) -> true | _ -> false) l with
		  | [Call(var, p)], constrs ->
		      Call(var, p), ensure_greater env var (part_of_eta_expand (l2or constrs))
		  | _ -> e, env))
	  | _ -> e, env
        )) env lower 
      in

      (* replace v !< _ -> v(expr)  with  v !< expr -> _ *)
      (* HACK: don't keep "expr -> _", remove it! *)
      let greater =
	filter_some (function
	  | Func(1, [ Local((0,0,_), _) ], Call(var', _expr)) when is_same var (rew_remap_locals (-1) var') -> 
	      failwith "unused490?" ;
	      None
	  | Func(0, [any], e) when is_anyval any ->
	      (* also handling _ -> C1 |&| v(expr) *)
	      (* giving _ -> C1 *)
	      (* HACK: should be Func(0, [ expr ], anyval) *)
	      let var' = rew_remap_locals 1 var in
	      let e' = allt2allt_option (l2allt (filter (function Call(v, _) -> not (is_same var' v) | _ -> true) (allt2l e))) in
	      map_option (fun e -> Func(0, [any], e)) e'
	  | e -> Some e
	) greater in

      (* 
	 i !> i|&|a | ...  =>  i !< a and i !> ...
         i !> a|&|b | a|&|b|&|c | ... =>  i !> a|&|b | ...

	 i !< i|a |&| ...  =>  i !> a and i !< ...
	 i !< a|b |&| a|b|c |&| ... =>  i !< a|b |&| ...
       *)
      let p = is_same var in

      let sp_lower l =
	let (r, l3) = fpartition (fun e ->
	  let list_allts = allt2l e in
	  if exists p list_allts then
	    (* needed for example "inits" *)
	    Some (l2allt (filter (o not p) list_allts))
	  else
	    None
        ) l in
	let l4 = keep_best (fun (a,b) -> 
	  let ab = normalize (l2allt [ a ; b ]) in
	  if is_same a ab then Some b else
	  if is_same b ab then Some a else None
	) l3 in
	r, l4
      in
      let sp_greater l =
	let (r, l3) = fpartition (fun e ->
	  let list_ors = or2l e in
	  if exists p list_ors then
	    Some (l2or (filter (o not p) list_ors))
	  else
	    None
        ) l in
	let l4 = keep_best (fun (a,b) -> 
	  let ab = normalize (l2or [ a ; b ]) in
	  if is_same a ab then Some b else
	  if is_same b ab then Some a else None
	) l3 in
	let l5 =
	  let abstracts, other = partition is_abstract2 (collect allt2l_func l4) in
	  if other = [] || abstracts = [] then l4
	  else
	    (* mixed abstracts and values *)

	    (match partition (fun e -> exists is_simple_const2 (or2l e)) other with
	    | [], _ -> l4
	    | consts, other ->
	        (* ensure constants follow the abstracts *)
		if simple_is_greater env (l2allt abstracts) (l2allt consts) then
		  consts @ other
		else raise(Type_greater(l2allt abstracts, l2allt consts)))
	in r, l5
      in
      let (rgreater, lower') = sp_lower lower in
      let (rlower, greater') = sp_greater greater in
      let lower, greater = lower' @ l2allt_ rlower, greater' @ l2or_ rgreater in
      let lower, greater = map_t2 (map (simplify_no_propagate env)) (lower, greater) in
      let t3 = 
	match lower, greater with
	| [], [] -> TFree
	| [], [sub] when is_subval sub -> TExact(subval)
	| [any], [] when is_anyval any -> TExact(anyval)
	| _ -> TConstr(normalize_constr { lower = l2or lower ; greater = l2allt greater }) in

      if debug && (rlower @ rgreater) <> [] then 
	printf "using the special rule for %s %s (was %s)\n" (rew2s var) (constr2s t3) (constr2s constr);

      t3, env
  in 
  pre_set_var_typ env' var t'

and propagate_var_constr env var =
  let constr = get_var_typ env var in
  let constr, env = propagate_constr env constr var in
  let r =
    match constr with
    | TFree
    | TExact _ -> None
    | TConstr c ->
      let lower, greater = or2l c.lower, allt2l c.greater in
      
      try
	(* e | ... < var < e |&| ...  =>  var = e *)
	let e = find (fun e -> exists (is_same e) greater) lower in
	let env = set_var_typ_raw env var (TExact e) in
	let lower', greater' = map_t2 (filter (o not (is_same e))) (lower, greater) in
	let env = ensure_greater env e (l2or lower') in
	let env = ensure_greater env (l2allt greater') e in
	Some env
      with Not_found -> 
        (* v     < var < v|&|A      =>  var = v     and v < A *)
        (* v|&|A < var < v|&|A|&|B  =>  var = v|&|A and v < B *)
        
        (* v|&|A < var < v|&|B  =>  var = v and v < A|&|B *)
        
        (* v|&|A     < var < v      =>  var = v and v < A  or  A     < var < v *)
        (* v|&|A|&|B < var < v|&|A  =>  var = v and v < A  or  A|&|B < var < A < v *)
        (* => we can't simplify those 2 *)
        
        (* so: split in v, common, left_lower, left_greater *)
        (* - left_greater must not be empty *)
        (* - set var = TExact(v |&| common) *)
        (* - ensure v < left_lower |&| left_upper *)
	match lower, partition is_var greater with
	| [lower], ([v], greater') ->
	    (match partition is_var (allt2l lower) with
	    | [v'], lower' when is_same v v' ->
		(* lower' is a list of allt's *)
		let common, left_lower, left_greater = intersection_and_differences is_same lower' greater' in
		if left_greater <> [] then
		  let env = set_var_typ_raw env var (TExact (l2allt (v :: common))) in
		  let env = ensure_greater env (l2allt (left_lower @ left_greater)) v in
		  Some env
		else
		  None
	    | _ -> None)
	| _ -> None
  in
  match r with
  | Some env -> env
  | None -> set_var_typ_raw env var constr

and propagate_stack env =
  let propagate_stack_ env var four =
    if (fla_ four).is_closed then (* don't bother *) env else
    let constr = typ_ four in
    let new_constr = constr_map (propagate_vartype_no env var) constr in
    if constr = new_constr then (* don't bother *) env else
      let env' = propagate_var_constr env var in
      if debug then printf "propagate_stack: %s %s (was %s)\n" (rew2s var) (constr2s (get_var_typ env' var)) (constr2s constr) ;
      env'
  in
  env_scan_while_modifying propagate_stack_ env

and resolve_best_type_ allow_packed_constrs same_or_greater use env var t = 
  let use2s = function
    | `Used_as_function -> "Used_as_function"
    | `Unused -> "Unused"
    | `Multi -> "Multi"
    | `Once io -> "Once " ^ io2s io
  in
  let force_resolve, want_upper =
    match use with
    | `Once In          -> true,  true
    | `Once Out         -> true,  false
    | `Once Inout
    | `Once Exact       -> t = TFree, false (* don't care *)
    | `Multi            -> false, false (* don't care *)
    | `Unused           -> true,  false (* don't care *)
    | `Used_as_function -> constr2lower t <> None, false
  in
  let has_var_or e =
    match partition is_local_var (or2l e) with
    | [], _ -> None
    | var::vars, l' -> Some(var, l2or (vars @ l'))
  in
  let has_var_allt e =
    match partition is_local_var (allt2l e) with
    | [], _ -> None
    | var::vars, l' -> Some(var, l2allt (vars @ l'))
  in
  if debug then printf "starting resolve_best_type for %s with use %s\n" !(get_var_name env var) (use2s use);
  let t', env' = 
    match t with
    | TExact(t) -> Some t, env

    | TConstr _ when is_tfree t -> 
	failwith "should be simplified"

    | TFree ->
	if force_resolve then Some anyval, env else None, env

    | TConstr(_) when not force_resolve -> 
	if debug then printf "resolve_best_type: keeping tied %s\n" (rew2s var);
	None, env

    | TConstr { lower = lo ; greater = gr } when is_subval lo ->
	(match has_var_allt gr with
	| Some(var', gr') -> 
	    (* needed for e1' *)
	    let env' = set_var_typ_raw env var (TExact var') in (* setting the var new val before ensure_greater *)
	    Some var', ensure_greater env' gr' var'
	| None ->
	    if not want_upper && allow_packed_constrs then
	      let name = !(get_var_name env var) in
	      Some(pack_local_constr name t), env
	    else
	      Some gr, env)

    | TConstr { lower = lo ; greater = gr } when is_superval gr ->
	(match has_var_or lo with
	| Some(var', lo') -> 
	    let env' = set_var_typ_raw env var (TExact var') in (* setting the var new val before ensure_greater *)
	    Some var', ensure_greater env' var' lo'
	| None ->
	    if want_upper && allow_packed_constrs then
	      let name = !(get_var_name env var) in
	      Some(pack_local_constr name t), env
	    else
	      Some lo, env)

    | TConstr { lower = lo ; greater = gr } ->
	(match 
	  (match has_var_or lo with
	  | Some(var', lo') -> Some(var', lo', gr)
	  | None -> match has_var_allt gr with
	    | Some(var', gr') -> Some(var', lo, gr')
	    | None -> None) with
	| Some(var', lo', gr') ->
	    let env2 = set_var_typ_raw env var (TExact var') in (* setting the var new val before ensure_greater *)
	    let env3 = ensure_greater env2 var' lo' in
	    let env4 = ensure_greater env3 gr' var' in
	    Some var', env4
	| None ->
	    (match io_cmppart want_upper false `Ensure same_or_greater env gr lo with
	    | Some t', env' -> Some t', env'
	    | None, _ -> 
		(*printf "can't ensure\n   %s\n>\n   %s\n\n" (rew2s (l2allt gr)) (rew2s (l2or lo)) ;*)
		raise (Type_greater(gr, lo))))
  in
  if t' <> None && debug then
    printf "resolve_best_type: %s is %s (was %s)\n" (rew2s var) (rew2s (some t')) (constr2s t);
  t', env'

and resolve_best_type same_or_greater use env var =
  let env = propagate_var_constr env var in
  let t = get_var_typ env var in
  let t, env = constr_map_withenv full_simplify env t in
  match resolve_best_type_ true same_or_greater use env var t with
  | Some t, env' -> set_var_typ_raw env' var (TExact t)
  | None, env' -> env'

and resolve_best_type_keep_abstract env var =
  let constr, env2 = propagate_constr env (get_var_typ env var) var in
  match resolve_best_type_ false `Greater (`Once In) env2 var constr with
  | Some t, env' -> 
      let env'' = set_var_typ_raw env' var (TExact t) in
      (t, some_or (constr2greater constr) superval), env''
  | None, _env' -> raise (Resolve_best_type "resolve_best_type_keep_abstract")

and resolve_aliased_vars env =
  let collect_local_vars_ind = rew_collect_ (function depth, Local((i,depth',_),_) when depth = depth' -> [i] | _ -> []) in
  let deps = env_local_map (fun v t -> 
    let l = match typ_ t with
    | TFree -> []
    | TExact(t) -> collect_local_vars_ind t
    | TConstr(c) -> collect_local_vars_ind c.lower @ collect_local_vars_ind c.greater
    in var_indice v, l) env in
  try
    let deps' = graph_sort deps in
    map fst deps', env
  with GraphSort_circular_deps(i,j) ->

    (* try to favour parameters, Inout *)
    let rec setting_one i j =
      if i = j then failwith "" ;
      match get_local_typ env i 0 with
      | TConstr(c) ->
	  let t = filter (fun e -> mem j (collect_local_vars_ind e)) (or2l c.lower @ allt2l c.greater) in
	  (match t with
	  | [] -> setting_one j i (* the other way, needed when the variables don't depend directly on each other (more than 2 nodes circular graph) *)
	  | e::l -> 
	      let k, t = if for_all (fun e' -> is_same e e') l then i, e else raise (Resolve_best_type "too complex") in
	      if debug then (printf "GraphSort_circular_deps setting Local(%d,0) to %s\n" k (rew2s t) ; dumpt env);
	      set_var_typ env (Local((k,0,""),bpos)) (TExact t))
      |	TExact(Local _) ->
	  (* means "j" must be propagated *)
	  propagate_var_constr env (Local((j,0,""),bpos))
      | _ -> internal_error "taking care of circular deps"
    in 
    let fl1, fl2 = map_t2 (fun i -> get_local_flags env i 0) (i, j) in
    let env' = 
      match fl1, fl2 with
      | { para_or_local = `Local }, { para_or_local = `Parameter } -> setting_one i j
      | { para_or_local = `Parameter }, { para_or_local = `Local } -> setting_one j i
      | { inoutness = None }, { inoutness = _ } -> setting_one i j
      | { inoutness = _ }, { inoutness = None } -> setting_one i j
      | { inoutness = Some io1 }, { inoutness = Some io2 } -> 
	  if less_io(io1, io2) then setting_one i j else setting_one j i
    in
    resolve_aliased_vars env'

and resolve_local_vars_ same_or_greater env in_expr out_expr = 

  let locals_sorted, env2 = resolve_aliased_vars env in

  let in_out_use, part_use = traverse_contraints env2 in_expr out_expr in
  let all_use = collect snd part_use @ in_out_use in

  let digested_use = map (fun i -> 
    i, match all_assoc i all_use with
    | [] -> `Unused
    | [io] -> `Once io
    | _ -> `Multi) locals_sorted
  in

  let env3 = fold_left (fun env i -> 
    resolve_best_type same_or_greater (assoc i digested_use) env (Local((i,0,""),bpos))
  ) env2 locals_sorted in

  let env4 = propagate_stack env3 in
  dumpt env4;

  env4

and resolve_local_vars env in_expr out_expr = resolve_local_vars_ `Greater env in_expr out_expr

and env_unpack_constr_raw env body =
  iter (set_var_name env) (rew_collect_vars 0 body) ;
  let _, body', constrs = unpack_constr body in
  (body', env), constrs

and env_unpack_constr_ env body =
  let (body2, env2), constrs = env_unpack_constr_raw env body in
  let env3 = fold_left (fun env (i, constr) ->
    set_var_typ env (Local((i, 0, ""), bpos)) (TConstr constr)
  ) env2 constrs in
  (body2, env3), constrs

and env_unpack_constr env body = fst (env_unpack_constr_ env body)

(*********************************************************************************)
(* cmppart_ **********************************************************************)
(*********************************************************************************)
and cmppart_funcs special want_upper same_or_greater env (a, b) =
  let para_a, para_b, body_a, body_b, nb_local_a, constr_a, env2 =
    match a, b with
    | Func(nblocal_a, pa, ba), Func(nblocal_b, _, _) ->
	let env2 = push_stackt env nblocal_a { def_flags with para_or_local = `Local } in
	let pa2 = map (eval_marked_val env2) pa in
	let (ba2, env3), constr_a = 
	  if special 
	  then env_unpack_constr_ env2 ba 
	  else env_unpack_constr env2 ba, [] in

	let ba3, pa3 = 
	  if special then
	    uncons (map (
		    rew_map (function
		      | Call(Constr("Pervasives::Builtin", "Var_forall", pos), e) ->
			  Call(Constr("Pervasives::Builtin", "Var_exists", pos), e)
		      | e -> e
	            )
		   ) (ba2 :: pa2))
	  else ba2, pa2 in

	let env4 = push_stackt env3 nblocal_b { def_flags with para_or_local = `Local } in
      
	let ba4, pa4 = uncons (map (rew_remap_locals 1) (ba3 :: pa3)) in

	(* removing Strict's *)
	let b' = rew_map (function
	  | Call(Constr("Pervasives::Builtin", "All_T", _), l) ->
	      let l' = filter (fun e -> not (is_strict e)) (sep2l l) in
	      l2allt l'
	  | e -> e
        ) b in

	let (pb2, bb2) = 
	  match rew_remap_locals 1 b' with
	  | Func(_, para, body) -> para, body
	  | _ -> internal_error "compare_funcs"
	in
	let pb3 = map (eval_marked_val env4) pb2 in
	let bb3, env5 = env_unpack_constr env4 bb2 in

	pa4, pb3, ba4, bb3, nblocal_a, constr_a, env5
    | _ -> failwith "inspect_funcs"
  in

  try
  let para_a', para_b' =
    let l_a, l_b = length para_a, length para_b in
    if l_a = l_b then para_a, para_b else
    if l_a = 1 then para_a, [ l2sep para_b ] else
    if l_b = 1 then (failwith "unused868?" ; [ l2sep para_a ], para_b) else
    raise (Type_greater(a,b))
  in
  let para_ab, env4 = map2_withenv (cooked_cmppart_ (io_cmppart want_upper false `Ensure same_or_greater)) env2 para_b' para_a' in

  let body_ab, env5 = io_cmppart want_upper false `Ensure same_or_greater env4 body_a body_b in

  match l_option2option_l para_ab, body_ab with
  | Some para_ab, Some body_ab ->
      (try
	let both, env6 = allocate_new_locals 0 (`PreAllocate (l2sep (body_ab :: para_ab))) env5 in
        let env7 = pop_stackt env6 in
	let body_ab', para_ab' = uncons (sep2l (rew_remap_locals (-1) both)) in
        try
	  let env8 =
	    if special then
	      fold_left (fun env i ->
		let constr = try TConstr (assoc i constr_a) with Not_found -> TFree in
		let var = Local((i, 0, ""), bpos) in
		set_var_typ env var (TExact (pack_local_constr_with_kind "Var_exists" !(get_var_name env var) constr))
	      ) env7 (range 0 nb_local_a)
	    else env7 in
          let env9 = resolve_local_vars_ same_or_greater env8 para_ab' [body_ab'] in
	  let body_ab'', para_ab'' = propagate_vartype env9 body_ab', map (propagate_vartype env9) para_ab' in

          Some(Func(0, para_ab'', body_ab'')), pop_stackt_allocate env9
        with Type_greater _ | Resolve_best_type _ | Interpreter.Parameter_number_mismatch _ -> None, pop_stackt_allocate env7
      with Type_greater _ | Interpreter.Parameter_number_mismatch _ -> None, pop_stackt_allocate (pop_stackt_allocate env5))
  | _ -> None, pop_stackt_allocate (pop_stackt_allocate env5)
  with Type_greater _ | Interpreter.Parameter_number_mismatch _ -> None, pop_stackt_allocate (pop_stackt_allocate env2)


and cmppart_simplify _want_upper _same_or_greater f env (a, b) =
  match a,b with

  | Ref _, _ -> f env (deref a) b
  | _, Ref _ -> f env a (deref b)

  (* special cases *)
  | _, Call(Call(Constr("Pervasives::Builtin", "Val", _), _), _) ->
      failwith "unused870?" ;
      (match eval_type_no_error env b with
      |	Call(Call(Constr("Pervasives::Builtin", "Val", _), _), _) -> raise No_match_found
      |	b' -> f env a b')

  | Call(Call(Constr("Pervasives::Builtin", "Val", _), _), _), _ ->
      failwith "unused876?" ;
      (match eval_type_no_error env a with
      |	Call(Call(Constr("Pervasives::Builtin", "Val", _), _), _) -> raise No_match_found
      |	a' -> f env a' b)

  | _, Call(Constr("Pervasives::Builtin", "Val", _), _) ->
      (match eval_type_no_error env b with
      |	Call(Constr("Pervasives::Builtin", "Val", _), _) -> raise No_match_found
      |	b' -> f env a b')
  | Call(Constr("Pervasives::Builtin", "Val", _), _), _ ->
      failwith "unused887?" ;
      (match eval_type_no_error env a with
      |	Call(Constr("Pervasives::Builtin", "Val", _), _) -> raise No_match_found
      |	a' -> f env a' b)

  | SepL(";", [_], _), _ | _, SepL(";", [_], _) -> failwith "bad expression in cmppart_simplify"

  | _ -> raise No_match_found

and cmppart_is_same want_upper same_or_greater f env (a, b) = 
  match a,b with
  | Float(va,_), Float(vb,_) -> if_some (va = vb) a, env
  | String(sa,_), String(sb,_) -> if_some(sa = sb) a, env
  | Constr(qa,ea,_), Constr(qb,eb,_) ->
      if Flags.check && (is_special_constr a || is_special_constr b) then failwith (sprintf "cmppart_is_same: %s and %s\n" (rew2s a) (rew2s b)) ;
      if_some (qa = qb && ea = eb) a, env

  | Call(Constr ("Pervasives::Builtin", "Abstract", _), _), Call(Constr ("Pervasives::Builtin", "Abstract", _), _) -> 
      if_some (is_same a b) a, env

  | Func(0, [any], ba), Func(0, [any'], bb) when is_anyval any && is_anyval any' ->
      let ba', bb' = map_t2 (rew_remap_locals (-1)) (ba, bb) in
      map_option_env (fun bab -> Func(0, [any], rew_remap_locals 1 bab)) (f env ba' bb')

  | Func _, Func _ ->
      cmppart_funcs false want_upper same_or_greater env (a, b)

    (* must have greater > b  and  greater > lower *)
  | Call(Constr("Pervasives::Builtin", "Var_forall", _), _), _ ->
      (match unpack_local_constr a with
      |	Some { greater = gr } when is_superval gr -> Some(if want_upper then a else b), env (* special case for returning the Var_forall instead of superval *)
      |	Some { greater = gr ; lower = lo } ->
	  (match f env gr lo with
	  | None, env2 -> None, env2
	  | _, env2 -> f env2 gr b)
      |	None -> None, env)
    (* must have lower > b  and  greater > lower *)
  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _), _ ->
      (match unpack_local_constr_ "Var_exists" a with
      |	Some { greater = gr ; lower = lo } ->
	  (match f env gr lo with
	  | None, env2 -> None, env2
	  | _, env2 -> f env2 lo b)
      |	None -> None, env)
    (* must have a > lower  and  greater > lower *)
  | _, Call(Constr("Pervasives::Builtin", "Var_forall", _), _) ->
      (match unpack_local_constr b with
      |	Some { lower = lo } when is_subval lo -> Some(if want_upper then a else b), env (* special case for returning the Var_forall instead of anyval *)
      |	Some { greater = gr ; lower = lo } ->
	  (match f env gr lo with
	  | None, env2 -> None, env2
	  | _, env2 -> f env2 a lo)
      |	None -> None, env)
    (* must have a > greater  and  greater > lower *)
  | _, Call(Constr("Pervasives::Builtin", "Var_exists", _), _) ->
      (match unpack_local_constr_ "Var_exists" b with
      |	Some { greater = gr ; lower = lo } ->
	  (match f env gr lo with
	  | None, env2 -> None, env2
	  | _, env2 -> f env2 a gr)
      |	None -> None, env)

    (* special case needed to handle (o where o > Foo) > Bar *)
  | Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], (Call(Constr("Pervasives::Builtin", "Var_forall", _), _) as a')), _ ->
      failwith "unused947?";
      (match unpack_local_constr a' with
      |	Some { greater = gr } when is_superval gr -> Some(if want_upper then a else b), env (* special case for returning the Var_forall instead of superval *)
      |	Some { greater = gr ; lower = lo } ->
	  let gr', lo' = map_t2 (fun e -> Func(0, [anyval], e)) (gr, lo) in
	  (match f env gr' lo' with
	  | None, env2 -> None, env2
	  | _, env2 -> f env2 gr' b)
      |	None -> None, env)
    (* special case needed to handle Bar > (o where o < Foo) *)
  | _, Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], (Call(Constr("Pervasives::Builtin", "Var_forall", _), _) as b')) ->
      (match unpack_local_constr b' with
      |	Some { lower = lo } when is_subval lo -> Some(if want_upper then a else b), env (* special case for returning the Var_forall instead of anyval *)
      |	Some { greater = gr ; lower = lo } ->
	  let gr', lo' = map_t2 (fun e -> Func(0, [anyval], e)) (gr, lo) in
	  (match f env gr' lo' with
	  | None, env2 -> None, env2
	  | _, env2 -> f env2 a lo')
      |	None -> None, env)


  | Sep(sa,la), Sep(sb,lb) -> 
      if sa = sb then
	for_all2_option_withenv (fun l -> Sep(sa, l)) f env la lb
      else
	None, env

  (* must be before eta_expanding *)
  | _, _ when is_var a || is_var b ->
      let _lo_a, gr_a = if is_var a then var2lower_and_greater env a else Some a, Some a in
      let lo_b, _gr_b = if is_var b then var2lower_and_greater env b else Some b, Some b in

      (match gr_a, lo_b with
      |	Some gr, Some lo ->
	  let gr', env2 = full_simplify env gr in
	  let lo', env3 = full_simplify env2 lo in
	  f env3 gr' lo'
      |	Some gr, None when want_upper -> Some gr, env
      |	None, Some lo when not want_upper -> Some lo, env
      |	_ -> Some (if want_upper then superval else subval), env)

  (* for things like  &Default_val |&| &Mapable(a) > o(b)  =>   (&Mapable |&| _->&Default_val)(a) > o(b) *)
  | Call(c, _), Call(fb, pb) when is_allt c && not (is_special_constr fb) ->
      let a' = normalize a in
      let ff, p = 
	match reformat_to_exhibit_call a' with
	| Some(f, p) -> f, p
	| None -> raise No_match_found
      in
      (match f env ff fb with
      |	Some fab, env' ->
	  (match p with
	  | Some pa ->map_option_env (fun pab -> Call(fab, pab)) (f env' pa pb)
	  | None -> (* try something else *) raise No_match_found)
      | None, _ -> (* try something else *) raise No_match_found)

  | _ -> raise No_match_found


and cmppart_is_same_only _want_upper _same_or_greater f env (a,b) = 
  match (a,b) with

  | Call(fa, pa), Call(fb, pb) when not (is_special_constr fa || is_special_constr fb) ->
      (match f env fa fb with
      |	Some fab, env2 -> map_option_env (fun p -> Call(fab, p)) (f env2 pa pb)
      |	None, env2 ->  None, env2)

  | Call(c, la), Call(c', lb) when (is_or c && is_or c') || (is_allt c && is_allt c') ->
      let bo, (l', env') = 
	for_all_withenv (fun (l, env) e ->
	  try
	    let e', env' = find_some (fun e' -> 
	      match f env e' e with
	      | Some _, env' -> Some(e', env')
	      | _ -> None
	    ) l in
	    true, (filter ((!=) (*on purpose*) e') l, env')
	  with Not_found -> false, (l, env)) (sep2l la, env) (sep2l lb)
      in if_some (bo && l' = []) a, env'

    (* special range cases *)
  | Call(Constr("Pervasives::Builtin", "Or_range", _), _), Call(Constr("Pervasives::Builtin", "Or_range", _), _) ->
      if_some (is_same a b) a, env

  | _ -> raise No_match_found

 
  (* subtyping relation only using "|&|", not "|" 
     needed for instances of Open_function's:
     x->x !< A->A|B  but A->A|B is not a valid instance of x->x
  *)
and cmppart_is_greater_allt want_upper _same_or_greater f env (a, b) = 
  match if is_abstract2_bounded a && not (is_abstract3 env b) then 
    (try
      let ab = wrap_close_the_world want_upper env a b in
      let ab = mark_as_val ab in
      (match f env ab b with
      |	None, _ -> failwith (sprintf "weird: closed type %s should be > %s" (rew2s ab) (rew2s b))
      |	Some _, env -> Some(ab, env))
    with Not_found ->
      if debug then printf "not accepting %s > %s\n" (rew2s a) (rew2s b) ;
      None) else None with
  | Some(ab, env') -> Some ab, env'
  | None ->

  match a,b with

  | Call _, Call(fb, pb) when is_abstract3 env a && is_simple_const fb ->
      (match reformat_to_exhibit_call a with
      |	Some(ca, Some(pa)) (*when is_var pa*) ->
	  let fab, p, env2 =
	    let ff, p = some (reformat_to_exhibit_call_factorizing_para b) in
	    let fab, env' = f env ca ff in
	    if fab <> None then 
	      fab, p, env' 
	    else
	      let ff, p = some (reformat_to_exhibit_call b) in
	      let fab, env'' = f env' ca ff in
	      fab, p, env''
	  in
	  (match fab, p with
	  | Some fab, Some p -> map_option_env (fun pab -> Call(fab, pab)) (f env2 pa p)
	  | Some fab, None -> Some(Call(fab, if want_upper then pa else anyval)), env2
	  | None, _ -> None, env2)
      |	None -> map_option_env (fun fab -> Call(fab, pb)) (f env (part_of_eta_expand a) fb)
      |	_ -> raise No_match_found)

  | Call(ca, _), Call _ when is_abstract3 env b && is_simple_const ca ->
      (match reformat_to_exhibit_call b with
      |	Some(cb, Some(pb)) (*when is_var pa*) ->
	  let fab, p, env2 =
	    let ff, p = some (reformat_to_exhibit_call_factorizing_para a) in
	    let fab, env' = f env ff cb in
	    if fab <> None then 
	      fab, p, env' 
	    else 
	      (if debug then printf "trying again using reformat_to_exhibit_call\n";
	      let ff, p = some (reformat_to_exhibit_call a) in
	      let fab, env'' = f env' ff cb in
	      fab, p, env'')
	  in
	  (match fab, p with
	  | Some fab, Some p -> map_option_env (fun pab -> Call(fab, pab)) (f env2 p pb)
	  | Some fab, None -> Some(Call(fab, if want_upper then superval else pb)), env2
	  | None, _ -> None, env2)
      |	_ -> raise No_match_found)

  | Call(ca, pa), Call(c, _) when is_var ca && is_var pa && (is_or c || is_simple_const c) ->
      let fab, p, env2 =
	let ff, p = some (reformat_to_exhibit_call b) in
	let fab, env' = f env ca ff in
	if fab <> None then 
	  fab, p, env' 
	else
	  let ff, p = some (reformat_to_exhibit_call_factorizing_para b) in
	  let fab, env'' = f env' ca ff in
	  fab, p, env''
      in
      (match fab, p with
      |	Some fab, Some p -> map_option_env (fun pab -> Call(fab, pab)) (f env2 pa p)
      |	Some fab, None -> Some(Call(fab, if want_upper then pa else subval)), env2
      |	None, _ -> None, env2)

  | Call(fa, pa), Call(fb, pb) when not (is_special_constr fa || is_special_constr fb) ->
      (match f env fa fb with
      |	Some fab, env2 -> map_option_env (fun p -> Call(fab, p)) (f env2 pa pb)
      |	None, env2 ->  None, env2)

   (* Default_val !> o(a)    =>   (_ -> Default_val) !> o  *)
  | Call _, Call(fb, pb) when is_abstract a && not (is_special_constr fb) ->
      map_option_env (fun fab -> Call(fab, pb)) (f env (part_of_eta_expand a) fb)

  | Call(var, pa), Call(c, _)  when is_var var && is_orallt c ->
      let ff, p = some (reformat_to_exhibit_call b) in
      let fab, env2 = f env var ff in
      (match fab, p with
      |	Some fab, Some p -> map_option_env (fun pab -> Call(fab, pab)) (f env2 pa p)
      |	Some fab, None -> Some(Call(fab, if want_upper then pa else subval)), env2
      |	None, _ -> None, env2)

  | Call(c, _), Call _ when is_abstract2 b && (is_or c || is_simple_const c) ->
      (* needed for f207 *)
      (match reformat_to_exhibit_call b with
      |	Some(cb, Some(pb)) when is_var pb ->
	  let ff, p = some (reformat_to_exhibit_call_factorizing_para a) in
	  let fab, env' = f env ff cb in
	  (match fab, p with
	  | Some fab, Some p -> map_option_env (fun pab -> Call(fab, pab)) (f env p pb)
	  | _ -> failwith "", env')
      |	_ -> raise No_match_found)

    (* (a |&| b(c)) > d(e)  -->  (_->a |&| b)(c) > d(e) *)
  | Call(c, _), Call(fb, pb) when is_allt c && not (is_special_constr fb) ->
      (match reformat_to_exhibit_call a with
      |	Some(fa, Some pa) -> 
	  (match f env fa fb with
	  | Some fab, env2 -> map_option_env (fun pab -> Call(fab, pab)) (f env2 pa pb)
	  | None, env2 -> None, env2)
      |	Some(fa, None) ->
	  map_option_env (fun fab -> Call(fab, if want_upper then superval else pb)) (f env fa fb)
      | _ -> 
          (* a|&|b > ...   => a > ... and b > ... *)
	  (* code taken from below *)
          for_all_option_withenv l2allt (fun env a -> f env a b) env (allt2l a))

  | Call _, Call(Constr _ as c, _) when is_abstract a && not (is_special_constr c) ->
      failwith "" ;
      f env (part_of_eta_expand a) c

  (* happens for things like  &C1 > (x -> &C1(x) |&| &C2),
     eta expanding: rewriting &C1  in  x -> &C1(x)
     must be before | and |&|
   *)

  (* a|&|b > ...   => a > ... and b > ... *)
  | Call(c, la), _ when is_allt c -> (* must be before .. > a|&|b => .. > a or .. > b *)
      for_all_option_withenv l2allt (fun env a -> f env a b) env (sep2l la)

  (* _ -> a|&|b > ...   =>  _ -> a > ...  and  _ -> b > ... *)
  | Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], Call(c, la)), _ when is_allt c -> (* must be before .. > _ -> a|&|b => .. > _ -> a or .. > _ -> b *)
      let la' = map (fun e -> Func(0, [anyval], e)) (sep2l la) in
      for_all_option_withenv l2allt (fun env a -> f env a b) env la'

  | Call(Call(c, _) as la, pa), Call(c', _) when is_allt c && is_allt c' && no_abstract la ->
      (* small hack: (X|&|Y)(a) > X(b) |&| Y(c)   > X(b) |&| Y(c)  and  Y(a) > X(b) |&| Y(c) *)
      for_all_option_withenv l2allt (fun env ca -> f env (Call(ca, pa)) b) env (allt2l la)

  | _, Call(c, lb) when is_allt c ->
      let l, env' = collect_some_withenv (fun env b -> f env a b) env (sep2l lb) in
      allt2allt_option (l2allt l), env'

  (* ... > _ -> a|&|b   =>  ... > _ -> a  and  ... > _ -> b *)
  | _, Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], Call(c, lb)) when is_allt c ->
      let lb' = map (fun e -> Func(0, [anyval], e)) (sep2l lb) in
      let l, env' = collect_some_withenv (fun env b -> f env a b) env lb' in
      l2allt_option l, env'

  | _ -> raise No_match_found

  (* meant to be before cmppart_is_greater_allt *)
and cmppart_is_greater want_upper _same_or_greater f env (a,b) = 
  match a,b with

    (* special case for &Object *)
  | Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])), _ ->
      Some(if want_upper then a else b), env

    (* HACK: hard coding of (a,b) !< Eq  =>  a !< Eq ; b !< Eq *)
  | Call(Constr("Pervasives::Builtin", "Abstract", _), Constr("Pervasives::Standard_classes", "Eq", _)), Sep(",", l) ->
      for_all_option_withenv (fun l -> Sep(",", l)) (fun env -> f env a) env l

    (* transforming a tuple in a list *)
  | Call(c, _), Constr("Pervasives::Builtin", "Empty_tuple", pos) when not (is_or c) ->
      f env a (Call(Constr("Pervasives::Builtin", "List", pos), subval))
  | Call(c, _), _ when not (is_or c) && for_all (function Sep(",", _) -> true | _ -> false) (or2l b) ->
      let b', env' = full_simplify env (l2or (collect sep2l (or2l b))) in
      f env' a (Call(Constr("Pervasives::Builtin", "List", bpos), b'))

    (* List > _ -> Empty_tuple  => true *)
  | Constr("Pervasives::Builtin", "List", _), Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], e) ->
      map_option (fun _ -> if want_upper then a else b) (tuple2l_ e), env

    (* special string case: Builtin::String contains *)
  | Constr("Pervasives::Builtin", "String", _), String _ -> Some(if want_upper then a else b), env
  | Constr("Pervasives::Builtin", "Char", _), String _ -> if_some (is_char b) (if want_upper then a else b), env

  | Constr("Pervasives::Builtin", "Numbers", _), Float _ -> Some(if want_upper then a else b), env
  | Constr("Pervasives::Builtin", "Numbers", _), Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float _ ; Float _ ])) -> Some(if want_upper then a else b), env

    (* special types cases (Float's, List's) *)

    (* special range cases *)
  | Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min, _) ; Float(max, _) ])), Float(nb, _) ->
      if_some (is_int nb && min <= nb && nb <= max) (if want_upper then a else b), env
  | Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min_a, _) ; Float(max_a, _) ])), 
    Call(Constr("Pervasives::Builtin", "Or_range", _), Sep(",", [ Float(min_b, _) ; Float(max_b, _) ])) ->
      if_some (min_a <= min_b && max_b <= max_a) (if want_upper then a else b), env

    (* &C |&| &D(x) > Nil  -->  (_ -> &C) |&| &D > _ -> Nil *)
    (* (&C |&| &D)(_) > Nil  -->  &C |&| &D > _ -> Nil *)
  | Call(c, _p), Constr _
  | Call(c, _p), String _  when is_allt c || is_abstract2 c ->
      (* remove it? since reformat_to_exhibit_call always fail *)
      (match reformat_to_exhibit_call a with
      |	Some(fa, Some _) -> f env fa (part_of_eta_expand b)
      | _ -> raise No_match_found)

    (* o(a) > &E  -->  o > _ -> &E *)
  | Call(c, _p), Call(Constr("Pervasives::Builtin", "Abstract", _), _) when is_var c ->
      failwith "unused1214?";
      (match reformat_to_exhibit_call a with
      |	Some(fa, Some _) -> f env fa (part_of_eta_expand b)
      | _ -> raise No_match_found)

    (* (a | b) > c |&| d  ->  a > c|&|d  or  b > c|&|d *)
    (* without this rule a|b > c  or  a|b > d  is chosen first *)
    (* which is no good for resolving A|&|B | C > A|&|B *)
    (* do not use it for Subval > ... *)
  | Call(c, la), Call(c', _) when is_or c && is_allt c' && sep2l la <> [] ->
      let l, env' = collect_some_withenv (fun env a -> f env a b) env (sep2l la) in
      l2or_option l, env'

  | _ -> raise No_match_found

  (* meant to be after cmppart_is_greater_allt *)
and cmppart2_is_greater want_upper _same_or_greater f env (a,b) = 
  match
    (match a,b with
    (* a(b) > (c | d(e))  -->  a(b) > (_->c | d)(e) *)
    | Call(fa, pa), Call(c, _) when is_orallt c && not (is_special_constr fa) && reformat_to_exhibit_call b <> None ->
	(match reformat_to_exhibit_call b with
	| Some(fb, Some pb) -> 
	    Some (match f env fa fb with
	    | Some fab, env2 -> map_option_env (fun pab -> Call(fab, pab)) (f env2 pa pb)
	    | None, env2 -> None, env2)
	| _ -> None)
    | _ -> None) with
  | Some(r, env) -> r, env
  | None ->

  match a,b with

  | _, Call(c, lb) when is_or c -> (* must be before a|b > .. => a > .. or b > .. *)
      for_all_option_withenv l2or (fun env b -> f env a b) env (sep2l lb)

  | _, Call _ when is_subval b ->
      Some(if want_upper then a else b), env

  | Call(c, la), _ when is_or c ->
      let l, env' = collect_some_withenv (fun env a -> f env a b) env (sep2l la) in
      l2or_option l, env'

  | Call(fa, pa), _ when is_var fa ->
      let _lo_a, gr_a = var2lower_and_greater env fa in
      (match gr_a with
      |	Some gr_a ->
	  let gr_a', env2 = full_simplify env (Call(gr_a, pa)) in
	  f env2 gr_a' b
      |	_ -> Some(if want_upper then a else b), env)

  | _, Call(fb, pb) when is_var fb ->
      failwith "unused1261?";
      let _lo_b, gr_b = var2lower_and_greater env fb in
      (match gr_b with
      |	Some gr_b ->
	  let gr_b', env2 = full_simplify env (Call(gr_b, pb)) in
	  f env2 a gr_b'
      |	_ -> Some(if want_upper then a else b), env)

  | _ -> raise No_match_found

and cmppart_ensure_same _want_upper _same_or_greater _f env (a,b) =
  if is_var a then
    Some a, set_var_typ env a (TExact b)
  else if is_var b then
    Some b, set_var_typ env b (TExact a)
  else
    raise No_match_found

and cmppart_ensure_greater want_upper _same_or_greater f env (a,b) =
  let (v_a, v_b) = is_var_constraintable env a, is_var_constraintable env b in

  if v_a then set_var_name env a ;
  if v_b then set_var_name env b ;

  match a,b with
  | Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", [])), _
  | _, Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", [])) ->
      Some(if want_upper then a else b), env

  | _, Call(c, lb) when is_or c && v_a ->
      for_all_option_withenv l2or (fun env b -> f env a b) env (sep2l lb)

  | _ -> raise No_match_found

and cmppart2_ensure_greater _want_upper _same_or_greater _f env (a,b) =
  let (v_a, v_b) = is_var_constraintable env a, is_var_constraintable env b in
  if v_a || v_b then
    try
      let env2 = if v_a then set_var_typ env a (TConstr { lower = b ; greater = superval }) else env in
      Some(if v_a then a else b), if v_b then set_var_typ env2 b (TConstr { lower = subval ; greater = a }) else env2
    with Type_greater _ -> None, env
  else
    raise No_match_found

and cmpparts_ensure = [
  `Same, [ cmppart_ensure_same ] ;
  `Greater,      [ cmppart_ensure_greater ; cmppart2_ensure_greater ] ;
  `Greater_allt, [ cmppart_ensure_greater ; cmppart2_ensure_greater ] ;
]
and cmpparts_is = [
  `Same, [ cmppart_is_same ; cmppart_is_same_only ] ;
  `Greater_allt, [ cmppart_is_same ; cmppart_is_greater_allt ; cmppart_is_same_only ];
  `Greater, [ cmppart_is_same ; cmppart_is_greater ; cmppart_is_greater_allt ; cmppart2_is_greater ];
]

(*********************************************************************************)
(* simple_is_ ensure_ cooked_ ****************************************************)
(*********************************************************************************)
and compute_cmppart want_upper is_or_ensure same_or_greater =
  let l = cmppart_simplify :: (if is_or_ensure = `Ensure then assoc same_or_greater cmpparts_ensure else []) @ assoc same_or_greater cmpparts_is in

  let rec compute_cmppart_ env a b =
  if debug then printf "compute_cmppart %s > %s\n" (rew2s a) (rew2s b);
  try
    find_some (fun f -> 
      try Some (f want_upper same_or_greater compute_cmppart_ env (a,b))
      with No_match_found -> None
    ) l
  with Not_found -> None, env
  in
  compute_cmppart_
  

and io_cmppart want_upper flexible_io is_or_ensure same_or_greater env a b = 
  if debug then printf "io_cmppart %s > %s\n" (rew2s a) (rew2s b);
  dumpt env ;
  let (io_a, a') = rew2io a in
  let (io_b, b'), env2 = if flexible_io then rew2best_io env io_a b else (rew2io b, env) in

  match io_a, io_b with
  | Exact, _
  | Inout, Inout ->
      compute_cmppart want_upper is_or_ensure `Same env2 b' a'
  | Out, Out | Out, Inout -> 
      compute_cmppart (not want_upper) is_or_ensure same_or_greater env2 b' a'
  | In, _ -> 
      compute_cmppart want_upper is_or_ensure same_or_greater env2 a' b'
  | _, _ -> None, env2

and option2bool_env (e,env) = e <> None, env

and simple_is_same_         env a b = option2bool_env (io_cmppart true false `Is     `Same         env a b)
and simple_is_greater_      env a b = option2bool_env (io_cmppart true false `Is     `Greater      env a b)
and ensure_same             env a b = io_cmppart true false `Ensure `Same         env a b
and ensure_greater_         env a b = io_cmppart true false `Ensure `Greater      env a b
and ensure_greater_allow_io env a b = io_cmppart true true  `Ensure `Greater      env a b

and simple_is_greater env a b = fst (simple_is_greater_ env a b)

and func_is_greater env same_or_greater a b = 
  if debug then printf "func_is_greater %s > %s\n" (rew2s a) (rew2s b);
  let r, _env = 
    match a, b with
    | Func _, Func _ -> cmppart_funcs true true same_or_greater env (a, b)
    | _ -> io_cmppart true false `Is same_or_greater env a b
  in r <> None

and ensure_same_or_greater_Var_exists env old new_ =
  match old, new_ with
  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _), Call(Constr("Pervasives::Builtin", "Var_exists", _), _) ->
      (match map_t2 (unpack_local_constr_ "Var_exists") (old, new_) with
      |	Some c_old, Some c_new_ ->
	  let env = ensure_greater env c_old.lower c_new_.lower in
	  let env = ensure_greater env c_new_.greater c_old.greater in
	  (Some new_, env), true
      |	_ -> failwith "ensure_same_or_greater_Var_exists")
  | Call(c, _), Call(Constr("Pervasives::Builtin", "Var_exists", _), _) when is_allt c ->
      (* v |&| a = (one x where x < gr)  =>  v = (one x where x < gr) and a > gr *)
      (match partition is_var (allt2l old), unpack_local_constr_ "Var_exists" new_ with
      |	([v], a), Some { lower = lo ; greater = gr } when is_subval lo ->
	  let env = ensure_greater env (l2allt a) gr in
	  let env = set_var_typ env v (TExact new_) in
	  (Some new_, env), false
      |	_ -> ensure_same env old new_, false)
  | _ -> ensure_same env old new_, false
	

and ensure_greater env a b =
  if debug then printf "ensure_greater %s > %s\n" (rew2s a) (rew2s b);
  match ensure_greater_ env a b with
  | None, _ -> raise(Type_greater(a,b))
  | Some ab, env -> 
      if debug then printf "ensure_greater %s > %s gave %s\n" (rew2s a) (rew2s b) (rew2s ab);
      env

and ensure_greater_l_e env la b = if la = [] then env else ensure_greater env (l2allt la) b
and ensure_greater_e_l env a lb = if lb = [] then env else ensure_greater env a (l2or lb)
and ensure_greater_l_l env la lb = if la = [] || lb = [] then env else ensure_greater env (l2allt la) (l2or lb)

and cooked_cmppart_ (f : env -> fromrewrite -> fromrewrite -> (fromrewrite option * env)) env a b =
  let (a,lz_a), (b,_lz_b) = map_t2 unlazy_and_alike (a, b) in
  let env' =
    if lz_a = Some `Dynamic then
      fold_left (fun env v ->
	let flag = get_var_flags env v in
	set_var_flags env v { flag with is_lazy_and_alike = Some `Dynamic }
      ) env (rew_collect_all_accessible_vars b)
    else env
  in
  f env' a b

and cooked_cmppart f env a b =
  if debug then printf "cooked_ensure_greater %s > %s\n" (rew2s a) (rew2s b);
  dumpt env ;
  match cooked_cmppart_ f env a b with
  | None, _ -> raise(Type_greater(a,b))
  | Some _, env -> env

and cooked_ensure_greater env a b = cooked_cmppart ensure_greater_ env a b
and cooked_ensure_greater_io env a b = cooked_cmppart ensure_greater_allow_io env a b


and func_have_common_subtype env a b = 
  if debug then printf "func_have_common_subtype %s and %s\n" (rew2s a) (rew2s b);
  match a, b with
  | Func(nblocal_a, pa, ba), Func _ ->
      let env = push_stackt env nblocal_a { def_flags with para_or_local = `Local } in
      let pa = map (eval_marked_val env) pa in
      let ba, env = env_unpack_constr env ba in
      
      let (bb, pb), env = 
	match rew_remap_locals 1 b with
	| Func(nblocal_b, pb, bb) -> 
	    let env = push_stackt env nblocal_b { def_flags with para_or_local = `Local } in
	    let pb = map (eval_marked_val env) pb in
	    let bb, env = env_unpack_constr env bb in
	    
	    let both, env = allocate_new_locals 0 (`PreAllocate (l2sep (bb :: pb))) env in
	    
	    uncons (sep2l (rew_remap_locals (-1) both)), pop_stackt env
	| _ -> internal_error "func_have_common_subtype"
      in

      let (i, env) = realloc_stack env 0 2 { def_flags with para_or_local = `Parameter } in
      let para, body = Local((i, 0, "para"), bpos), Local((i + 1, 0, "body"), bpos) in
      set_var_name env para ; set_var_name env body ;

      let bo =
	try
	  let env = ensure_greater_l_e env [ l2sep_ pa ; l2sep_ pb ] para in
	  let env = ensure_greater_e_l env body [ ba ; bb ]  in
      
	  let env = resolve_local_vars env [para] [body] in

	  if debug then dumpt_forced_ env true ;
	  true
	with Type_greater _ -> false in 
      let _ = pop_stackt env in
      bo

  | _ -> 
      if is_anyval a || is_anyval b then true else 
      true (* failwith "todo1433" *)

(*********************************************************************************)
(* infere_ ***********************************************************************)
(*********************************************************************************)

and alias_or_allt_paras env _expr =
  env (*
  let ors = rew_collect_ (function depth, Call(Constr("Pervasives::Builtin", funct, _), Sep(",", l)) when mem funct ["Or"; "All_T"] -> [map (rew_remap_locals (-depth)) l] | _ -> []) expr in
  fold_left (fun env l ->
    match filter is_var l with
    | [] -> env
    | e :: l -> fold_left (fun env v -> set_var_typ env v (TExact(e))) env l
  ) env ors *)
   
and infere_func is_pattern result_context env func_name expr =
  let is_unset nb env =
    match func_name with
    | None -> true
    | Some var -> get_var_typ env (rew_remap_locals nb var) = TFree
  in
  match expr with
  | Func(nblocal, para, body) ->
      let was_unset = is_unset 0 env in
      let env2 = push_stackt env nblocal { def_flags with para_or_local = `Parameter } in
      let is_lazy_and_alike = map unlazy_and_alike para in
      let (both, env3) = map_withenv (infere_run_time true None) env2 (map fst is_lazy_and_alike) in
      let (para_o, tpara) = split both in
      let env4 = fold_left alias_or_allt_paras env3 tpara in

      (* set is_lazy *after* infering para's, otherwise you get unlazy(x) -> unlazy(x) *)
      let env4' = fold_left (fun env -> function 
	| v, Some lazy_and_alike ->
	    let flag = get_var_flags env v in
	    set_var_flags env v { flag with is_lazy_and_alike = Some lazy_and_alike }
	| _ -> env
       ) env4 is_lazy_and_alike in
      
      let ((body_o, tbody), env5) = (infere_run_time is_pattern result_context) env4' body in

      let dead_vars, env6 = if was_unset && not (is_unset 1 env5) then (
        let func_name' = rew_remap_locals 1 (some func_name) in
	match get_var_typ env5 func_name' with
	| TExact(Func(_, para', body')) ->
	    let (para'', body'') = map (rew_remap_locals (-1)) para', rew_remap_locals (-1) body' in
	    let env7 = fold_left2 ensure_greater env5 para'' tpara in
	    let env8 = ensure_greater env7 body'' tbody in
	    let _vars = collect rew_collect_all_accessible_vars (body'' :: para'') in

	    let env8' =
	      let flags = get_var_flags env8 func_name' in
	      set_var_flags env8 func_name' { flags with is_attached_to_a_variable = false } in

	    let dead_vars = find_created_vars_now_dead env8' in

	    (* create new variables in current depth, aka more local than dead_vars *)
	    let env9 = fold_left (fun env v ->
	      let l, env2 = allocate_new_vars_in_depth env 1 (get_var_flags env v) v in
	      if debug then printf "creating new vars in depth %s\n" (String.concat ", " (map rew2s l)) ;
	      let v' = hd l in
	      let env3 = set_var_typ_raw env2 v' (get_var_typ env2 v) in
	      let env4 = set_var_typ_raw env3 v (TExact v') in
	      env4
	    ) env8' dead_vars in

	    let env10 = propagate_stack env9 in
	    dead_vars, env10
	| _ -> internal_error "infere_func"
      ) else [], env5 in

      let env7 = env6 in

      let para_o' = map (lazyfy_and_alike env7) para_o in
      let tboth = map (propagate_vartype env7) (tbody :: map (lazyfy_and_alike env7) tpara) in
      let tboth', env8_ = map_withenv full_simplify env7 tboth in
      let env8 = resolve_local_vars env8_ (tl tboth') [hd tboth'] in

      let tboth2 = map (propagate_vartype env8) tboth' in

      (* removing bindings before allocate_new_locals which would create new vars *)
      let env8' = fold_left (fun env v -> set_var_typ_raw env v TFree) env8 dead_vars in

      let still_alive = find_still_alive_local_vars env8' in

      let constraints =
	if still_alive = [] then
	  rev (env_local_collect (fun v t -> 
	    if not (member_ is_same v still_alive) then pack_constr v (typ_ t) else []
	  ) env8')
	else
	  [] in

      let tbody2, tpara' = add_packed_constr constraints (hd tboth2), tl tboth2 in
      let tboth2', env9 = map_withenv full_simplify env8' (tbody2 :: tpara') in

      let tboth3, nblocal' = if still_alive = [] then rew_pack_locals tboth2' else tboth2', nblocal in

      let c = if still_alive = [] then (fun e -> `NoPreAllocate e) else (fun e -> `PreAllocate e) in
      let tboth4, env10 = map_fst sep2l (allocate_new_locals 0 (c (l2sep tboth3)) env9) in

      let (val_, typ_, env11) = 
	Func(nblocal, para_o', body_o), 
	Func(nblocal', tl tboth4, hd tboth4), pop_stackt_allocate env10 in
      let typ_', env12 = full_simplify env11 typ_ in

      Run_time(val_, typ_'), env12
  | _ -> internal_error "infere_func (2)"


and infere_simple_type_call_noerror env tpara tfunc =
  let e = Call(tfunc, tpara) in
  match eval_type_some env e with
  | Some e' -> e', env
  | None -> infere_simple_type_call env tpara tfunc (tfunc, tpara)

and infere_simple_type_call env tpara tfunc (f_for_error,para_for_pos) =
  let tpara = flatten_tuple tpara in
  match tfunc with
  | Func(nblocal, in_para, tbody) ->
      let nb_para = sep2nb tpara in
      let nb_in_para_min, strictness = compute_nbpara_bounds in_para in
      let para' =
	if nb_para < nb_in_para_min then raise (Interpreter.Parameter_number_mismatch (sprintf "missing parameters to function %s%s (%d < %d)" (rew2s f_for_error) (pos2sfull (getpos para_for_pos)) nb_para nb_in_para_min))
	else if strictness = `Strict then
	  if nb_in_para_min = 1 then
	    [ tpara ]
	  else if nb_para > nb_in_para_min then
	    raise (Interpreter.Parameter_number_mismatch (sprintf "too many parameters (%d > %d)%s" nb_para nb_in_para_min (pos2sfull (getpos para_for_pos))))
	  else 
	    sep2l tpara
	else (* `NoLimit *)
	  let a,b = break_at_indice nb_in_para_min (sep2l tpara) in
	  a @ [ Call(Constr("Pervasives::Builtin", "Remaining", bpos), l2sep b) ]
      in

      let env = push_stackt env nblocal def_flags in
      let t, env = env_unpack_constr env tbody in
	  
      let in_para' = map (eval_marked_val env) in_para in
      let env = fold_left2 cooked_ensure_greater_io env in_para' (map (rew_remap_locals 1) para') in
	  
      let env = resolve_local_vars env [] [t] in

      let t2 = propagate_vartype env t in
      let t3, env = full_simplify env (eval_marked_val env t2) in
      (* for newly created locals that don't depend on entry parameters, create them in the stack of caller *)
      let t'', env = allocate_new_locals 0 (`PreAllocate t3) env in
      rew_remap_locals (-1) t'', pop_stackt env
  | _ -> failwith "infere_simple_type_call"

and infere_call is_pattern result_context env f_o tf para = 
  let infere_call_compile_time env valtyps = 
    let para' = to_compile_time valtyps in
    let f_o' = if is_var f_o then ! ! (get_var_val env f_o) else f_o in
    if may_be_a_func f_o' && not (has_dynamic_para f_o') then
      match eval_some env (Call(rew_remove_type_constraints f_o', para')) with
      |	Some v -> Compile_time (v), env
      |	None -> raise Interpreter.Not_compile_time
    else
      raise Interpreter.Not_compile_time
  in
  let tf = un_inout tf in

  let is_special = (* special variables are hard-coded, so don't create a prototype for them *)
    match tf with
    | Global((i, "Pervasives::Builtin", _), _) -> i = builtin_op__assign || is_constraint_op i
    | _ -> false
  in

  let tf', env = 
    if is_var tf && not is_special then (
      let env = resolve_best_type `Greater `Used_as_function env tf in
      propagate_vartype_simple env tf, env
     ) else tf, env
  in

  (* verifying tf is Func(...), otherwise need creating one *)
  let tf'', env = 
    if is_var tf' && not is_special then
      let flag = get_var_flags env tf' in
      if is_pattern then
	tf', set_var_flags env tf' { flag with is_constr = true }
      else if flag.is_constr then
	tf', env
      else
	begin
	  dump env ;
	  let paras = match para with Sep(",", l) | SepL(",", _, l) -> l | e -> [e] in
	  let vars, env = allocate_new_vars env (length paras + 1) (get_var_flags env tf') tf' in
	  let vars' = map (rew_remap_locals 1) vars in
	  let paras_io = map (function
	    | Call(Global((i, "Pervasives::Builtin", _), _), _) when i = builtin_op__ref -> Inout
	    | _ -> In
          ) paras in
	  let new_paras = map2 io2rew (tl vars') paras_io in
	  let tf'' = Func(0, new_paras, hd vars') in
	  if debug then printf "creating a prototype for %s: %s\n" (rew2s tf') (rew2s tf'');
	  tf'', set_var_typ env tf' (TExact tf'')
	end
    else tf', env
  in
  (match tf'' with
	(* HACK: Dynamic transform the argument in a runtime value *)
      |	Constr(_, "Dynamic", _) ->
	  map_fst (function t,v -> Run_time(t,v)) (infere_run_time is_pattern result_context env para)

      |	Constr("Pervasives::Builtin", "Virtual_class", _) ->
	  (match para with
	  | SepL(",", _, name :: methods) ->
	      let class_ = rew_remove_unused_sepl (Interpreter.eval_during_type_inference env name) in
	      let class_, env = simplify env class_ in

	      let cmethods = map (fun m -> global2name name, m) methods in

	      let env = env_add_class env class_ cmethods in

	      let class4pkg = 
		let class4pkg = env.virtual_classes.class4pkg in
		{ class4pkg with p_classes = (class_, { p_methods = cmethods }) :: class4pkg.p_classes } in

	      Compile_time(unit), { env with virtual_classes = { env.virtual_classes with class4pkg = class4pkg } }
	  | _ -> failwith "")

      |	Constr("Pervasives::Builtin", "Constant", _) ->
	  (match para with
	  | String(s, pos) -> Compile_time(Constr(env.current_package, s, pos)), env
	  | e -> failwith ("Pervasives::Builtin::Constant: bad value " ^ rew2s e))

      |	Constr("Pervasives::Builtin", "Pattern_matching_vars", _) ->
	  (match para with
	  | Sep(",", [ vars ; expr ]) ->
	      let env = fold_left (fun env v ->
		let flag = get_var_flags env v in
		set_var_flags env v { flag with pattern_matching__implying_closed_world = true }
	      ) env (sep2l_ vars) in
	      infere is_pattern result_context env expr
	  | _ -> failwith "bad Restrict_those_vars_types")

      | Constr("Pervasives::Builtin", "Freeze", _) ->
	  (* needed for tf509 *)
	  let (i, env) = realloc_stack env 0 1 def_flags in
	  let t = Local((i, 0, "Freeze"), bpos) in
	  set_var_name env t ;
	  Run_time(Call(f_o, para), t), env

      | Call(Constr("Pervasives::Builtin", "Val", _), var) when is_var var ->
	  failwith "unused1623?" ;
	  let ((_v,tpara), env) = infere_run_time is_pattern None env para in
	  let t = eval_type env (Call(tf'', tpara)) in
	  Run_time(bottom, t), env

      |	Constr("Pervasives::Builtin", "Or", pos) ->
	  let ((v,t), env) = infere_run_time is_pattern None env para in
	  let t' = l2sep (map detype (sep2l t)) in
	  Run_time(Call(f_o, v), Call(Constr("Pervasives::Builtin", "Type", pos), Call(tf', t'))), env

      | Call(Constr("Pervasives::Builtin", "Overload", _), _) ->
	  failwith "unused1634?" ;
	  (match unoverload tf'' with
	  | [ val_, typ ] -> 
	      infere_call is_pattern result_context env val_ typ para
	  | _ -> failwith "overloaded function")

      | Func(_nblocal, in_para, _) ->
	  let (valtyps, env) = infere is_pattern (adjust_para para in_para) env para in
	  (try
	    infere_call_compile_time env valtyps
	  with Interpreter.Not_compile_time ->
	    let (para_o, tpara) = to_run_time valtyps in
	    let t', env = infere_simple_type_call env tpara tf'' (f_o,para) in
	    Run_time(Call(f_o, para_o), t'), env)

      | Call(c, e) when is_allt c -> 
	  if is_const e then
	    let (valtyp, env) = infere is_pattern None env para in
	    map_valtyp (fun v -> Call(f_o, v)) (fun t -> Call(Call(c, e), t)) valtyp, env
	  else
	    let l = allt2l tf'' in
	    let nbpara = sep2nb para in
	    let l_nb = map (function Func(_, para, _) as e -> length para, e | e -> 0, e) l in
	    let l_nb' = 
	      let l_nb' = if nbpara == 1 then [] else filter (fun (nb,_) -> nb = nbpara) l_nb in
	      let l_nb' = if l_nb' <> [] then l_nb' else filter (fun (nb,_) -> nb = 1) l_nb in
	      if l_nb' <> [] then l_nb' 
	      else
		let nbs = map fst l_nb in
		let error =
		  if for_all (fun nb -> nbpara > nb) nbs then "too many parameters" else
		  if for_all (fun nb -> nbpara < nb) nbs then "not enough parameters" else "bad number of parameters" in
		raise (Interpreter.Parameter_number_mismatch (error ^ pos2sfull (getpos para)))
	    in
	    let l' = map snd l_nb' in
	    let in_para = match l' with Func(_, p, _) :: _ -> p | _ -> failwith "" in
	    let valtyps, env = infere is_pattern (adjust_para para in_para) env para in
	    (try
	      infere_call_compile_time env valtyps
	    with Interpreter.Not_compile_time ->
	      let (para_o, tpara) = to_run_time valtyps in
	      match partition_either (fun t -> 
		try 
		  Left (infere_simple_type_call env tpara t (f_o,para))
		with e -> 
		  (match e with
		  | Type_greater _ | Resolve_best_type _ | Interpreter.Parameter_number_mismatch _ -> Right e
		  | _ -> raise e)) l' with
	      | [], errs -> raise (hd errs)
	      | (t', env) :: [], _ -> Run_time(Call(f_o, para_o), t'), env
	      | _ -> failwith "TODO")

      |	Call(c, _) when is_or c ->
	  (* needed for f44 *)
	  let (both, env) = map_withenv (fun env tf -> infere_call is_pattern result_context env f_o tf para) env (or2l tf'') in
	  let (lv, lt) = map_t2 l2or (split (map to_run_time both)) in
	  Run_time(lv, lt), env

      |	any when is_anyval any -> 
	  Run_time(Call(f_o, para), anyval), env

      |	Global((i, "Pervasives::Builtin", _), _) when i = builtin_op__assign ->
	  (match para with
	  | Sep(",", [ lvalue ; value ]) -> 
	      if is_global_var lvalue && not Flags.quiet then (eprintf "Starting %s\n" (rew2s lvalue); flush stderr);
	      let env = fold_left (fun env v -> 
		set_var_name env v ; snd (set_var_inoutness env (Some Out) v)
	      ) env (rew_collect_all_accessible_vars lvalue) in
	      let (value', tvalue), env = 
		if is_func value then
		  map_fst to_run_time (infere_func is_pattern None env (Some lvalue) value)
		else
		  infere_run_time is_pattern None env value
	      in
	      let isfunc = 
		try
		  may_be_a_func value || constr_is_open_function (get_var_typ env lvalue) ||
		  may_be_a_func ! !(get_var_val env value) || is_func (propagate_vartype_simple env value)
		with _ -> false in
	      if isfunc then (
		let env = set_var_allt env lvalue value' tvalue in
		(if is_global_var lvalue then
		  (* no need to keep it *)
		  Compile_time(unit)
		else
		  Run_time(Call(f_o, Sep(",", [ lvalue ; value' ])), unit)), env
	      ) else
		Run_time(Call(f_o, Sep(",", [ lvalue ; value' ])), unit),
		(match lvalue with
		| Global _ ->
		    let _, env = type_declare env lvalue value' in
		    let env = set_var_val env lvalue value' in
		    set_var_typ env lvalue 
		      (if get_var_typ env lvalue = TFree then TExact tvalue else TConstr { lower = tvalue ; greater = superval })
		| _ -> cooked_ensure_greater_io env lvalue tvalue)
	  | _ -> failwith "bad para to :=")
	
      |	Global((op, "Pervasives::Builtin", _), _) when is_special ->
	  (match para with
	  | Sep(",", [ left; right ]) ->

	      let left = rew_remove_unused_sepl left in

	      let env = fold_left (fun env var ->
		set_var_name env var ;
		let flags = get_var_flags env var in
		set_var_flags env var { flags with is_constr = true }
	      ) env (rew_collect_all_accessible_vars left) in

	      let rec rewrite env e =
		let expand_var env var =
		  let (value, typ, flags) = get_var env var in
		  if not flags.is_type && TExact(value) <> typ || is_same value bottom then 
		    var
		  else 
		    unoverload_simple value
		in
		rew_map_once_adapting_env (fun env e -> 
		  match e with
		  | Call(Constr("Pervasives::Builtin", "Lvalue", _), e) -> 
		      failwith "unused1751?" ;
		      Some (rewrite env e)
		  | Call(v, _) when is_var v -> 
		      let e' = eval_some env e in
		      map_option (fun e -> e, env) e'
		  | var when is_var var -> Some (expand_var env var, env)
		  | _ -> None) env e
	      in
	      let tright, env = rewrite env (rew_remove_unused_sepl right) in

	      (*let tright_ = rew_map unoverload_simple tright in*)
	      let tright' = mark_as_val tright in
	      let tright'', env = full_simplify env (rew_remove_unused_sepl tright') in

	      let env = 
		if op = builtin_op__has_type then
		  let t = rew_map_once (function
		      | Func(_nb, para, Sep(";", l)) as e ->
			  let (_constraints, body) = split_last l in
			  let l1 = count_uniq (collect rew_collect_local_vars para) in
			  let l2 = count_uniq (rew_collect_local_vars body) in
			  let vars_needing_strict = filter_some (fun (e,nb) -> if nb > 1 then Some e else None) (l1 @ l2) in
			  Some (add_strict_to_abstract_constraints (fun i -> mem i vars_needing_strict) e)
		      |	_ -> None) tright'' in
		  add_var_typ env left t
		else if op = builtin_op__supertype then
		  ensure_greater env left tright''
		else if op = builtin_op__subtype then
		  if not (is_var left) || is_capitalised_var left then
		     (* HACK for now, hard coded special case:
			Use the Int !< Addable as a hint to know instances.
			It should only be a verification
		      *)
		      let instance = rew_remove_unused_sepl (unoverload_simple (Interpreter.eval_during_type_inference env left)) in
		      let instances, env = collect_withenv (fun env -> function
			  | [ Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], body) ] when rew_collect_all_accessible_vars body = [] ->
			      let instance, env' = simplify env (Call(instance, anyval)) in
			      map (fun l -> l2allt l, instance) (list_of_nonempty_sublists (allt2l body)), env'
			  | l ->
			      let l' = lvector_product (map (function
				| Func(0, [ Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) ], body) when rew_collect_all_accessible_vars body = [] ->
				    map (fun l -> Func(0, [anyval], l2allt l)) (list_of_nonempty_sublists (allt2l body))
				| l -> [ l ]
			      ) l) in
			      map (fun e -> l2allt e, instance) l', env
		      ) env (list_of_nonempty_sublists (allt2l tright'')) in
		      
		      let instances = filter (fun (class_, _) -> assoc_has_key class_ env.virtual_classes.class2info) instances in

		      let env = env_add_instances env instances in

		      let class4pkg = 
			let class4pkg = env.virtual_classes.class4pkg in
			{ class4pkg with p_instances = instances @ class4pkg.p_instances }
		      in

		      { env with virtual_classes = { env.virtual_classes with class4pkg = class4pkg } }
		  else
		    ensure_greater env tright'' left
		else failwith "" in
	      
		Run_time(Call(f_o, Sep(",", [ left ; tright ])), unit), env
	  | _ -> failwith "")

      |	c when is_const c ->
	  let result_context' = match result_context with Some(Call(_,e)) -> Some e | _ -> None in
	  let (valtyp, env) = infere is_pattern result_context' env para in
	  map_valtyp (fun v -> Call(f_o, v)) (fun t -> Call(c, t)) valtyp, env
      |	v when is_var v && (is_pattern || (get_var_flags env v).is_constr) ->
	  failwith "unused1815?" ;
	  let result_context' = match result_context with Some(Call(_,e)) -> Some e | _ -> None in
	  let (valtyp, env) = infere is_pattern result_context' env para in
	  map_valtyp (fun v -> Call(f_o, v)) (fun t -> Call(v, t)) valtyp, env

      |	_ -> 
	  let expr = Call(f_o,para) in
	  failwith (sprintf "missing type for %s%s in %s" (rew2s tf'') (pos2sfull (getpos expr)) (rew2s expr)))

  (* HACK for ; *)
and hack_for_semicolon env = 
  let f l = 
    let l', last = split_last l in
    last, fold_left (fun env e -> ensure_greater env e unit) env l'
  in function
    | Sep(";", l) -> f l
    | SepL(";", _, l) -> failwith "unused1831?" ; f l
    | e -> e, env

and infere_sepL_ed env s l =
  let t, env2 = hack_for_semicolon env (Sep(s, l)) in
  let env3 = resolve_local_vars env2 [] [t] in

  let locals = env_local_map (fun v _ -> propagate_vartype_simple env3 v) env3 in

  let local_constraints = collect (constr_collect (fun e -> [e])) (env_local_map (fun _ t -> typ_ t) env3) in
  let env4 = fold_left alias_or_allt_paras env3 (t :: local_constraints) in
  let t' = propagate_vartype env4 t in
  let t'', env4 = allocate_new_locals 0 (`PreAllocate t') env4 in
  t'', locals, env4

(********************************************************************************)
and (infere : bool -> fromrewrite option -> env -> fromrewrite -> valtyp * env) = fun is_pattern result_context env expr ->
  if debug then printf "infering %s\n" (rew2s expr);
  check_expr true env None expr ;

  let result_context_io = match result_context with
  | None -> In
  | Some e -> fst(rew2io e)
  in
  let valtyp, env' = match expr with

  | Local(_, pos)
  | Global(_, pos) -> 
      set_var_name env expr ;
      let (val_, typ, flags) = get_var env expr in
      (match
	if is_same val_ bottom then
	  Some expr
	else if may_be_a_func val_ || typ <> TExact(val_) then
	  Some val_
	else
	  None with
      |	None -> Compile_time(val_), env
      | Some val2 -> 
	  let expr2 = 
	    if flags.is_type then 
	      val2 
	    else if flags.is_lazy_and_alike = Some `Lazy then
	      Call(External("unlazy", pos), expr)
	    else expr
	  in
	  let typ, env2 = full_simplify env (propagate_vartype_simple env expr) in
	  let typ2, env3 = rew_map_shallow_withenv env_unpack_constr env2 typ in
	  let typ3 = 
	    if flags.is_attached_to_a_variable && expr <> typ2 && result_context_io <> In then 
	      inout typ2 
	    else 
	      typ2 
	  in Run_time(expr2, typ3), env3)
  | Float(_,_)
  | String(_,_)
  | Constr(_,_,_) -> Compile_time(expr), env

  | External("Superval", _) -> Compile_time(superval), env
  | External("Subval", _) -> Compile_time(subval), env

  | External(_,_) -> Run_time(expr, anyval), env

  | Ref _ -> failwith "No Ref here, please!"

  | Func(_, _, _) -> infere_func is_pattern result_context env None expr

  | Call(Constr("Pervasives::Builtin", "Lvalue", _), e) -> 
      let valtyp, env' = infere true result_context env e in
      (match valtyp with
      |	Compile_time _ -> valtyp, env'
      |	Run_time(_, t) ->
	  valtyp, fold_left (fun env v -> 
	    let flags = get_var_flags env' v in
	    set_var_flags env v { flags with para_or_local = `Parameter }
	  ) env' (rew_collect_vars 0 t))

  | Call(f, para) -> 
      let ((f_o, tf), env2) = infere_run_time is_pattern None env f in
      infere_call is_pattern result_context env2 f_o tf para

  | SepL(s, locals, l) -> 
      let l' = (*extract_and_sort_defs*) id l in
      let result_context' = match result_context with
      | Some(Sep(_, l)) when length l = length l' -> map (fun e -> Some e) l
      |	_ -> map (fun _ -> None) l' in
      let env2 = push_stackt env (length locals) def_flags in
      let (valtyps, env3) = map2_withenv (fun env cont e -> infere is_pattern cont env e) env2 result_context' l' in
      (try
	let val_ = Sep(s, map to_compile_time valtyps) in
	let val_', env4 = hack_for_semicolon env3 val_ in
	Compile_time(val_'), pop_stackt env4
      with Interpreter.Not_compile_time ->
	let (l_o, tl) = split (map to_run_time valtyps) in
	let t, locals', env4 = infere_sepL_ed env3 s tl in
	Run_time(SepL(s, map (fun e -> [e]) locals', l_o), rew_remap_locals (-1) t), pop_stackt env4
      )

  | Sep (s, l) ->
      let result_context' = match result_context with
      | Some(Sep(_, l)) -> map (fun e -> Some e) l
      |	_ -> map (fun _ -> None) l in
      let (valtyps, env2) = map2_withenv (fun env cont e -> infere is_pattern cont env e) env result_context' l in
      (try
	let val_ = Sep(s, map to_compile_time valtyps) in
	let val_', env3 = hack_for_semicolon env2 val_ in
	Compile_time(val_'), env3
      with Interpreter.Not_compile_time ->
	let (l_o, tl) = split (map to_run_time valtyps) in
	let t, env3 = hack_for_semicolon env2 (Sep(s, tl)) in
	Run_time(Sep(s, l_o), t), env3
      )
  in 
  if debug then printf "INFERED %s --> %s\n" (rew2s expr) (valtyp2s valtyp);
  iter_valtyp (check_expr true env' None) (check_expr false env' None) valtyp ;
  valtyp, env'

and infere_run_time is_pattern result_context env expr =
  let valtyp, env' = infere is_pattern result_context env expr in
  to_run_time valtyp, env'

(*********************************************************************************)
(* short naming ******************************************************************)
(*********************************************************************************)
and (rew2names_ : ((Types.fromrewrite * int) * Types.fromrewrite) list -> env -> fromrewrite -> fromrewrite * env) = fun type_vars env e ->
  rew_map_once_adapting_env (fun env e ->
    let e' = match e with
    | Call(c, _) when is_orallt c ->
	let rec find_matching env matches non t = function
	  | [] -> (length matches, non), env
	  | e::l -> 
	      (*printf "creating rew2names_ %s %s\n" (rew2s e) (rew2s t) ; dumpt env ;*)
	      let bo = 
		try 
		  func_is_greater env `Greater (l2orallt c (matches @ [e])) t
		with Type_greater _ | Resolve_best_type _ | Interpreter.Parameter_number_mismatch _ -> false
	      in
	      if bo
	      then find_matching env (matches @ [e]) non t l
	      else find_matching env matches (non @ [e]) t l
	in 
	let rec group env found l =
	  let l2, env2 = map_withenv (fun env (var,t) -> 
	    let r, env' = find_matching env [] [] t l in
	    (var, r), env') env type_vars in
	  let l3 = sort (fun ((_,sa),(a,_)) ((_,sb),(b,_)) -> 
	    match compare b a with
	    | 0 -> compare sa sb
	    | n -> n
          ) l2 in
	  match l3 with
	  | [] -> found, env2
	  | (_,(0,_)) :: _ -> 
	      let l', env3 = map_withenv (rew2names_ type_vars) env2 l in
	      found @ l', env3
	  | ((t,_),(_,non)) :: _ -> group env2 (found @ [t]) non
	in 
	let found, env2 = group env [] (orallt2l c e) in
	if found = [] then None else Some(l2orallt c found, env2)
    | _ -> 
	try
	  let ((v,_),_), env' = find_withenv (fun env (_,t) -> 
	    let r, env' = simple_is_same_ env t e in
	    r, env'
	  ) env type_vars in
	  Some(v, env')
	with Not_found -> None
    in
    if debug && e' <> None then 
      printf "type_vars: replacing %s with %s\n" (rew2s e) (rew2s (fst (some e')));
    e'
  ) env e

and get_type_vars env =
  let l = ref [] in
  env_iter (fun var four ->
    let t = ! !(val_ four) in
    let starts_with_lowercase =
      !(nam_ four) <> "" && (
      let c = String.get !(nam_ four) 0 in
      c = Char.uppercase c)
    in
    if (fla_ four).is_type && (fla_ four).is_closed && not (is_special_constr t) && starts_with_lowercase then
      l := ((var, String.length (rew2s t)), mark_as_val t) :: !l
  ) env ;
  if debug then iter (fun ((var,_),t) -> printf "type_vars: %s !! %s\n" (rew2s var) (rew2s t)) !l ;
  !l

(*and rew2names env e = fst (rew2names_ (get_type_vars env) env e)*)

let infere_main env expr = 
  Interpreter.during_type_inference := true ;
  let valtyp, env = infere false None env expr in

  (* resolve the local variables still living in the stack *)
  let locals_sorted, env = resolve_aliased_vars env in
  let env = fold_left (fun env i -> 
    resolve_best_type `Greater `Unused env (Local((i,0,""),bpos))
  ) env locals_sorted in
  let env = propagate_stack env in

  valtyp, env



let help_interpreter_resolving_constraints env tf para = 
  if debug then printf "help_interpreter_resolving_constraints\n";
  match tf with
  | Func(nblocal, in_para, body) ->
      let env2 = push_stackt env nblocal { def_flags with para_or_local = `Local } in
      (try 
	let _, env3 = env_unpack_constr env2 body in

	if Flags.check then (match rew_collect_all_accessible_vars para with
	| [] -> ()
	| _ -> failwith "help_interpreter_resolving_constraints called with a variable");

	let para2 = rew_remap_locals 1 para in
	let paras = if length in_para > 1 then sep2l para2 else [para2] in
	if length in_para <> length paras then raise (Interpreter.Parameter_number_mismatch "");
	
	let in_para2, env4 = map_withenv (fun env e ->
	  let e2, _ = unlazy_and_alike e in
	  let io, e3 = rew2io e2 in
	  if is_local_var e3 then (e, env) else

	  let (i, env') = realloc_stack env 0 1 def_flags in
	  let var = io2rew (Local((i, 0, "i"), bpos)) io in
	  var, cooked_ensure_greater env' e var
	) env3 in_para in

	let env5 = fold_left2 cooked_ensure_greater env4 in_para2 paras in
	let locals_sorted, env6 = resolve_aliased_vars env5 in

	let locals_types, env7 = map_withenv (fun env i -> 
	  let (t, upper_t), env' = resolve_best_type_keep_abstract env (Local((i,0,""),bpos)) in
	  ((i, t), (i, upper_t)), env'
        ) env6 locals_sorted in

	let (typs, upper_typs) = map_t2 (fun locals_types ->
	  map (rew_map_leaf (function
	    | 0, Local((i,0,_),_) -> assoc i locals_types
	    | _, e -> e
	  )) in_para2
        ) (split locals_types) in

	let r = 
	  match collect rew_collect_all_accessible_vars_no_val (typs @ upper_typs) with
	  | [] -> Some(l2sep typs, l2sep upper_typs)
	  | _ -> None
	in
	r, pop_stackt env7
      with Type_greater _ | Resolve_best_type _ | Interpreter.Parameter_number_mismatch _ -> 
	None, pop_stackt env2)
      
  | _ -> failwith "bad"

;;
Interpreter.resolve_constraints := help_interpreter_resolving_constraints
